import logging
logger = logging.getLogger(__file__)

import requests
from http.client import responses
import os
import json
from datetime import datetime
from lofar.common.datetimeutils import formatDatetime

# usage example:
#
# with TMSSsession('paulus', 'pauluspass', 'localhost', 8000) as tmsssession:
#    response = tmsssession.session.get(url='http://localhost/api/task_draft/')
#    print(response)


#TODO: add unittests!
class TMSSsession(object):

    OPENID = "openid"
    BASICAUTH = "basicauth"

    def __init__(self, username, password, host, port: int=8000, authentication_method=OPENID):
        self.session = requests.session()
        self.username = username
        self.password = password
        self.base_url = "http://%s:%d/api" % (host, port)
        self.authentication_method = authentication_method

    @staticmethod
    def create_from_dbcreds_for_ldap(dbcreds_name: str=None):
        '''Factory method to create a TMSSSession object which uses the credentials in the ~/.lofar/dbcredentials/<dbcreds_name>.ini file
           (mis)use the DBCredentials to get a url with user/pass for tmss
           the contents below are used to contruct a url like this: http://localhost:8000/api
           [database:TMSS]
            host=localhost
            user=<username>
            password=<password>
            type=http
            port=8000
         '''
        if dbcreds_name is None:
            dbcreds_name = os.environ.get("TMSS_CLIENT_DBCREDENTIALS", "TMSSClient")

        from lofar.common.dbcredentials import DBCredentials
        dbcreds = DBCredentials().get(dbcreds_name)
        return TMSSsession(username=dbcreds.user, password=dbcreds.password,
                           host=dbcreds.host,
                           port=dbcreds.port,
                           authentication_method=TMSSsession.BASICAUTH)

    def __enter__(self):
        self.open()

        # return the request session for use within the context
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def open(self):
        '''open the request session and login'''
        self.session.__enter__()
        self.session.verify = False

        if self.authentication_method == self.OPENID:
            # get authentication page of OIDC through TMSS redirect
            response = self.session.get(self.base_url.replace('/api', '/oidc/authenticate/'), allow_redirects=True)
            csrftoken = self.session.cookies['csrftoken']

            # post user credentials to login page, also pass csrf token
            data = {'username': self.username, 'password': self.password, 'csrfmiddlewaretoken': csrftoken}
            response = self.session.post(url=response.url, data=data, allow_redirects=True)

            # raise when sth went wrong
            if "The username and/or password you specified are not correct" in response.content.decode('utf8'):
                raise ValueError("The username and/or password you specified are not correct")
            if response.status_code != 200:
                raise ConnectionError(response.content.decode('utf8'))

        if self.authentication_method == self.BASICAUTH:
            self.session.auth = (self.username, self.password)

    def close(self):
        '''close the request session and logout'''
        try:
            # logout user
            self.session.get(self.base_url + '/logout/', allow_redirects=True)
            self.session.close()
        except:
            pass

    def set_subtask_status(self, subtask_id: int, status: str) -> {}:
        '''set the status for the given subtask, and return the subtask with its new state, or raise on error'''
        response = self.session.patch(url='%s/subtask/%s/' % (self.base_url, subtask_id),
                                      json={'state': "%s/subtask_state/%s/" % (self.base_url, status)},
                                      params={'format':'json'})

        if response.status_code >= 200 and response.status_code < 300:
            return json.loads(response.content.decode('utf-8'))

        content = response.content.decode('utf-8')
        raise Exception("Could not set status with url %s - %s %s - %s" % (response.request.url, response.status_code, responses.get(response.status_code), content))

    def get_subtask_parset(self, subtask_id) -> str:
        '''get the lofar parameterset (as text) for the given subtask'''
        result = self.session.get(url='%s/subtask/%s/parset' % (self.base_url, subtask_id))
        if result.status_code >= 200 and result.status_code < 300:
            return result.content.decode('utf-8')
        raise Exception("Could not get parameterset for subtask %s.\nResponse: %s" % (subtask_id, result))

    def get_subtask_predecessors(self, subtask_id: int, state: str=None) -> list:
        '''get the subtask's predecessors as list of dict for the given subtask'''

        clauses = {}
        if state is not None:
            clauses["state__value"] = state

        path = 'subtask/%s/predecessors' % (subtask_id,)
        return self.get_path_as_json_object(path, clauses)

    def get_subtask_successors(self, subtask_id: int, state: str=None) -> list:
        '''get the subtask's successors as list of dict for the given subtask'''
        clauses = {}
        if state is not None:
            clauses["state__value"] = state

        path = 'subtask/%s/successors' % (subtask_id,)
        return self.get_path_as_json_object(path, clauses)

    def get_subtask(self, subtask_id: int) -> dict:
        '''get the subtask as dict for the given subtask'''
        path = 'subtask/%s' % (subtask_id,)
        return self.get_path_as_json_object(path)

    def get_subtasks(self, state: str=None,
                     cluster: str=None,
                     start_time_less_then: datetime=None, start_time_greater_then: datetime=None,
                     stop_time_less_then: datetime = None, stop_time_greater_then: datetime = None) -> list:
        '''get subtasks (as list of dicts) filtered by the given parameters'''
        clauses = {}
        if state is not None:
            clauses["state__value"] = state
        if cluster is not None:
            clauses["cluster__name"] = cluster
        if start_time_less_then is not None:
            clauses["start_time__lt="] = formatDatetime(start_time_less_then)
        if start_time_greater_then is not None:
            clauses["start_time__gt"] = formatDatetime(start_time_greater_then)
        if stop_time_less_then is not None:
            clauses["stop_time__lt"] = formatDatetime(stop_time_less_then)
        if stop_time_greater_then is not None:
            clauses["stop_time__gt"] = formatDatetime(stop_time_greater_then)

        return self.get_path_as_json_object("subtask", clauses)

    def get_path_as_json_object(self, path: str, params={}) -> object:
        '''get resource at the given path, interpret it as json, and return it as as native object (usually a dict or a list of dicts)'''
        full_url = '%s/%s/' % (self.base_url, path.strip('/'))
        return self.get_url_as_json_object(full_url, params=params)

    def get_url_as_json_object(self, full_url: str, params={}) -> object:
        '''get resource at the given full url (including http://<base_url>, interpret it as json, and return it as as native object (usually a dict or a list of dicts)'''
        if "format=json" not in full_url or params.get("format") != "json":
            params['format'] ='json'

        response = self.session.get(url=full_url, params=params)
        logger.info("%s %s %s on %s", response.request.method.upper(), response.status_code, responses.get(response.status_code), response.request.url)

        if response.status_code >= 200 and response.status_code < 300:
            result = json.loads(response.content.decode('utf-8'))
            if isinstance(result, dict):
                result_object = result.get('results', result) # return the 'results' list if any, or lese just the object itself

                if result.get('next'):
                    # recurse, get the 'next' url, and return a concatenation of the results
                    return result_object + self.get_url_as_json_object(result['next'])
                return result_object
            return result

        # ugly error message parsing
        content = response.content.decode('utf-8')
        try:
            error_msg = content.split('\n')[1] # magic! error message is at 2nd line of response...
        except:
            error_msg= content

        raise Exception("Could not get %s - %s %s - %s" % (full_url, response.status_code, responses.get(response.status_code), error_msg))

    def _get_template(self, template_type_name: str, name: str, version: str=None) -> dict:
        '''get the template of the given type as dict for the given name (and version)'''
        clauses = {}
        if name is not None:
            clauses["name"] = name
        if version is not None:
            clauses["version"] = version
        result = self.get_path_as_json_object(template_type_name, clauses)
        if isinstance(result, list):
            if len(result) > 1:
                raise ValueError("Found more then one %s for clauses: %s" % (template_type_name, clauses))
            elif len(result) == 1:
                return result[0]
            return None
        return result

    def get_schedulingunit_template(self, name: str, version: str=None) -> dict:
        '''get the schedulingunit_template as dict for the given name (and version)'''
        return self._get_template('scheduling_unit_template', name, version)

    def get_task_template(self, name: str, version: str=None) -> dict:
        '''get the task_template as dict for the given name (and version)'''
        return self._get_template('task_template', name, version)

    def get_subtask_template(self, name: str, version: str=None) -> dict:
        '''get the subtask_template as dict for the given name (and version)'''
        return self._get_template('subtask_template', name, version)

    def get_subtask_output_dataproducts(self,  subtask_id: int) -> []:
        '''get the output dataproducts of the subtask with the given subtask_id'''
        return self.get_path_as_json_object('subtask/%s/output_dataproducts' % subtask_id)

    def get_subtask_input_dataproducts(self,  subtask_id: int) -> []:
        '''get the input dataproducts of the subtask with the given subtask_id'''
        return self.get_path_as_json_object('subtask/%s/input_dataproducts' % subtask_id)

    def specify_observation_task(self, task_id: int) -> requests.Response:
        """specify observation for the given draft task by just doing a REST API call """
        result = self.session.get(url='%s/api/task/%s/specify_observation' % (self.base_url, task_id))
        if result.status_code >= 200 and result.status_code < 300:
            return result.content.decode('utf-8')
        raise Exception("Could not specify observation for task %s.\nResponse: %s" % (task_id, result))

    def create_blueprints_and_subtasks_from_scheduling_unit_draft(self, scheduling_unit_draft_id: int) -> {}:
        """create a scheduling_unit_blueprint, its specified taskblueprints and subtasks for the given scheduling_unit_draft_id.
        returns the scheduled subtask upon success, or raises."""
        return self.get_path_as_json_object('scheduling_unit_draft/%s/create_blueprints_and_subtasks' % scheduling_unit_draft_id)

    def schedule_subtask(self, subtask_id: int) -> {}:
        """schedule the subtask for the given subtask_id.
        returns the scheduled subtask upon success, or raises."""
        return self.get_path_as_json_object('subtask/%s/schedule' % subtask_id)

    def get_setting(self, setting_name: str) -> {}:
        """get the value of a TMSS setting.
        returns the setting value upon success, or raises."""
        response = self.session.get(url='%s/setting/%s/' % (self.base_url, setting_name),
                                    params={'format': 'json'})

        if response.status_code >= 200 and response.status_code < 300:
            return json.loads(response.content.decode('utf-8'))['value']

        content = response.content.decode('utf-8')
        raise Exception("Could not get setting with name %s.\nResponse: %s" % (setting_name, content))

    def set_setting(self, setting_name: str, setting_value: bool) -> {}:
        """set a value for a TMSS setting.
        returns the setting value upon success, or raises."""
        response = self.session.patch(url='%s/setting/%s/' % (self.base_url, setting_name),
                                      json={'value': setting_value})

        if response.status_code >= 200 and response.status_code < 300:
            return json.loads(response.content.decode('utf-8'))['value']

        content = response.content.decode('utf-8')
        raise Exception("Could not set status with url %s - %s %s - %s" % (response.request.url, response.status_code, responses.get(response.status_code), content))

