#!/usr/bin/env python3

# TMSSBusListener.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: TMSSBusListener.py 1580 2015-09-30 14:18:57Z loose $

"""
TMSSBusListener listens on the lofar notification message bus and calls (empty) on<SomeMessage> methods when such a message is received.
Typical usage is to derive your own subclass from TMSSBusListener and implement the specific on<SomeMessage> methods that you are interested in.
"""

from lofar.messaging.messagebus import BusListener, AbstractMessageHandler
from lofar.messaging import DEFAULT_BROKER, DEFAULT_BUSNAME, EventMessage
from lofar.common.util import waitForInterrupt, single_line_with_single_spaces

import logging
logger = logging.getLogger(__name__)


_DEFAULT_TMSS_NOTIFICATION_PREFIX_TEMPLATE = 'TMSS.%s.notification'
DEFAULT_TMSS_TASK_NOTIFICATION_PREFIX = _DEFAULT_TMSS_NOTIFICATION_PREFIX_TEMPLATE % 'Task'
DEFAULT_TMSS_SUBTASK_NOTIFICATION_PREFIX = _DEFAULT_TMSS_NOTIFICATION_PREFIX_TEMPLATE % 'SubTask'
DEFAULT_TMSS_ALL_NOTIFICATION_PREFIX = _DEFAULT_TMSS_NOTIFICATION_PREFIX_TEMPLATE + '#'


class TMSSSubTaskEventMessageHandler(AbstractMessageHandler):
    '''
    Base-type messagehandler for handling TMSS event messages.
    Typical usage is to derive your own subclass from TMSSSubTaskEventMessageHandler and implement the specific on<SomeMessage> methods that you are interested in.
    '''
    def handle_message(self, msg: EventMessage):
        if not isinstance(msg, EventMessage):
            raise ValueError("%s: Ignoring non-EventMessage: %s" % (self.__class__.__name__, msg))

        stripped_subject = msg.subject.replace("%s." % DEFAULT_TMSS_SUBTASK_NOTIFICATION_PREFIX, '')

        logger.info("%s.on%s: %s" % (self.__class__.__name__, stripped_subject, single_line_with_single_spaces(msg.content)))

        if stripped_subject == 'Defining':
            self.onSubTaskDefining(**msg.content)
        elif stripped_subject == 'Defined':
            self.onSubTaskDefined(**msg.content)
        elif stripped_subject == 'Scheduling':
            self.onSubTaskScheduling(**msg.content)
        elif stripped_subject == 'Scheduled':
            self.onSubTaskScheduled(**msg.content)
        elif stripped_subject == 'Queueing':
            self.onSubTaskQueueing(**msg.content)
        elif stripped_subject == 'Queued':
            self.onSubTaskQueued(**msg.content)
        elif stripped_subject == 'Starting':
            self.onSubTaskStarting(**msg.content)
        elif stripped_subject == 'Started':
            self.onSubTaskStarted(**msg.content)
        elif stripped_subject == 'Finishing':
            self.onSubTaskFinishing(**msg.content)
        elif stripped_subject == 'Finished':
            self.onSubTaskFinished(**msg.content)
        elif stripped_subject == 'Cancelling':
            self.onSubTaskCancelling(**msg.content)
        elif stripped_subject == 'Cancelled':
            self.onSubTaskCancelled(**msg.content)
        elif stripped_subject == 'Error':
            self.onSubTaskError(**msg.content)
        else:
            raise ValueError("TMSSBusListener.handleMessage: unknown subject: %s" %  msg.subject)

    def onSubTaskDefining(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskDefining is called upon receiving a SubTaskDefining message, which is sent when a SubTasks changes state to "Defining".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskDefined(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskDefined is called upon received a SubTaskDefined message, which is sent when a SubTasks changes state to "Defined".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskScheduling(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskScheduling is called upon receiving a SubTaskScheduling message, which is sent when a SubTasks changes state to "Scheduling".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskScheduled(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskScheduled is called upon received a SubTaskScheduled message, which is sent when a SubTasks changes state to "Scheduled".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskQueueing(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskQueueing is called upon receiving a SubTaskQueueing message, which is sent when a SubTasks changes state to "Queueing".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskQueued(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskQueued is called upon received a SubTaskQueued message, which is sent when a SubTasks changes state to "Queued".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskStarting(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskStarting is called upon receiving a SubTaskStarting message, which is sent when a SubTasks changes state to "Starting".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskStarted(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskStarted is called upon received a SubTaskStarted message, which is sent when a SubTasks changes state to "Started".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskFinishing(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskFinishing is called upon receiving a SubTaskFinishing message, which is sent when a SubTasks changes state to "Finishing".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskFinished(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskFinished is called upon received a SubTaskFinished message, which is sent when a SubTasks changes state to "Finished".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskCancelling(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskCancelling is called upon receiving a SubTaskCancelling message, which is sent when a SubTasks changes state to "Cancelling".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskCancelled(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskCancelled is called upon received a SubTaskCancelled message, which is sent when a SubTasks changes state to "Cancelled".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass

    def onSubTaskError(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskError is called upon receiving a SubTaskError message, which is sent when a SubTasks changes state to "Error".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        pass


class TMSSSubTaskBusListener(BusListener):
    def __init__(self,
                 handler_type: TMSSSubTaskEventMessageHandler.__class__ = TMSSSubTaskEventMessageHandler,
                 handler_kwargs: dict = None,
                 exchange: str = DEFAULT_BUSNAME,
                 routing_key: str = DEFAULT_TMSS_SUBTASK_NOTIFICATION_PREFIX+".#",
                 num_threads: int = 1,
                 broker: str = DEFAULT_BROKER):
        """
        TMSSSubTaskBusListener listens on the lofar notification message bus and calls on<SomeMessage> methods in the TMSSSubTaskEventMessageHandler when such a message is received.
        Typical usage is to derive your own subclass from TMSSSubTaskEventMessageHandler and implement the specific on<SomeMessage> methods that you are interested in.
        """
        if not issubclass(handler_type, TMSSSubTaskEventMessageHandler):
            raise TypeError("handler_type should be a TMSSSubTaskEventMessageHandler subclass")

        super().__init__(handler_type, handler_kwargs, exchange, routing_key, num_threads, broker)


if __name__ == '__main__':
    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    class ExampleTMSSSubTaskEventMessageHandler(TMSSSubTaskEventMessageHandler):
        def onSubTaskDefined(self, **kwargs):
            logger.debug("MyTMSSSubTaskEventMessageHandler.onSubTaskDefined(%s)", kwargs)

    with TMSSSubTaskBusListener(handler_type=ExampleTMSSSubTaskEventMessageHandler):
        waitForInterrupt()

