#!/usr/bin/env python3

# subtask_scheduling.py
#
# Copyright (C) 2015
# ASTRON (Netherlands Institute for Radio Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.
#
# $Id: subtask_scheduling.py 1580 2015-09-30 14:18:57Z loose $

"""
The subtask_scheduling service schedules TMSS subtasks.
It listens on the lofar notification message bus for state changes of TMSS subtasks; when a task finished,
it schedules (rest action) all successors that are in state 'defined'.
"""

import os
from optparse import OptionParser
import logging
logger = logging.getLogger(__name__)

from lofar.sas.tmss.client.tmssbuslistener import *
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession

class TMSSSubTaskSchedulingEventMessageHandler(TMSSSubTaskEventMessageHandler):
    '''
    '''
    def __init__(self, tmss_client_credentials_id: str=None):
        super().__init__()
        self.tmss_client = TMSSsession.create_from_dbcreds_for_ldap(tmss_client_credentials_id)

    def start_handling(self):
        self.tmss_client.open()
        super().start_handling()

    def stop_handling(self):
        super().stop_handling()
        self.tmss_client.close()

    def onSubTaskFinished(self, subtask_id: int, old_state: str, new_state:str):
        '''onSubTaskFinished is called upon received a SubTaskFinished message, which is sent when a SubTasks changes state to "Finished".
        :param subtask_id: the TMSS id of the SubTask
        :param old_state: the previous state of the SubTask
        :param new_state: the new state of the SubTask
        '''
        logger.info("subtask %s finished. Trying to schedule defined successor subtasks...", subtask_id)

        successors = self.tmss_client.get_subtask_successors(subtask_id, state="defined")
        successor_ids = [s['id'] for s in successors]

        logger.info("subtask %s finished. trying to schedule defined successors: %s",
                    subtask_id,
                    ', '.join(str(id) for id in successor_ids) or 'None')

        for successor in successors:
            try:
                suc_subtask_id = successor['id']
                suc_subtask_state = successor['state_value']

                if suc_subtask_state == "defined":
                    logger.info("trying to schedule successor subtask %s for finished subtask %s", suc_subtask_id, subtask_id)
                    scheduled_successor = self.tmss_client.schedule_subtask(suc_subtask_id)
                    suc_subtask_state = scheduled_successor['state_value']
                    logger.info("successor subtask %s for finished subtask %s now has state '%s', see %s", suc_subtask_id, subtask_id, suc_subtask_state, scheduled_successor['url'])
                else:
                    logger.warning("skipping scheduling of successor subtask %s for finished subtask %s because its state is '%s'", suc_subtask_id, subtask_id, suc_subtask_state)

            except Exception as e:
                logger.error(e)

def create_service(exchange: str=DEFAULT_BUSNAME, broker: str=DEFAULT_BROKER, tmss_client_credentials_id: str=None):
    return TMSSSubTaskBusListener(handler_type=TMSSSubTaskSchedulingEventMessageHandler,
                                  handler_kwargs={'tmss_client_credentials_id': tmss_client_credentials_id},
                                  exchange=exchange,
                                  broker=broker)

def main():
    # make sure we run in UTC timezone
    os.environ['TZ'] = 'UTC'

    logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

    # Check the invocation arguments
    parser = OptionParser('%prog [options]',
                          description='run the tmss_subtask_scheduling_service which automatically schedules the defined successor tasks for finished subtasks')
    parser.add_option('-q', '--broker', dest='broker', type='string', default=DEFAULT_BROKER, help='Address of the messaging broker, default: %default')
    parser.add_option('--exchange', dest='exchange', type='string', default=DEFAULT_BUSNAME, help='Name of the exchange on the messaging broker, default: %default')
    parser.add_option('-t', '--tmss_client_credentials_id', dest='tmss_client_credentials_id', type='string',
                      default=os.environ.get("TMSS_CLIENT_DBCREDENTIALS", "TMSSClient"),
                      help='the credentials id for the file in ~/.lofar/dbcredentials which holds the TMSS http REST api url and credentials, default: %default')
    (options, args) = parser.parse_args()

    with create_service(options.exchange, options.broker, options.tmss_client_credentials_id):
        waitForInterrupt()

if __name__ == '__main__':
    main()
