#!/usr/bin/env python3

# Copyright (C) 2012-2015  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

import unittest
import uuid

import logging
logger = logging.getLogger(__name__)

from lofar.sas.tmss.test.test_utils import TMSSTestEnvironment
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator

from lofar.messaging.messagebus import TemporaryExchange, BusListenerJanitor
from lofar.sas.tmss.services.subtask_scheduling import create_service
from lofar.common.test_utils import integration_test
from time import sleep
from datetime import datetime, timedelta

@integration_test
class TestSubtaskSchedulingService(unittest.TestCase):
    '''
    Tests for the SubtaskSchedulingService
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.TEST_UUID = uuid.uuid1()

        cls.tmp_exchange = TemporaryExchange("%s_%s" % (cls.__name__, cls.TEST_UUID))
        cls.tmp_exchange.open()

        cls.tmss_test_env = TMSSTestEnvironment(exchange=cls.tmp_exchange.address)
        cls.tmss_test_env.start()

        cls.test_data_creator = TMSSRESTTestDataCreator(cls.tmss_test_env.django_server.url,
                                                        (tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password))

    @classmethod
    def tearDownClass(cls) -> None:
        cls.tmss_test_env.stop()
        cls.tmp_exchange.close()

    def test_01_for_expected_behaviour(self):
        '''
        This test starts a scheduling service and tmss, creates a chain of subtasks, finishes the first, and checks if the successors are then scheduled.
        '''

        logger.info(' -- test_01_for_expected_behaviour -- ')

        # create and start the service (the object under test)
        service = create_service(exchange=self.tmp_exchange.address, tmss_client_credentials_id=self.tmss_test_env.client_credentials.dbcreds_id)
        with BusListenerJanitor(service):
            # -------------------------
            # long setup of objects....

            # setup proper template
            subtask_template_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.SubtaskTemplate(subtask_type_url=self.test_data_creator.django_api_url + '/subtask_type/qa_files/'), '/subtask_template/')

            # create two subtasks
            subtask1_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.Subtask(state="defined", specifications_template_url=subtask_template_url), '/subtask/')
            subtask2_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.Subtask(state="defined", specifications_template_url=subtask_template_url), '/subtask/')

            # ugly
            subtask1_id = subtask1_url.split('/')[subtask1_url.split('/').index('subtask') + 1]
            subtask2_id = subtask2_url.split('/')[subtask2_url.split('/').index('subtask') + 1]

            # connect them
            output_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.SubtaskOutput(subtask1_url), '/subtask_output/')
            input_url = self.test_data_creator.post_data_and_get_url(self.test_data_creator.SubtaskInput(subtask_url=subtask2_url, subtask_output_url=output_url), '/subtask_input/')

            # ... end of long setup of objects
            # --------------------------------

            # now for the real test: set subtask1_id status to finished, and check that subtask2 is then properly scheduled
            with self.tmss_test_env.create_tmss_client() as tmss_client:
                subtask1 = tmss_client.get_subtask(subtask1_id)
                subtask2 = tmss_client.get_subtask(subtask2_id)

                self.assertEqual(subtask1['state_value'], 'defined')
                self.assertEqual(subtask2['state_value'], 'defined')

                # the first subtask ran, and is now finished... set it's status. This should trigger the scheduling service to schedule the second subtask.
                tmss_client.set_subtask_status(subtask1_id, 'finished')

                # allow some time for the scheduling service to do its thing...
                start = datetime.utcnow()
                while subtask2['state_value'] != 'scheduled':
                    subtask2 = tmss_client.get_subtask(subtask2_id)
                    sleep(0.5)
                    if datetime.utcnow() - start >  timedelta(seconds=2):
                        raise TimeoutError()

                # subtask2 should now be scheduled
                self.assertEqual(subtask2['state_value'], 'scheduled')

logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

if __name__ == '__main__':
    #run the unit tests
    unittest.main()
