#!/usr/bin/python3

# Copyright (C) 2020  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.parameterset import parameterset
from lofar.common.datetimeutils import formatDatetime
from lofar.common.json_utils import add_defaults_to_json_object_for_schema
from lofar.sas.tmss.tmss.exceptions import *
from datetime import datetime

def _convert_to_parset_dict_for_observationcontrol_schema(subtask: models.Subtask) -> dict:
    # make sure the spec is complete (including all non-filled in properties with default)
    spec = add_defaults_to_json_object_for_schema(subtask.specifications_doc, subtask.specifications_template.schema)

    # -----------------------------------------------------------------------------------------------
    # Historic rationale: in TMSS-183 we made MAC run an actual observation from a TMSS specification.
    # With the help of Auke and Jan-David I could generate the parset as defined below.
    # MAC turned out to be very sensitive for having specific keys with very specific prefixes etc.
    # As a result, the generated parset contains many "duplicate"(nested) keys.
    # We all agree that this is ugly, and we should not want this, but hey... it works.
    # We decided to keep it like this, and maybe do more tuning/pruning later in the TMSS project.
    # Or, we can just get rid of this to-parset-adaper when MAC has been rewritten to the new station API.
    # -----------------------------------------------------------------------------------------------


    parset = dict() # parameterset has no proper assignment operators, so take detour via dict...
    parset["Observation.ObsID"] = subtask.pk
    parset["Observation.momID"] = 0 # Needed by MACScheduler
    parset["Observation.otdbID"] = 0 # Needed by MACScheduler; should/can this be the same as subtask.pk?
    parset["Observation.tmssID"] = subtask.pk
    parset["Observation.processType"] = subtask.specifications_template.type.value.capitalize()
    parset["Observation.processSubtype"] = "Beam Observation" # TODO: where to derive the processSubtype from?
    parset["Observation.Campaign.name"] = subtask.task_blueprint.scheduling_unit_blueprint.draft.scheduling_set.project.name
    parset["Observation.startTime"] = formatDatetime(subtask.start_time) if isinstance(subtask.start_time, datetime) else subtask.start_time
    parset["Observation.stopTime"] = formatDatetime(subtask.stop_time) if isinstance(subtask.stop_time, datetime) else subtask.stop_time
    parset["Observation.VirtualInstrument.minimalNrStations"] = 1  # maybe not mandatory?
    parset["Observation.VirtualInstrument.stationSet"] = "Custom"  # maybe not mandatory?
    parset["Observation.VirtualInstrument.stationList"] = "[%s]" % ','.join(s for s in spec["stations"]["station_list"])
    parset["Observation.antennaArray"] = "HBA" if "HBA" in spec["stations"]["antenna_set"] else "LBA" # maybe not mandatory?
    parset["Observation.antennaSet"] = spec["stations"]["antenna_set"]
    parset["Observation.bandFilter"] = spec["stations"]["filter"]
    parset["Observation.sampleClock"] = 200 # why is this not part of the schema? for example as a required setting with a single allowed value.
    parset["Observation.nrBitsPerSample"] = 8 # why is this not part of the schema? for example as a required setting with a single allowed value.
    parset["Observation.strategy"] = "default"  # maybe not mandatory?

    digi_beams = spec['stations']['digital_pointings']
    parset["Observation.nrBeams"] = len(digi_beams)
    for beam_nr, digi_beam in enumerate(digi_beams):
        beam_prefix = "Observation.Beam[%d]." % beam_nr
        parset[beam_prefix+"directionType"] = digi_beam['pointing']['direction_type']
        parset[beam_prefix+"angle1"] = digi_beam['pointing']['angle1']
        parset[beam_prefix+"angle2"] = digi_beam['pointing']['angle2']
        parset[beam_prefix+"target"] = digi_beam['name']
        parset[beam_prefix+"subbandList"] = digi_beam['subbands']

        phase_centers = spec['COBALT']['correlator']['phase_centers']
        if phase_centers:
            # for now, cobalt can handle only one phase_center
            # assume the first is the one
            phase_center = phase_centers[0]
            parset[beam_prefix+"Correlator.phaseCenterOverride"] = phase_center['index'] == beam_nr
            parset[beam_prefix+"Correlator.directionType"] = phase_center['pointing']['direction_type']
            parset[beam_prefix+"Correlator.angle1"] = phase_center['pointing']['angle1']
            parset[beam_prefix+"Correlator.angle2"] = phase_center['pointing']['angle2']

    analog_beam = spec['stations']['analog_pointing']
    parset["Observation.nrAnaBeams"] = 1
    beam_prefix = "Observation.AnaBeam[0]."
    parset[beam_prefix+"directionType"] = analog_beam['direction_type']
    parset[beam_prefix+"angle1"] = analog_beam['angle1']
    parset[beam_prefix+"angle2"] = analog_beam['angle2']

    for prefix in ["", "Observation.ObservationControl.OnlineControl."]:
        parset[prefix+"Cobalt.realTime"] = True
        parset[prefix+"Cobalt.blockSize"] = spec['COBALT']['blocksize']
        parset[prefix+"Cobalt.correctBandPass"] = spec['COBALT']['bandpass_correction']
        parset[prefix+"Cobalt.delayCompensation"] = spec['COBALT']['delay_compensation']

        parset[prefix+"Cobalt.Correlator.nrChannelsPerSubband"] = spec['COBALT']['correlator']['channels_per_subband']
        parset[prefix+"Cobalt.Correlator.nrBlocksPerIntegration"] = spec['COBALT']['correlator']['blocks_per_integration']
        parset[prefix+"Cobalt.Correlator.nrIntegrationsPerBlock"] = spec['COBALT']['correlator']['integrations_per_block']

    parset["Observation.Cluster.ProcessingCluster.clusterName"] = subtask.cluster.name
    
    parset["Observation.DataProducts.Output_Correlated.enabled"] = True
    parset["Observation.DataProducts.Output_Correlated.storageClusterName"] = subtask.cluster.name
    parset["Observation.DataProducts.Output_Correlated.storageClusterPartition"] = "/data/test-projects"
    parset["Observation.DataProducts.Output_Correlated.filenames"] = []
    parset["Observation.DataProducts.Output_Correlated.locations"] = []
    # TODO: do not use SubtaskOutput.objects.filter but make subtask.subtask_outputs work
    subtask_outputs = list(models.SubtaskOutput.objects.filter(subtask_id=subtask.id))
    # TODO don't we have to append to dataproducts here and then fill in the combined list in the end?
    for output_nr, subtask_output in enumerate(subtask_outputs):
        dataproducts = list(models.Dataproduct.objects.filter(producer_id=subtask_output.id))
        parset["Observation.DataProducts.Output_Correlated.filenames"] = "[%s]" % ','.join(dp.filename for dp in dataproducts)
        parset["Observation.DataProducts.Output_Correlated.locations"] = "[%s]" % ','.join("%s:%s" % (subtask.cluster.name, dp.directory) for dp in dataproducts)
        # mimic MoM placeholder thingy (the resource assigner parses this)
        parset["Observation.DataProducts.Output_Correlated.identifications"] = "[TMSS_subtask_%s.SAP%03d]" % (subtask.id, output_nr)

    # various additional 'Control' settings which seem to be needed for MAC
    parset["prefix"] = "LOFAR."
    parset["Observation.claimPeriod"] = 35
    parset["Observation.preparePeriod"] = 20
    for prefix in ["", "Observation."]:
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.CorrProc._executable"] = "CN_Processing"
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.CorrProc._hostname"] = "cbmmaster"
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.CorrProc._nodes"] = []
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.CorrProc._startstopType"] = "bgl"
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.CorrProc.workingdir"] = "/opt/lofar/bin/"
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl._hostname"] = "cbmmaster"
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.extraInfo"] = '["PIC","Cobalt"]'
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.procesOrder"] = []
        parset[prefix+"ObservationControl.OnlineControl.CorrAppl.processes"] = '["CorrProc"]'
        parset[prefix+"ObservationControl.OnlineControl._hostname"] = 'CCU001'
        parset[prefix+"ObservationControl.OnlineControl.applOrder"] = '["CorrAppl"]'
        parset[prefix+"ObservationControl.OnlineControl.applications"] = '["CorrAppl"]'
        parset[prefix+"ObservationControl.OnlineControl.inspectionHost"] = 'head01.cep4.control.lofar'
        parset[prefix+"ObservationControl.OnlineControl.inspectionProgram"] = 'inspection-plots-observation.sh'
        parset[prefix+"ObservationControl.StationControl._hostname"] = parset["Observation.VirtualInstrument.stationList"]
        parset[prefix+"ObservationControl.StationControl.aartfaacPiggybackAllowed"] = False
        parset[prefix+"ObservationControl.StationControl.tbbPiggybackAllowed"] = False

    return parset


def _convert_to_parset_dict_for_pipelinecontrol_schema(subtask: models.Subtask) -> dict:
    # see https://support.astron.nl/confluence/pages/viewpage.action?spaceKey=TMSS&title=UC1+JSON

    # make sure the spec is complete (including all non-filled in properties with default)
    spec = add_defaults_to_json_object_for_schema(subtask.specifications_doc, subtask.specifications_template.schema)

    # -----------------------------------------------------------------------------------------------
    # Historic rationale: in TMSS-183 we made MAC run an actual observation from a TMSS specification.
    # With the help of Auke and Jan-David I could generate the parset as defined below.
    # MAC turned out to be very sensitive for having specific keys with very specific prefixes etc.
    # As a result, the generated parset contains many "duplicate"(nested) keys.
    # We all agree that this is ugly, and we should not want this, but hey... it works.
    # We decided to keep it like this, and maybe do more tuning/pruning later in the TMSS project.
    # Or, we can just get rid of this to-parset-adaper when MAC has been rewritten to the new station API.
    # -----------------------------------------------------------------------------------------------

    parset = dict()  # parameterset has no proper assignment operators, so take detour via dict...

    # General
    parset["prefix"] = "LOFAR."
    parset["Observation.processType"] = "Pipeline"
    parset["Observation.processSubtype"] = "Averaging Pipeline" # TODO: where to derive the processSubtype from?
    parset["Observation.ObservationControl.PythonControl.pythonProgram"] = "preprocessing_pipeline.py"
    parset["Observation.ObservationControl.PythonControl.softwareVersion"] = ""
    parset["Observation.Campaign.name"] = subtask.task_blueprint.scheduling_unit_blueprint.draft.scheduling_set.project.name
    parset["Observation.Scheduler.taskName"] = subtask.task_blueprint.name
    parset["Observation.Scheduler.predecessors"] = []
    parset["Observation.Cluster.ProcessingCluster.clusterName"] = subtask.cluster.name
    parset["Observation.Cluster.ProcessingCluster.clusterPartition"] = 'cpu'
    parset["Observation.Cluster.ProcessingCluster.numberOfTasks"] = 110 # TODO: compute numberOfTasks
    parset["Observation.Cluster.ProcessingCluster.numberOfCoresPerTask"] = 2 # TODO: compute numberOfCoresPerTask

    # DPPP steps
    dppp_steps = []
    if "preflagger0" in spec:
        dppp_steps.append('preflagger[0]')
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].chan"] = "[%s]" % spec["preflagger0"]["channels"]
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].abstime"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].azimuth"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].baseline"] = ""
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].blrange"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].corrtype"] = ""
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].count.path"] = "-"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].count.save"] = "false"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].elevation"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].expr"] = ""
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].freqrange"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].lst"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].reltime"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].timeofday"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].timeslot"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[0].type"] = "preflagger"

    if 'preflagger1' in spec:
        dppp_steps.append('preflagger[1]')
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].corrtype"] = spec["preflagger1"]["corrtype"]
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].abstime"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].azimuth"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].baseline"] = ""
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].blrange"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].chan"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].count.path"] = "-"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].count.save"] = "false"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].elevation"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].expr"] = ""
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].freqrange"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].lst"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].reltime"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].timeofday"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].timeslot"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.preflagger[1].type"] = "preflagger"

    if 'aoflagger' in spec:
        dppp_steps.append('aoflagger')
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.strategy"] = spec["aoflagger"]["strategy"]
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.autocorr"] = "F"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.count.path"] = "-"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.count.save"] = "FALSE"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.keepstatistics"] = "T"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.memorymax"] = "10"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.memoryperc"] = "0"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.overlapmax"] = "0"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.overlapperc"] = "0"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.pedantic"] = "F"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.pulsar"] = "F"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.timewindow"] = "0"
        parset["Observation.ObservationControl.PythonControl.DPPP.aoflagger.type"] = "aoflagger"

    if "demixer" in spec:
        dppp_steps.append('demixer')
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.baseline"] = spec["demixer"]["baselines"]
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.demixfreqstep"] = spec["demixer"]["demix_frequency_steps"]
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.demixtimestep"] = spec["demixer"]["demix_time_steps"]
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.freqstep"] = spec["demixer"]["frequency_steps"]
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.timestep"] = spec["demixer"]["time_steps"]
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.ignoretarget"] = spec["demixer"]["ignore_target"]
        parset["Observation.ObservationControl.PythonControl.PreProcessing.demix_always"] = spec["demixer"]["demix_always"]
        parset["Observation.ObservationControl.PythonControl.PreProcessing.demix_if_needed"] = spec["demixer"]["demix_if_needed"]
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.blrange"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.corrtype"] = "cross"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.elevationcutoff"] = "0.0deg"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.instrumentmodel"] = "instrument"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.modelsources"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.ntimechunk"] = "0"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.othersources"] = "[]"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.skymodel"] = "sky"
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.subtractsources"] = ""
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.targetsource"] = ""
        parset["Observation.ObservationControl.PythonControl.DPPP.demixer.type"] = "demixer"

    parset["Observation.ObservationControl.PythonControl.DPPP.steps"] = "[%s]" % ",".join(dppp_steps)
    parset["Observation.ObservationControl.PythonControl.DPPP.msout.storagemanager.name"] = spec["storagemanager"] # todo: needs to be emptystring when standard/basic/non-dysco?

    # Dataproducts
    parset["Observation.DataProducts.Input_Correlated.enabled"] = "true"

    in_dataproducts = []
    for input_nr, subtask_input in enumerate(subtask.inputs.all()):
        in_dataproducts = subtask_input.dataproducts.all()
        parset["Observation.DataProducts.Input_Correlated.filenames"] = "[%s]" % ",".join([dp.filename for dp in in_dataproducts])
        parset["Observation.DataProducts.Input_Correlated.locations"] = "[%s]" % ",".join(["%s:%s" % (subtask.cluster.name, dp.directory) for dp in in_dataproducts])
        # mimic MoM placeholder thingy (the resource assigner parses this)
        # should be expanded with SAPS and datatypes
        parset["Observation.DataProducts.Input_Correlated.identifications"] = "[TMSS_subtask_%s.SAP%03d]" % (subtask_input.producer.subtask.id, input_nr)

    parset["Observation.DataProducts.Input_Correlated.skip"] = "[%s]" % ",".join(['0']*len(in_dataproducts))

    # TODO: do not use SubtaskOutput.objects.filter but make subtask.subtask_outputs work
    subtask_outputs = list(models.SubtaskOutput.objects.filter(subtask_id=subtask.id))
    out_dataproducts = []
    for subtask_output in subtask_outputs:
        out_dataproducts = list(models.Dataproduct.objects.filter(producer_id=subtask_output.id))

    parset["Observation.DataProducts.Output_Correlated.enabled"] = "true"
    parset["Observation.DataProducts.Output_Correlated.filenames"] = "[%s]" % ",".join([dp.filename for dp in out_dataproducts])
    parset["Observation.DataProducts.Output_Correlated.locations"] = "[%s]" % ",".join(["%s:%s" % (subtask.cluster.name, dp.directory) for dp in out_dataproducts])
    parset["Observation.DataProducts.Output_Correlated.skip"] = "[%s]" % ",".join(['0']*len(out_dataproducts))
    parset["Observation.DataProducts.Output_Correlated.identifications"] = "[TMSS_subtask_%s.SAP%03d]" % (subtask.id, 0)
    parset["Observation.DataProducts.Output_Correlated.storageClusterName"] = subtask.cluster.name

    # Other
    parset["Observation.ObservationControl.PythonControl.PreProcessing.SkyModel"] = "Ateam_LBA_CC"
    parset["Observation.ObservationControl.PythonControl.DPPP.checkparset"] = "-1"

    parset["Observation.ObservationControl.PythonControl.DPPP.msin.autoweight"] = "true"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.band"] = "-1"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.baseline"] = ""
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.blrange"] = "[]"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.corrtype"] = ""
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.datacolumn"] = "DATA"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.forceautoweight"] = "false"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.missingdata"] = "false"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.nchan"] = "nchan"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.orderms"] = "false"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.sort"] = "false"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.startchan"] = "0"
    parset["Observation.ObservationControl.PythonControl.DPPP.msin.useflag"] = "true"
    parset["Observation.ObservationControl.PythonControl.DPPP.msout.overwrite"] = "false"
    parset["Observation.ObservationControl.PythonControl.DPPP.msout.tilenchan"] = "8"
    parset["Observation.ObservationControl.PythonControl.DPPP.msout.tilesize"] = "4096"
    parset["Observation.ObservationControl.PythonControl.DPPP.msout.vdsdir"] = "A"
    parset["Observation.ObservationControl.PythonControl.DPPP.msout.writefullresflag"] = "true"

    parset["Observation.ObservationControl.PythonControl.DPPP.showprogress"] = "F"
    parset["Observation.ObservationControl.PythonControl.DPPP.showtimings"] = "F"
    parset["Observation.ObservationControl.PythonControl.DPPP.uselogger"] = "T"

    return parset


# dict to store conversion methods based on subtask.specifications_template.name
_convertors = {'observationcontrol schema': _convert_to_parset_dict_for_observationcontrol_schema,
               'pipelinecontrol schema': _convert_to_parset_dict_for_pipelinecontrol_schema }


def convert_to_parset(subtask: models.Subtask) -> parameterset:
    '''
    Convert the specifications in the subtask to a LOFAR parset for MAC/COBALT
    :raises ConversionException if no proper conversion is available.
    '''
    return parameterset(convert_to_parset_dict(subtask))

def convert_to_parset_dict(subtask: models.Subtask) -> dict:
    '''
    Convert the specifications in the subtask to a LOFAR parset dict with typed values for MAC/COBALT
    :raises ConversionException if no proper conversion is available.
    '''
    try:
        convertor = _convertors[subtask.specifications_template.name]
    except KeyError:
        raise ConversionException("Cannot convert subtask id=%d to parset. No conversion routine available for specifications_template='%s'" % (
                                  subtask.id, subtask.specifications_template.name))

    return convertor(subtask)
