"""
This file contains the database models
"""

from django.db.models import Model, CharField, DateTimeField, BooleanField, ForeignKey, CASCADE, IntegerField, FloatField, SET_NULL, PROTECT, ManyToManyField, UniqueConstraint, QuerySet
from django.contrib.postgres.fields import ArrayField, JSONField
from django.contrib.postgres.indexes import GinIndex
from enum import Enum
from django.db.models.expressions import RawSQL
from django.db.models.deletion import ProtectedError
from lofar.sas.tmss.tmss.tmssapp.validation import validate_json_against_schema
from django.core.exceptions import ValidationError
from rest_framework import status
import datetime

#
# Common
#

# abstract models

class BasicCommon(Model):
    # todo: we cannot use foreign keys in the array here, so we have to keep the Tags table up to date by trigger or so.
    # todo: we could switch to a manytomany field instead?
    tags = ArrayField(CharField(max_length=128), size=8, blank=True, help_text='User-defined search keywords for object.', default=list)
    created_at = DateTimeField(auto_now_add=True, help_text='Moment of object creation.')
    updated_at = DateTimeField(auto_now=True, help_text='Moment of last object update.')

    class Meta:
        abstract = True
        indexes = [GinIndex(fields=['tags'])]


class NamedCommon(BasicCommon):
    name = CharField(max_length=128, help_text='Human-readable name of this object.', null=False) # todo: check if we want to have this primary_key=True
    description = CharField(max_length=255, help_text='A longer description of this object.')

    def __str__(self):
        return self.name

    class Meta:
        abstract = True


# todo: check if we want to generally use this flavour, i.e. make everything named addressable by name rather than int id. (This then does not allow for multiple items of same name, of course.)
class NamedCommonPK(NamedCommon):
    name = CharField(max_length=128, help_text='Human-readable name of this object.', null=False, primary_key=True)

    class Meta:
        abstract = True


# concrete models

class Tags(Model):
    # todo: figure out how to keep this in sync with tags columns (->BasicCommon)
    # todo: Or remove this altogether without keeping track of tags?
    title = CharField(max_length=128)
    description = CharField(max_length=255)


#
# I/O
#

# choices

class AbstractChoice(Model):
    """
    Abstract class for all derived 'choices' models.
    We define a 'choice' as an item that you can pick from a predefined list.
    In the derived classes, we use an enum.Enum to define such a predefined list here in code.
    All values of the enums are then put automagically into the database in the populate module, which
    is/can_be/should_be called in the last migration step to populate the database with inital values
    for our 'static choices'.

    Design decision: Django also provides the 'choices' property on fields which sort of limits the number of choices
    one can make, and which sort of does some validation. In our opinion the validation is done in the wrong place, and
    no data consistency is enforced.
    So, we decided to follow Django's own hint, see https://docs.djangoproject.com/en/2.0/ref/models/fields/#choices
    "you’re probably better off using a proper database table with a ForeignKey"

    You can find the derived AbstractChoice classes being used as ForeignKey in other models, thus enforcing data
    consistency at database level.
    """
    value = CharField(max_length=128, editable=True, null=False, blank=False, unique=True, primary_key=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.value


class Role(AbstractChoice):
    """Defines the model and predefined list of possible Role's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        CORRELATOR = "correlator"
        BEAMFORMER = "beamformer"
        INSPECTION_PLOTS = "inspection plots"
        CALIBRATOR = "calibrator"
        TARGET = "target"
        INPUT = "input"
        OUTPUT = "output"


class Datatype(AbstractChoice):
    """Defines the model and predefined list of possible Datatype's for TaskConnectorType.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        VISIBILITIES = "visibilities"
        TIME_SERIES = "time series"
        INSTRUMENT_MODEL = "instrument model"
        IMAGE = "image"
        QUALITY = "quality"


class Dataformat(AbstractChoice):
    """Defines the model and predefined list of possible Dataformat's for TaskRelationDraft and TaskRelationBlueprint.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        MEASUREMENTSET = "MeasurementSet"
        BEAMFORMED = "Beamformed"
        QA_HDF5 = "QA_HDF5"
        QA_PLOTS = "QA_Plots"


class CopyReason(AbstractChoice):
    """Defines the model and predefined list of possible CopyReason's for SchedulingUnitDraft and TaskDraft.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        TEMPLATE = "template"
        REPEATED = "repeated"


class SchedulingRelationPlacement(AbstractChoice):
    """Defines the model and predefined list of possible Placements for Task Scheduling Relation.
        The items in the Choices class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        AFTER = "after"
        BEFORE = "before"
        PARALLEL = "parallel"


class Flag(AbstractChoice):
    """Defines the model and predefined list of possible Flags to be used in Setting.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        AUTOSCHEDULE = "allow_scheduling_observations"


class PeriodCategory(AbstractChoice):
    """Defines the model and predefined list of possible period categories to be used in Project.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        SINGLE_CYCLE = "single_cycle"
        LONG_TERM = "long_term"
        UNBOUNDED = "unbounded"


class ProjectCategory(AbstractChoice):
    """Defines the model and predefined list of possible project categories to be used in Project.
        The items in the Choices class below are automagically populated into the database via a data migration."""

    class Choices(Enum):
        REGULAR = "regular"
        USER_SHARED_SUPPORT = "user_shared_support"
        COMMISSIONING = "commissioning"
        DDT = "ddt"
        TEST = "test"


# concrete models

class Setting(BasicCommon):
    name = ForeignKey('Flag', null=False, on_delete=PROTECT, unique=True, primary_key=True)
    value = BooleanField(null=False)


class TaskConnectorType(BasicCommon):
    role = ForeignKey('Role', null=False, on_delete=PROTECT)
    datatype = ForeignKey('Datatype', null=False, on_delete=PROTECT)
    dataformats = ManyToManyField('Dataformat', blank=True)
    output_of = ForeignKey("TaskTemplate", related_name='output_connector_types', on_delete=CASCADE)
    input_of = ForeignKey("TaskTemplate", related_name='inpput_connector_types', on_delete=CASCADE)


#
# Templates
#

# abstract models

class Template(NamedCommon):
    version = CharField(max_length=128, help_text='Version of this template (with respect to other templates of the same name).')
    schema = JSONField(help_text='Schema for the configurable parameters needed to use this template.')

    class Meta:
        abstract = True
        # TODO: remove all <class>_unique_name_version UniqueConstraint's from the subclasses and replace by this line below when we start using django 3.0
        # constraints = [UniqueConstraint(fields=['name', 'version'], name='%(class)s_unique_name_version')]

# concrete models

class GeneratorTemplate(Template):
    create_function = CharField(max_length=128, help_text='Python function to call to execute the generator.')

    class Meta:
        # TODO: move up to the abstract base class and replace with django 3.0 UniqueConstraint(... name='%*class)s_unique_name_version)
        constraints = [UniqueConstraint(fields=['name', 'version'], name='GeneratorTemplate_unique_name_version')]


class DefaultGeneratorTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("GeneratorTemplate", on_delete=PROTECT)


class SchedulingUnitTemplate(Template):
    class Meta:
        # TODO: move up to the abstract base class and replace with django 3.0 UniqueConstraint(... name='%*class)s_unique_name_version)
        constraints = [UniqueConstraint(fields=['name', 'version'], name='SchedulingUnitTemplate_unique_name_version')]


class DefaultSchedulingUnitTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("SchedulingUnitTemplate", on_delete=PROTECT)


class TaskTemplate(Template):
    validation_code_js = CharField(max_length=128, help_text='JavaScript code for additional (complex) validation.')

    class Meta:
        # TODO: move up to the abstract base class and replace with django 3.0 UniqueConstraint(... name='%*class)s_unique_name_version)
        constraints = [UniqueConstraint(fields=['name', 'version'], name='TaskTemplate_unique_name_version')]

class DefaultTaskTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("TaskTemplate", on_delete=PROTECT)


class TaskRelationSelectionTemplate(Template):
    class Meta:
        # TODO: move up to the abstract base class and replace with django 3.0 UniqueConstraint(... name='%*class)s_unique_name_version)
        constraints = [UniqueConstraint(fields=['name', 'version'], name='TaskRelationSelectionTemplate_unique_name_version')]


class DefaultTaskRelationSelectionTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("TaskRelationSelectionTemplate", on_delete=PROTECT)


#
# Instance Objects
#

class Cycle(NamedCommonPK):
    start = DateTimeField(help_text='Moment at which the cycle starts, that is, when its projects can run.')
    stop = DateTimeField(help_text='Moment at which the cycle officially ends.')

    @property
    def duration(self) -> datetime.timedelta:
        '''the duration of the cycle (stop-start date)'''
        return self.stop - self.start

    # Project to Cycle references are ManyToMany now, so cannot be protected via on_delete on db level.
    # Instead, explicitly check whether there are projects linked to the Cycle and prevent delete
    # todo: verify correct behavior
    #  - Implemented: Cycles cannot be deleted if they have projects assigned to them
    #  - Also possible: Projects need at least one cycle
    def delete(self, *args, **kwargs):
        if len(self.projects.all()) > 0:
            raise ProtectedError(protected_objects=self.projects.all(), msg='This Cycle is referenced by %s project(s) and cannot be deleted.' % len(self.projects.all()))
        else:
            super().delete(*args, **kwargs)


class CycleQuota(Model):
    cycle = ForeignKey('Cycle', related_name="cycle_quota", on_delete=PROTECT, help_text='Cycle to which these quota apply.')
    value = FloatField(help_text='Resource Quota value')
    resource_type = ForeignKey('ResourceType', on_delete=PROTECT, help_text='Resource type.')


class Project(NamedCommonPK):
    cycles = ManyToManyField('Cycle', blank=True, related_name='projects', help_text='Cycles to which this project belongs (NULLable).')
    priority_rank = FloatField(null=False, help_text='Priority of this project w.r.t. other projects. Projects can interrupt observations of lower-priority projects.') # todo: add if needed: validators=[MinValueValidator(0.0), MaxValueValidator(1.0)]
    trigger_priority = IntegerField(default=1000, help_text='Priority of this project w.r.t. triggers.') # todo: verify meaning and add to help_text: "Triggers with higher priority than this threshold can interrupt observations of projects."
    can_trigger = BooleanField(default=False, help_text='True if this project is allowed to supply observation requests on the fly, possibly interrupting currently running observations (responsive telescope).')
    private_data = BooleanField(default=True, help_text='True if data of this project is sensitive. Sensitive data is not made public.')
    expert = BooleanField(default=False, help_text='Expert projects put more responsibility on the PI.')
    filler = BooleanField(default=False, help_text='Use this project to fill up idle telescope time.')
    project_category = ForeignKey('ProjectCategory', null=True, on_delete=PROTECT, help_text='Project category.')
    period_category = ForeignKey('PeriodCategory', null=True, on_delete=PROTECT, help_text='Period category.')

    # JK, 29/07/20 - after discussion with Sander, it turns out that the ticket TMSS-277 was a misunderstanding.
    #  'default' does not refer to 'default values' that are supposed to be filled in by the backend.
    #  It was meant to be 'resource_types displayed in the frontend by default', where the other resource_types are
    #  optionally added to the set of quota. These can then be customized in the frontend and are created by the
    #  frontend in the backend, but no quota are intended to be added automatically. So nothing is really  needed in
    #  the backend for this (apart from the set of predefined resource_types).
    #  There was some open question on whether there may be a required subset of quota that have to be enforced. So
    #  I'll leave this in for now, until that question is cleared up.
    #
    # # also create default project quotas when projects are created
    # def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
    #     creating = self._state.adding  # True on create, False on update
    #     super().save(force_insert, force_update, using, update_fields)
    #     if creating:
    #         # todo: review these defaults for being sensible
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="lta_storage"), value=1024^4, project=self)
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="cep_storage"), value=1024^4, project=self)
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="cep_processing_time"), value=60*60*24, project=self)
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="lofar_observing_time"), value=60*60*24, project=self)
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="lofar_observing_time_prio_a"), value=60*60*12, project=self)
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="lofar_observing_time_prio_b"), value=60*60*12, project=self)
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="support_time"), value=60*60*6, project=self)
    #         ProjectQuota.objects.create(resource_type=ResourceType.objects.get(name="number_of_triggers"), value=42, project=self)


class ProjectQuota(Model):
    project = ForeignKey('Project', related_name="project_quota", on_delete=PROTECT, help_text='Project to wich this quota belongs.')  # protected to avoid accidents
    value = FloatField(help_text='Resource Quota value')
    resource_type = ForeignKey('ResourceType', on_delete=PROTECT, help_text='Resource type.')  # protected to avoid accidents


class ResourceType(NamedCommonPK):
    pass


class SchedulingSet(NamedCommon):
    generator_doc = JSONField(null=True, help_text='Parameters for the generator (NULLable).')
    generator_template = ForeignKey('GeneratorTemplate', on_delete=SET_NULL, null=True, help_text='Generator for the scheduling units in this set (NULLable).')
    generator_source = ForeignKey('SchedulingUnitDraft', on_delete=SET_NULL, null=True, help_text='Reference for the generator to an existing collection of specifications (NULLable).')
    project = ForeignKey('Project', related_name="scheduling_sets", on_delete=PROTECT, help_text='Project to which this scheduling set belongs.')  # protected to avoid accidents

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.generator_doc and self.generator_template_id and self.generator_template.schema:
            validate_json_against_schema(self.generator_doc, self.generator_template.schema)

        super().save(force_insert, force_update, using, update_fields)


class SchedulingUnitDraft(NamedCommon):
    requirements_doc = JSONField(help_text='Scheduling and/or quality requirements for this run.')
    copies = ForeignKey('SchedulingUnitDraft', related_name="copied_from", on_delete=SET_NULL, null=True, help_text='Source reference, if we are a copy (NULLable).')
    copy_reason = ForeignKey('CopyReason', null=True, on_delete=PROTECT, help_text='Reason why source was copied (NULLable).')
    generator_instance_doc = JSONField(null=True, help_text='Parameter value that generated this run draft (NULLable).')
    scheduling_set = ForeignKey('SchedulingSet', related_name='scheduling_unit_drafts', on_delete=CASCADE, help_text='Set to which this scheduling unit draft belongs.')
    requirements_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc.') # todo: 'schema'?

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.requirements_doc and self.requirements_template_id and self.requirements_template.schema:
            validate_json_against_schema(self.requirements_doc, self.requirements_template.schema)

        super().save(force_insert, force_update, using, update_fields)

    @property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of all tasks of this scheduling unit
        '''
        return self.relative_stop_time - self.relative_start_time

    @property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all tasks of this scheduling unit
        '''
        task_drafts = list(self.task_drafts.all())
        if task_drafts:
            return min(task_drafts, key=lambda x: x.relative_start_time).relative_start_time
        else:
            return datetime.timedelta(seconds=0)

    @property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all tasks of this scheduling unit
        '''
        task_drafts = list(self.task_drafts.all())
        if task_drafts:
            return max(task_drafts, key=lambda x: x.relative_stop_time).relative_stop_time
        else:
            return datetime.timedelta(seconds=0)


class SchedulingUnitBlueprint(NamedCommon):
    requirements_doc = JSONField(help_text='Scheduling and/or quality requirements for this scheduling unit (IMMUTABLE).')
    do_cancel = BooleanField()
    requirements_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc (IMMUTABLE).')
    draft = ForeignKey('SchedulingUnitDraft', related_name='scheduling_unit_blueprints', on_delete=CASCADE, help_text='Scheduling Unit Draft which this run instantiates.')

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.requirements_doc and self.requirements_template_id and self.requirements_template.schema:
            validate_json_against_schema(self.requirements_doc, self.requirements_template.schema)

        super().save(force_insert, force_update, using, update_fields)

    @property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of all tasks of this scheduling unit
        '''
        if self.start_time is None or self.stop_time is None:
            return self.relative_stop_time - self.relative_start_time
        else:
            return self.stop_time - self.start_time   # <- todo: do we ever want this?

    @property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all tasks of this scheduling unit
        '''
        task_blueprints = list(self.task_blueprints.all())
        if task_blueprints:
            return min(task_blueprints, key=lambda x: x.relative_start_time).relative_start_time
        else:
            return datetime.timedelta(seconds=0)

    @property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all tasks of this scheduling unit
        '''
        task_blueprints = list(self.task_blueprints.all())
        if task_blueprints:
            return max(task_blueprints, key=lambda x: x.relative_stop_time).relative_stop_time
        else:
            return datetime.timedelta(seconds=0)

    @property
    def start_time(self) -> datetime or None:
        '''return the earliest start time of all tasks of this scheduling unit
        '''
        tasks_with_start_time = list(filter(lambda x: x.start_time is not None, self.task_blueprints.all()))
        if tasks_with_start_time:
            return min(tasks_with_start_time, key=lambda x: x.start_time).start_time
        else:
            return None

    @property
    def stop_time(self) -> datetime or None:
        '''return the latest stop time of all tasks of this scheduling unit
        '''
        tasks_with_stop_time = list(filter(lambda x: x.stop_time is not None, self.task_blueprints.all()))
        if tasks_with_stop_time:
            return max(tasks_with_stop_time, key=lambda x: x.stop_time).stop_time
        else:
            return None


class TaskDraft(NamedCommon):
    specifications_doc = JSONField(help_text='Specifications for this task.')
    copies = ForeignKey('TaskDraft', related_name="copied_from", on_delete=SET_NULL, null=True, help_text='Source reference, if we are a copy (NULLable).')
    copy_reason = ForeignKey('CopyReason', on_delete=PROTECT, null=True, help_text='Reason why source was copied (NULLable).')
    scheduling_unit_draft = ForeignKey('SchedulingUnitDraft', related_name='task_drafts', on_delete=CASCADE, help_text='Scheduling Unit draft to which this task draft belongs.')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc.') # todo: 'schema'?

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.specifications_doc and self.specifications_template_id and self.specifications_template.schema:
            validate_json_against_schema(self.specifications_doc, self.specifications_template.schema)

        super().save(force_insert, force_update, using, update_fields)

    @property
    def successors(self) -> QuerySet:
        '''return the connect successor taskdraft(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
           If you want the result, add .all() like so: my_taskdraft.successors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskDraft.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskdraft as successor_task\n"
                                                      "INNER JOIN tmssapp_taskrelationdraft as task_rel on task_rel.consumer_id = successor_task.id\n"
                                                      "WHERE task_rel.producer_id = %s", params=[self.id]))

    @property
    def predecessors(self) -> QuerySet:
        '''return the connect predecessor taskdraft(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
        If you want the result, add .all() like so: my_taskdraft.predecessors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskDraft.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskdraft as successor_task\n"
                                                      "INNER JOIN tmssapp_taskrelationdraft as task_rel on task_rel.producer_id = successor_task.id\n"
                                                      "WHERE task_rel.consumer_id = %s", params=[self.id]))

    @property
    def duration(self) -> datetime.timedelta:
         '''returns the overall duration of this task
         '''
         return self.relative_stop_time - self.relative_start_time

    @property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all subtasks of this task
        '''
        scheduling_relations = list(self.first_to_connect.all()) + list(self.second_to_connect.all())
        for scheduling_relation in scheduling_relations:
            if scheduling_relation.first.id == self.id and scheduling_relation.placement.value == "after":
                previous_related_task_draft = TaskDraft.objects.get(id=scheduling_relation.second.id)
                time_offset = scheduling_relation.time_offset
                # todo: max of several relations
                if previous_related_task_draft.relative_stop_time:
                    return previous_related_task_draft.relative_stop_time + datetime.timedelta(seconds=time_offset)

            if scheduling_relation.second.id == self.id and scheduling_relation.placement.value == "before":
                previous_related_task_draft = TaskDraft.objects.get(id=scheduling_relation.first.id)
                time_offset = scheduling_relation.time_offset
                # todo: max of several relations
                if previous_related_task_draft.relative_stop_time:
                    return previous_related_task_draft.relative_stop_time + datetime.timedelta(seconds=time_offset)
        return datetime.timedelta(seconds=0)

    @property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all subtasks of this task
        '''
        # todo: when it was added, check if self.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
        try:
            duration = self.specifications_doc["duration"]
            return self.relative_start_time + datetime.timedelta(seconds=duration)
        except:
            pass
        return self.relative_start_time


    # JK, 28/07/20: After discussion with Sander, we probably only want the
    #  - duration on the scheduling_unit draft (based on relative start/stop times)
    #  - duration plus relative start/stop on the task draft.
    #  This provides an estimate of what is currently planned out in the draft, but does not confuse with timestamps of actual start/stop of the blueprints.
    #  Only on the blueprints, we also aggregate start_stop times as they are in the system
    #  I'll leave these code bits here for now, until we made up our minds about this, but this can probably be removed
    #
    # @property
    # def duration(self) -> datetime.timedelta:
    #     '''returns the overall duration in seconds of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     if self.start_time is None or self.stop_time is None:
    #         # todo: calculate?
    #         return None
    #     else:
    #         return self.stop_time - self.start_time
    #
    # @property
    # def start_time(self) -> datetime or None:
    #     '''return the earliest start time of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     blueprints_with_start_time = list(filter(lambda x: x.start_time is not None, self.task_blueprints.all()))
    #     if blueprints_with_start_time:
    #         return min(blueprints_with_start_time, key=lambda x: x.start_time).start_time
    #     else:
    #         # todo: calculate?
    #         return None
    #
    # @property
    # def stop_time(self) -> datetime or None:
    #     '''return the latest stop time of all blueprints of this task
    #     # todo: is this the wanted behavior? Do you want to consider all the blueprints created from your draft or do you want to preview a new blueprint?
    #     '''
    #     blueprints_with_stop_time = list(filter(lambda x: x.stop_time is not None, self.task_blueprints.all()))
    #     if blueprints_with_stop_time:
    #         return max(blueprints_with_stop_time, key=lambda x: x.stop_time).stop_time
    #     else:
    #         # todo: calculate?
    #         return None


class TaskBlueprint(NamedCommon):
    specifications_doc = JSONField(help_text='Schedulings for this task (IMMUTABLE).')
    do_cancel = BooleanField(help_text='Cancel this task.')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc (IMMUTABLE).')
    draft = ForeignKey('TaskDraft', related_name='task_blueprints', on_delete=CASCADE, help_text='Task Draft which this task instantiates.')
    scheduling_unit_blueprint = ForeignKey('SchedulingUnitBlueprint', related_name='task_blueprints', on_delete=CASCADE, help_text='Scheduling Unit Blueprint to which this task belongs.')

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.specifications_doc and self.specifications_template_id and self.specifications_template.schema:
            validate_json_against_schema(self.specifications_doc, self.specifications_template.schema)

        super().save(force_insert, force_update, using, update_fields)

    @property
    def successors(self) -> QuerySet:
        '''return the connect successor taskblueprint(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
           If you want the result, add .all() like so: my_taskblueprint.successors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskBlueprint.objects.filter(id__in=RawSQL("SELECT successor_task.id FROM tmssapp_taskblueprint as successor_task\n"
                                                          "INNER JOIN tmssapp_taskrelationblueprint as task_rel on task_rel.consumer_id = successor_task.id\n"
                                                          "WHERE task_rel.producer_id = %s", params=[self.id]))

    @property
    def predecessors(self) -> QuerySet:
        '''return the connect predecessor taskblueprint(s) as queryset (over which you can perform extended queries, or return via the serializers/viewsets)
        If you want the result, add .all() like so: my_taskblueprint.predecessors.all()
        '''
        # see comment on Subtask for why this needs to be raw sql
        return TaskBlueprint.objects.filter(id__in=RawSQL("SELECT predecessor_task.id FROM tmssapp_taskblueprint as predecessor_task\n"
                                                          "INNER JOIN tmssapp_taskrelationblueprint as task_rel on task_rel.producer_id = predecessor_task.id\n"
                                                          "WHERE task_rel.consumer_id = %s", params=[self.id]))

    @property
    def duration(self) -> datetime.timedelta:
        '''return the overall duration of this task
        '''
        if self.start_time is None or self.stop_time is None:
            return self.relative_stop_time - self.relative_start_time
        else:
            return self.stop_time - self.start_time

    @property
    def relative_start_time(self) -> datetime.timedelta:
        '''return the earliest relative start time of all subtasks of this task
        '''
        scheduling_relations = list(self.first_to_connect.all()) + list(self.second_to_connect.all())
        for scheduling_relation in scheduling_relations:
                if scheduling_relation.first.id == self.id and scheduling_relation.placement.value == "after":
                    previous_related_task_blueprint = TaskBlueprint.objects.get(id=scheduling_relation.second.id)
                    time_offset = scheduling_relation.time_offset
                    # todo: max of several relations
                    if previous_related_task_blueprint.relative_stop_time:
                        return previous_related_task_blueprint.relative_stop_time + datetime.timedelta(seconds=time_offset)

                if scheduling_relation.second.id == self.id and scheduling_relation.placement.value == "before":
                    previous_related_task_blueprint = TaskBlueprint.objects.get(id=scheduling_relation.first.id)
                    time_offset = scheduling_relation.time_offset
                    # todo: max of several relations
                    if previous_related_task_blueprint.relative_stop_time:
                        return previous_related_task_blueprint.relative_stop_time + datetime.timedelta(seconds=time_offset)
        return datetime.timedelta(seconds=666660)

    @property
    def relative_stop_time(self) -> datetime.timedelta:
        '''return the latest relative stop time of all subtasks of this task
        '''
        # todo: when it was added, check if subtask.specifications_template.type.value == TaskType.Choices.OBSERVATION.value:
        try:
            duration = self.specifications_doc["duration"]
            return self.relative_start_time + datetime.timedelta(seconds=duration)
        except:
            pass
        return self.relative_start_time

    @property
    def start_time(self) -> datetime or None:
        '''return the earliest start time of all subtasks of this task
        '''
        subtasks_with_start_time = list(filter(lambda x: x.start_time is not None, self.subtasks.all()))
        if subtasks_with_start_time:
            return min(subtasks_with_start_time, key=lambda x: x.start_time).start_time
        else:
            return None

    @property
    def stop_time(self) -> datetime or None:
        '''return the latest stop time of all subtasks of this task
        '''
        subtasks_with_stop_time = list(filter(lambda x: x.stop_time is not None, self.subtasks.all()))
        if subtasks_with_stop_time:
            return max(subtasks_with_stop_time, key=lambda x: x.stop_time).stop_time
        else:
            return None


class TaskRelationDraft(BasicCommon):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')
    selection_template = ForeignKey('TaskRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.') # todo: 'schema'?
    dataformat = ForeignKey('Dataformat', null=False, on_delete=PROTECT, help_text='Selected data format to use. One of (MS, HDF5).')

    # caveat: it might look like producer has an incorrect related_name='consumed_by'. But it really is correct, denends on the way you look at it
    producer = ForeignKey('TaskDraft', related_name='consumed_by', on_delete=CASCADE, help_text='Task Draft that has the output connector. NOTE: The producer does typically, but not necessarily, belong to the same Scheduling Unit (or even the same Project) as the consumer.')
    # caveat: it might look like consumer has an incorrect related_name='produced_by'. But it really is correct, denends on the way you look at it
    consumer = ForeignKey('TaskDraft', related_name='produced_by', on_delete=CASCADE, help_text='Task Draft that has the input connector.')

    input_role = ForeignKey('TaskConnectorType', related_name='taskrelationdraft_input_roles', on_delete=CASCADE, help_text='Input connector type (what kind of data can be taken as input).')
    output_role = ForeignKey('TaskConnectorType', related_name='taskrelationdraft_output_roles', on_delete=CASCADE, help_text='Output connector type (what kind of data can be created as output).')

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.selection_doc and self.selection_template_id and self.selection_template.schema:
            validate_json_against_schema(self.selection_doc, self.selection_template.schema)

        super().save(force_insert, force_update, using, update_fields)


class TaskRelationBlueprint(BasicCommon):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')
    dataformat = ForeignKey('Dataformat', null=False, on_delete=PROTECT, help_text='Selected data format to use.')

    # caveat: it might look like producer has an incorrect related_name='consumed_by'. But it really is correct, denends on the way you look at it
    producer = ForeignKey('TaskBlueprint', related_name='consumed_by', on_delete=CASCADE, help_text='Task Blueprint that has the output connector.')
    # caveat: it might look like consumer has an incorrect related_name='produced_by'. But it really is correct, denends on the way you look at it
    consumer = ForeignKey('TaskBlueprint', related_name='produced_by', on_delete=CASCADE, help_text='Task Blueprint that has the input connector.')

    input_role = ForeignKey('TaskConnectorType', related_name='taskrelationblueprint_input_roles', on_delete=CASCADE, help_text='Input connector type (what kind of data can be taken as input).')
    output_role = ForeignKey('TaskConnectorType', related_name='taskrelationblueprint_output_roles', on_delete=CASCADE, help_text='Output connector type (what kind of data can be created as output).')
    draft = ForeignKey('TaskRelationDraft', on_delete=CASCADE, related_name='related_task_relation_blueprint', help_text='Task Relation Draft which this work request instantiates.')
    selection_template = ForeignKey('TaskRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.')  # todo: 'schema'?

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.selection_doc and self.selection_template_id and self.selection_template.schema:
            validate_json_against_schema(self.selection_doc, self.selection_template.schema)

        super().save(force_insert, force_update, using, update_fields)


class TaskSchedulingRelationBlueprint(BasicCommon):
    first = ForeignKey('TaskBlueprint', related_name='first_to_connect', on_delete=CASCADE, help_text='First Task Blueprint to connect.')
    second = ForeignKey('TaskBlueprint', related_name='second_to_connect', on_delete=CASCADE, help_text='Second Task Blueprint to connect.')
    placement = ForeignKey('SchedulingRelationPlacement', null=False, default="after", on_delete=PROTECT, help_text='Task scheduling relation placement.')
    time_offset = IntegerField(default=60, help_text='Time offset of start of second task with respect to start of first task.')  
  
    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.first == self.second:
            raise ValidationError("First and Second Task Draft must be different.")
        if (str(self.placement) == SchedulingRelationPlacement.Choices.BEFORE.value or str(self.placement) == SchedulingRelationPlacement.Choices.AFTER.value) and self.time_offset<0:
            raise ValidationError("Time_offset must be >= 0")
        super().save(force_insert, force_update, using, update_fields)
    

class TaskSchedulingRelationDraft(BasicCommon):
    first = ForeignKey('TaskDraft', related_name='first_to_connect', on_delete=CASCADE, help_text='First Task Draft to connect.')
    second = ForeignKey('TaskDraft', related_name='second_to_connect', on_delete=CASCADE, help_text='Second Task Draft to connect.')
    placement = ForeignKey('SchedulingRelationPlacement', null=False, on_delete=PROTECT, help_text='Task scheduling relation placement.')
    time_offset = IntegerField(default=60, help_text='Time offset of start of second task with respect to start of first task.')  

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        if self.first == self.second:
            raise ValidationError("First and Second Task Draft must be different.")
        if (str(self.placement) == SchedulingRelationPlacement.Choices.BEFORE.value or str(self.placement) == SchedulingRelationPlacement.Choices.AFTER.value) and self.time_offset<0:
            raise ValidationError("Time_offset must be >= 0")
        super().save(force_insert, force_update, using, update_fields)

