"""
This module 'populate' defines methods to populate the database with predefined ('static') data,
according to the proposed Django way: https://docs.djangoproject.com/en/2.0/topics/migrations/#data-migrations

import this module in your empty migration step file, and add the following migration:

from ..populate import *

class Migration(migrations.Migration):

    dependencies = [ <the dependency is automatically inserted here> ]

    operations = [ migrations.RunPython(populate_choices) ]

"""

import logging
logger = logging.getLogger(__name__)

import json
import os
from datetime import datetime, timezone
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.models.specification import *
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import *
from lofar.common.json_utils import *
from lofar.common import isTestEnvironment, isDevelopmentEnvironment

working_dir = os.path.dirname(os.path.abspath(__file__))


def populate_choices(apps, schema_editor):
    '''
    populate each 'choice' table in the database with the 'static' list of 'choice'.Choices for
    each 'choice'type in Role, Datatype, Dataformat, CopyReason
    :return: None
    '''
    for choice_class in [Role, Datatype, Dataformat, CopyReason,
                         SubtaskState, SubtaskType, StationType, Algorithm, ScheduleMethod, SchedulingRelationPlacement,
                         Flag, ProjectCategory, PeriodCategory]:
        choice_class.objects.bulk_create([choice_class(value=x.value) for x in choice_class.Choices])

def populate_settings(apps, schema_editor):
    Setting.objects.create(name=Flag.objects.get(value='allow_scheduling_observations'), value=True)


def populate_lofar_json_schemas(apps, schema_editor):

    _populate_scheduling_unit_schema()
    # populate task schema's
    _populate_preprocessing_schema()
    _populate_observation_with_stations_schema()
    _populate_calibrator_addon_schema()

    _populate_dataproduct_specifications_templates()
    _populate_taskrelation_selection_templates()
    _populate_dataproduct_feedback_templates()
    _populate_obscontrol_schema()
    _populate_pipelinecontrol_schema()
    _populate_connectors()

    _populate_qa_files_subtask_template()
    _populate_qa_plots_subtask_template()


def populate_test_scheduling_set(apps, schema_editor):
    """
    Create a Test Schedule Set to be able to refer to when Scheduling Unit Draft is created from a
    scheduling unit json
    :return:
    """
    try:
        from lofar.sas.tmss.test.tmss_test_data_django_models import SchedulingSet_test_data, SchedulingUnitDraft_test_data
        if isTestEnvironment() or isDevelopmentEnvironment():
            # create a Test Scheduling Set UC1 under project TMSS-Commissioning
            tmss_project = models.Project.objects.get(name="TMSS-Commissioning")
            scheduling_set_data = SchedulingSet_test_data(name="Test Scheduling Set UC1", project=tmss_project)
            scheduling_set = models.SchedulingSet.objects.create(**scheduling_set_data)

            # construct a scheduling_unit_doc, i.e.: a specification of interrelated tasks which conforms the scheduling unit schema
            # by default, this scheduling_unit_doc holds no tasks, so lets setup the UC1 sequence of tasks here, and add it to the scheduling_unit_doc
            scheduling_unit_template = models.SchedulingUnitTemplate.objects.get(name="scheduling unit schema")
            scheduling_unit_doc = get_default_json_object_for_schema(scheduling_unit_template.schema)

            # create and add a calibrator task spec
            scheduling_unit_doc['tasks'].append({"name": "Calibrator Observation 1",
                                                 "description": "Calibrator Observation for UC1 HBA scheduling unit",
                                                 "specifications_doc": get_default_json_object_for_schema(models.TaskTemplate.objects.get(name="calibrator schema").schema),
                                                 "specifications_template": "calibrator schema"})

            # create and add a calibrator preprocessing spec
            scheduling_unit_doc['tasks'].append({"name": "Pipeline Calibrator1",
                                                 "description": "Preprocessing Pipeline for Calibrator Observation 1",
                                                 "specifications_doc": get_default_json_object_for_schema(models.TaskTemplate.objects.get(name="preprocessing schema").schema),
                                                 "specifications_template": "preprocessing schema"})

            # create and add a target obs spec
            scheduling_unit_doc['tasks'].append({"name": "Target Observation",
                                                 "description": "Target Observation for UC1 HBA scheduling unit",
                                                 "specifications_doc": get_default_json_object_for_schema(models.TaskTemplate.objects.get(name="observation schema").schema),
                                                 "specifications_template": "observation schema"})

            # create and add a target pipeline spec for sap0
            scheduling_unit_doc['tasks'].append({"name": "Preprocessing Pipeline SAP0",
                                                 "description": "Preprocessing Pipeline for Target Observation SAP0",
                                                 "specifications_doc": get_default_json_object_for_schema(models.TaskTemplate.objects.get(name="preprocessing schema").schema),
                                                 "specifications_template": "preprocessing schema"})

            # create and add a target pipeline spec for sap1
            scheduling_unit_doc['tasks'].append({"name": "Preprocessing Pipeline SAP1",
                                                 "description": "Preprocessing Pipeline for Target Observation SAP1",
                                                 "specifications_doc": get_default_json_object_for_schema(models.TaskTemplate.objects.get(name="preprocessing schema").schema),
                                                 "specifications_template": "preprocessing schema"})

            # create and add a calibrator task spec
            scheduling_unit_doc['tasks'].append({"name": "Calibrator Observation 2",
                                                 "description": "Calibrator Observation for UC1 HBA scheduling unit",
                                                 "specifications_doc": get_default_json_object_for_schema(models.TaskTemplate.objects.get(name="calibrator schema").schema),
                                                 "specifications_template": "calibrator schema"})

            # create and add a calibrator preprocessing spec
            scheduling_unit_doc['tasks'].append({"name": "Pipeline Calibrator2",
                                                 "description": "Preprocessing Pipeline for Calibrator Observation 2",
                                                 "specifications_doc": get_default_json_object_for_schema(models.TaskTemplate.objects.get(name="preprocessing schema").schema),
                                                 "specifications_template": "preprocessing schema"})

            # ----- end of tasks

            # setup task_scheduling_relations between Target and Calibrator observations
            scheduling_unit_doc['task_scheduling_relations'].append({"first": "Calibrator Observation 1",
                                                                     "second": "Target Observation",
                                                                     "placement": "before",
                                                                     "time_offset": 60 })
            scheduling_unit_doc['task_scheduling_relations'].append({"first": "Calibrator Observation 2",
                                                                     "second": "Target Observation",
                                                                     "placement": "after",
                                                                     "time_offset": 60 })

            # ----- end of task_scheduling_relations

            #TODO: check various input/output datatypes and roles for each task_relation
            scheduling_unit_doc['task_relations'].append({"producer": "Calibrator Observation 1",
                                                          "consumer": "Pipeline Calibrator1",
                                                          "tags": [],
                                                          "input": { "role": "input", "datatype": "visibilities" },
                                                          "output": { "role": "correlator", "datatype": "visibilities" },
                                                          "dataformat": "MeasurementSet",
                                                          "selection_doc": {},
                                                          "selection_template": "All" })

            scheduling_unit_doc['task_relations'].append({"producer": "Calibrator Observation 2",
                                                          "consumer": "Pipeline Calibrator2",
                                                          "tags": [],
                                                          "input": { "role": "input", "datatype": "visibilities" },
                                                          "output": { "role": "correlator", "datatype": "visibilities" },
                                                          "dataformat": "MeasurementSet",
                                                          "selection_doc": {},
                                                          "selection_template": "All" })

            scheduling_unit_doc['task_relations'].append({"producer": "Target Observation",
                                                          "consumer": "Preprocessing Pipeline SAP0",
                                                          "tags": [],
                                                          "input": { "role": "input", "datatype": "visibilities" },
                                                          "output": { "role": "correlator", "datatype": "visibilities" },
                                                          "dataformat": "MeasurementSet",
                                                          "selection_doc": {"sap": [0]},
                                                          "selection_template": "SAP" })

            scheduling_unit_doc['task_relations'].append({"producer": "Target Observation",
                                                          "consumer": "Preprocessing Pipeline SAP1",
                                                          "tags": [],
                                                          "input": { "role": "input", "datatype": "visibilities" },
                                                          "output": { "role": "correlator", "datatype": "visibilities" },
                                                          "dataformat": "MeasurementSet",
                                                          "selection_doc": {"sap": [1]},
                                                          "selection_template": "SAP" })

            # finally... add the scheduling_unit_doc to a new SchedulingUnitDraft instance, and were ready to use it!
            scheduling_unit_data = SchedulingUnitDraft_test_data(name="Test Scheduling Unit UC1", scheduling_set=scheduling_set,
                                                                 template=scheduling_unit_template, requirements_doc=scheduling_unit_doc)
            models.SchedulingUnitDraft.objects.create(**scheduling_unit_data)

    except ImportError:
        pass


def populate_cycles(apps, schema_editor):
    for nr in range(0, 18):
        cycle = models.Cycle.objects.create(name="Cycle %02d" % nr,
                                            description="Lofar Cycle %s" % nr,
                                            start=datetime(2013+nr//2, 6 if nr%2==0 else 11, 1, 0, 0, 0, 0, tzinfo=timezone.utc),
                                            stop=datetime(2013+(nr+1)//2, 6 if nr%2==1 else 11, 1, 0, 0, 0, 0, tzinfo=timezone.utc))
        models.CycleQuota.objects.create(cycle=cycle,
                                         resource_type=ResourceType.objects.get(name="lofar_observing_time"),
                                         value=0.8*cycle.duration.total_seconds()) # rough guess. 80% of total time available for observing
        models.CycleQuota.objects.create(cycle=cycle,
                                         resource_type=ResourceType.objects.get(name="cep_processing_time"),
                                         value=0.8*cycle.duration.total_seconds())
        models.CycleQuota.objects.create(cycle=cycle,
                                         resource_type=ResourceType.objects.get(name="lta_storage"),
                                         value=0) # needs to be filled in by user (SOS)
        models.CycleQuota.objects.create(cycle=cycle,
                                         resource_type=ResourceType.objects.get(name="support_time"),
                                         value=0)  # needs to be filled in by user (SOS)
        models.CycleQuota.objects.create(cycle=cycle,
                                         resource_type=ResourceType.objects.get(name="lofar_observing_time_commissioning"),
                                         value=0.05*cycle.duration.total_seconds()) # rough guess. 5% of total time available for observing
        models.CycleQuota.objects.create(cycle=cycle,
                                         resource_type=ResourceType.objects.get(name="lofar_observing_time_prio_a"),
                                         value=0) # needs to be filled in by user (SOS)
        models.CycleQuota.objects.create(cycle=cycle,
                                         resource_type=ResourceType.objects.get(name="lofar_observing_time_prio_b"),
                                         value=0) # needs to be filled in by user (SOS)


def populate_projects(apps, schema_editor):
    tmss_project = models.Project.objects.create(name="TMSS-Commissioning",
                                             description="Project for all TMSS tests and commissioning",
                                             priority_rank=1.0,
                                             can_trigger=False,
                                             private_data=True,
                                             expert=True,
                                             filler=False)
    tmss_project.cycles.set([models.Cycle.objects.get(name="Cycle 14")])


def populate_resources(apps, schema_editor):
    ResourceType.objects.create(name="lta_storage", description="Amount of storage in the LTA (in bytes)")
    ResourceType.objects.create(name="cep_storage", description="Amount of storage on the CEP processing cluster (in bytes)")
    ResourceType.objects.create(name="cep_processing_time", description="Processing time on the CEP processing cluster (in seconds)")
    ResourceType.objects.create(name="lofar_observing_time", description="Observing time (in seconds)")
    ResourceType.objects.create(name="lofar_observing_time_prio_a", description="Observing time with priority A (in seconds)")
    ResourceType.objects.create(name="lofar_observing_time_prio_b", description="Observing time with priority B (in seconds)")
    ResourceType.objects.create(name="lofar_observing_time_commissioning", description="Observing time for Commissioning/DDT (in seconds)")
    ResourceType.objects.create(name="support_time", description="Support time by human (in seconds)")
    ResourceType.objects.create(name="number_of_triggers", description="Number of trigger events (as integer)")


def populate_misc(apps, schema_editor):
    cluster = Cluster.objects.create(name="CEP4", location="CIT")
    fs = Filesystem.objects.create(name="LustreFS", cluster=cluster, capacity=3.6e15)


def _populate_scheduling_unit_schema():
    with open(os.path.join(working_dir, "schemas/scheduling-unit.json")) as json_file:
        json_data = json.loads(json_file.read())
        scheduling_unit_template_data = {"name": "scheduling unit schema",
                                         "description": 'Schema for scheduling unit',
                                         "version": '0.1',
                                         "tags": ["UC1"],
                                         "schema": json_data}
    SchedulingUnitTemplate.objects.create(**scheduling_unit_template_data)


def _populate_observation_with_stations_schema():
    with open(os.path.join(working_dir, "schemas/task-observation-with-stations.json")) as json_file:
        json_data = json.loads(json_file.read())
        task_template_data = {"name": "observation schema",
                              "description": 'schema for observations',
                              "version": '0.1',
                              "tags": [],
                              "schema": json_data}
    TaskTemplate.objects.create(**task_template_data)


def _populate_calibrator_addon_schema():
    with open(os.path.join(working_dir, "schemas/task-calibrator-addon.json")) as json_file:
        json_data = json.loads(json_file.read())
        task_template_data = {"name": "calibrator schema",
                              "description": 'addon schema for calibrator observations',
                              "version": '0.1',
                              "tags": [],
                              "schema": json_data}
    TaskTemplate.objects.create(**task_template_data)


def _populate_correlator_calibrator_schema():
    with open(os.path.join(working_dir, "schemas/task-correlator.json")) as json_file:
        json_data = json.loads(json_file.read())
        task_template_data = {"name": "correlator schema",
                              "description": 'correlator schema for calibrator observations',
                              "version": '0.1',
                              "tags": ["obsolete?"],
                              "schema": json_data}
    TaskTemplate.objects.create(**task_template_data)


def _populate_obscontrol_schema():
    with open(os.path.join(working_dir, "schemas/subtask-observation-control.json")) as json_file:
        json_data = json.loads(json_file.read())
        subtask_template_data = {"type": SubtaskType.objects.get(value='observation'),
                                 "name": "observationcontrol schema",
                                 "description": 'observationcontrol schema for observation subtask',
                                 "version": '0.1',
                                 "realtime": True,
                                 "queue": False,
                                 "tags": [],
                                 "schema": json_data}
    SubtaskTemplate.objects.create(**subtask_template_data)


def _populate_stations_schema():
    with open(os.path.join(working_dir, "schemas/task-stations.json")) as json_file:
        json_data = json.loads(json_file.read())
        task_template_data = {"name": "stations schema",
                              "description": 'Generic station settings and selection',
                              "version": '0.1',
                              "tags": [],
                              "schema": json_data}
    TaskTemplate.objects.create(**task_template_data)


def _populate_taskrelation_selection_templates():
    # All
    with open(os.path.join(working_dir, "schemas/empty-schema.json")) as json_file:
        json_data = json.loads(json_file.read())
        data = {"name": "All",
                "description": 'Select all, apply no filtering.',
                "version": '1',
                "schema": json_data}
    TaskRelationSelectionTemplate.objects.create(**data)

    # SAP
    with open(os.path.join(working_dir, "schemas/task-relation-sap.json")) as json_file:
        json_data = json.loads(json_file.read())
        data = {"name": "SAP",
                "description": 'Select by SAP.',
                "version": '1',
                "tags": [],
                "schema": json_data}
    TaskRelationSelectionTemplate.objects.create(**data)


def _populate_dataproduct_specifications_templates():
    # Note: to some extend, this reflects the TaskRelationSelectionTemplates. That is expected since they define
    # the filters that operate on the DataproductSpecificationTemplates defined here. However, filters probably
    # will only use a subset of dataproduct specs, but could allow selecting several values of which only one
    # can be met by a single dataproduct.
    with open(os.path.join(working_dir, "schemas/empty-schema.json")) as json_file:
        json_data = json.loads(json_file.read())
        template_data = {"name": "Empty",
                         "description": 'Empty DataproductSpecificationsTemplate with an empty schema',
                         "version": '1',
                         "tags": [],
                         "schema": json_data}
    DataproductSpecificationsTemplate.objects.create(**template_data)

    # SAP
    with open(os.path.join(working_dir, "schemas/task-relation-sap.json")) as json_file:
        json_data = json.loads(json_file.read())
        template_data = {"name": "SAP",
                         "description": 'Select by SAP.',
                         "version": '1',
                         "schema": json_data}
    DataproductSpecificationsTemplate.objects.create(**template_data)


def _populate_dataproduct_feedback_templates():
    with open(os.path.join(working_dir, "schemas/empty-schema.json")) as json_file:
        json_data = json.loads(json_file.read())
        template_data = {"name": "Empty",
                         "description": 'Empty DataproductFeedbackTemplate with an empty schema',
                         "version": '1',
                         "tags": [],
                         "schema": json_data}
    DataproductFeedbackTemplate.objects.create(**template_data)


def _populate_qa_files_subtask_template():
    with open(os.path.join(working_dir, "schemas/subtask-qa-files.json")) as json_file:
        json_data = json.loads(json_file.read())
        subtask_template_data = {"type": SubtaskType.objects.get(value=SubtaskType.Choices.QA_FILES.value),
                                 "name": "QA file conversion",
                                 "description": 'QA file conversion subtask template',
                                 "version": '0.1',
                                 "realtime": False,
                                 "queue": True,
                                 "tags": [],
                                 "schema": json_data}
    SubtaskTemplate.objects.create(**subtask_template_data)


def _populate_qa_plots_subtask_template():
    with open(os.path.join(working_dir, "schemas/subtask-qa-plots.json")) as json_file:
        json_data = json.loads(json_file.read())
        subtask_template_data = {"type": SubtaskType.objects.get(value=SubtaskType.Choices.QA_PLOTS.value),
                                 "name": "QA plots",
                                 "description": 'QA plots subtask template',
                                 "version": '0.1',
                                 "realtime": False,
                                 "queue": True,
                                 "tags": [],
                                 "schema": json_data}
    SubtaskTemplate.objects.create(**subtask_template_data)


def _populate_preprocessing_schema():
    with open(os.path.join(working_dir, "schemas/task-preprocessing.json")) as json_file:
        json_data = json.loads(json_file.read())
        task_template_data = {"name": "preprocessing schema",
                              "description": 'preprocessing settings',
                              "version": '0.1',
                              "tags": [],
                              "schema": json_data}
    TaskTemplate.objects.create(**task_template_data)


def _populate_pipelinecontrol_schema():
    with open(os.path.join(working_dir, "schemas/subtask-pipeline-control.json")) as json_file:
        json_data = json.loads(json_file.read())
        subtask_template_data = {"type": SubtaskType.objects.get(value='pipeline'),
                                 "name": "pipelinecontrol schema",
                                 "description": 'pipelinecontrol schema for pipeline subtask',
                                 "version": '0.1',
                                 "realtime": True,
                                 "queue": False,
                                 "tags": [],
                                 "schema": json_data}
    SubtaskTemplate.objects.create(**subtask_template_data)


def _populate_connectors():
    # the TaskConnectorType's define how the Task[Draft/Blueprint] *can* be connected.
    # TODO Need overview which we do actually need
    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.INPUT.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                 output_of=TaskTemplate.objects.get(name='observation schema'),
                                 input_of=TaskTemplate.objects.get(name='preprocessing schema'))

    TaskConnectorType.objects.create(role=Role.objects.get(value=Role.Choices.CORRELATOR.value),
                                 datatype=Datatype.objects.get(value=Datatype.Choices.VISIBILITIES.value),
                                 output_of=TaskTemplate.objects.get(name='calibrator schema'),
                                 input_of=TaskTemplate.objects.get(name='preprocessing schema'))