"""
This file contains the serializers (for the elsewhere defined data models)
"""

from rest_framework import serializers
from .. import models
from .widgets import JSONEditorField
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from rest_framework import decorators
import json


class RelationalHyperlinkedModelSerializer(serializers.HyperlinkedModelSerializer):
    _accepted_pk_names = ('id', 'name')

    def get_field_names(self, declared_fields, info):
        field_names = super().get_field_names(declared_fields, info)
        field_names.remove(self.url_field_name) # is added later, see retun statement

        if getattr(self.Meta, 'extra_fields', None):
            field_names += self.Meta.extra_fields

        # add 'plain'-values of the fieldnames which relate to AbstractChoice-'lookup-tables'
        choice_field_names = [name+'_value' for name, rel in info.forward_relations.items()
                              if issubclass(rel.related_model, models.AbstractChoice)]

        # add 'plain-id(s)'-values of the fieldnammes which to forward_relations
        forward_related_field_names = [name+'_ids' if rel.to_many else name+'_id'
                                       for name, rel in info.forward_relations.items()
                                       if rel.related_model._meta.pk.name in self._accepted_pk_names
                                       and name in field_names]

        # always add 'plain-id'-values of the fieldnammes which to reverse_relations
        reverse_related_field_names = [name+'_ids' for name, rel in info.reverse_relations.items()
                                       if rel.related_model._meta.pk.name in self._accepted_pk_names
                                       and name in field_names]

        # return them sorted alphabetically, with id and url first so it's easy to identify and 'click' them
        return [info.pk.name, self.url_field_name] + sorted(field_names + choice_field_names + forward_related_field_names + reverse_related_field_names)

    def build_field(self, field_name, info, model_class, nested_depth):
        '''override of super.build_field to handle 'choice' fields'''
        try:
            return super().build_field(field_name, info, model_class, nested_depth)
        except ImproperlyConfigured:
            if field_name.endswith('_ids'):
                return self.build_reverse_relations_ids_field(field_name, info, model_class, nested_depth)
            if field_name.endswith('_value'):
                return self.build_choice_field(field_name, info)
            raise

    def build_reverse_relations_ids_field(self, field_name, info, model_class, nested_depth):
        '''builds a PrimaryKeyRelatedField serializer for the 'reverse_relations_ids' fields'''
        return serializers.PrimaryKeyRelatedField, {'label':field_name,
                                                    'source':field_name[:-4], # cut '_ids' from end
                                                    'many':True,
                                                    'read_only':True}

    def build_choice_field(self, field_name, info):
        '''builds a StringRelatedField serializer for the 'choice' fields'''
        original_field_name = field_name[:-6] # cut '_value' from end
        if original_field_name in info.forward_relations.keys():
            return serializers.StringRelatedField, {'label':field_name,
                                                    'source': original_field_name,
                                                    'read_only':True}


class FloatDurationField(serializers.FloatField):

    # Turn datetime to float representation in seconds.
    # (Timedeltas are otherwise by default turned into a string representation)
    def to_representation(self, value):
        return value.total_seconds()


# This is required for keeping a user reference as ForeignKey in other models
# (I think so that the HyperlinkedModelSerializer can generate a URI)
class UserSerializer(serializers.Serializer):
    class Meta:
        model = User
        fields = '__all__'


class TagsSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Tags
        fields = '__all__'


class GeneratorTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.GeneratorTemplate
        fields = '__all__'


class DefaultGeneratorTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultGeneratorTemplate
        fields = '__all__'


class SchedulingUnitTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.SchedulingUnitTemplate
        fields = '__all__'


class DefaultSchedulingUnitTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultSchedulingUnitTemplate
        fields = '__all__'


class TaskTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskTemplate
        fields = '__all__'


class DefaultTaskTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultTaskTemplate
        fields = '__all__'


class TaskRelationSelectionTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskRelationSelectionTemplate
        fields = '__all__'


class DefaultTaskRelationSelectionTemplateSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.DefaultTaskRelationSelectionTemplate
        fields = '__all__'


class RoleSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Role
        fields = '__all__'

class SchedulingRelationPlacementSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.SchedulingRelationPlacement
        fields = '__all__'


class DatatypeSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Datatype
        fields = '__all__'


class DataformatSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Dataformat
        fields = '__all__'


class CopyReasonSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.CopyReason
        fields = '__all__'


class TaskConnectorTypeSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskConnectorType
        fields = '__all__'


class CycleSerializer(RelationalHyperlinkedModelSerializer):
    duration = FloatDurationField(read_only=True, help_text="Duration of the cycle [seconds]")

    class Meta:
        model = models.Cycle
        fields = '__all__'
        extra_fields = ['projects', 'name', 'duration', 'cycle_quota']

class CycleQuotaSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.CycleQuota
        fields = '__all__'
        extra_fields = ['resource_type']

class ProjectSerializer(RelationalHyperlinkedModelSerializer):
#    scheduling_sets = serializers.PrimaryKeyRelatedField(source='scheduling_sets', read_only=True, many=True)

    class Meta:
        model = models.Project
        fields = '__all__'
        extra_fields = ['name','project_quota'] #, 'scheduling_sets']


class ProjectQuotaSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ProjectQuota
        fields = '__all__'
        extra_fields = ['resource_type']


class ResourceTypeSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ResourceType
        fields = '__all__'
        extra_fields = ['name']


class FlagSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Flag
        fields = '__all__'


class SettingSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.Setting
        fields = '__all__'


class ProjectCategorySerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.ProjectCategory
        fields = '__all__'


class PeriodCategorySerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.PeriodCategory
        fields = '__all__'


class SchedulingSetSerializer(RelationalHyperlinkedModelSerializer):

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields['generator_doc'] = JSONEditorField(self.instance.generator_template.schema)
        except Exception as e:
            pass

    class Meta:
        model = models.SchedulingSet
        fields = '__all__'
        extra_fields = ['scheduling_unit_drafts']


class SchedulingUnitDraftSerializer(RelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(required=False)

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields['requirements_doc'] = JSONEditorField(self.instance.requirements_template.schema)
        except Exception as e:
            pass

    class Meta:
        model = models.SchedulingUnitDraft
        fields = '__all__'
        extra_fields = ['scheduling_unit_blueprints', 'task_drafts', 'duration']


class SchedulingUnitBlueprintSerializer(RelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(required=False)

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields['requirements_doc'] = JSONEditorField(self.instance.requirements_template.schema)
        except Exception as e:
            pass

    class Meta:
        model = models.SchedulingUnitBlueprint
        fields = '__all__'
        extra_fields = ['task_blueprints', 'duration', 'start_time', 'stop_time']


class TaskDraftSerializer(RelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(required=False)
    relative_start_time = FloatDurationField(required=False)
    relative_stop_time = FloatDurationField(required=False)

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields['specifications_doc'] = JSONEditorField(self.instance.specifications_template.schema)
        except Exception as e:
            pass

    class Meta:
        model = models.TaskDraft
        fields = '__all__'
        extra_fields = ['task_blueprints', 'produced_by', 'consumed_by', 'first_to_connect', 'second_to_connect', 'duration', 'relative_start_time', 'relative_stop_time']


class TaskBlueprintSerializer(RelationalHyperlinkedModelSerializer):

    duration = FloatDurationField(required=False)
    relative_start_time = FloatDurationField(required=False)
    relative_stop_time = FloatDurationField(required=False)

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields['specifications_doc'] = JSONEditorField(self.instance.specifications_template.schema)
        except Exception as e:
            pass

    class Meta:
        model = models.TaskBlueprint
        fields = '__all__'
        extra_fields = ['subtasks', 'produced_by', 'consumed_by', 'first_to_connect', 'second_to_connect', 'duration', 'start_time', 'stop_time', 'relative_start_time', 'relative_stop_time']


class TaskRelationDraftSerializer(RelationalHyperlinkedModelSerializer):

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields['selection_doc'] = JSONEditorField(self.instance.selection_template.schema)
        except Exception as e:
            pass

    class Meta:
        model = models.TaskRelationDraft
        fields = '__all__'
        extra_fields = ['related_task_relation_blueprint']


class TaskRelationBlueprintSerializer(RelationalHyperlinkedModelSerializer):

    # Create a JSON editor form to replace the simple text field based on the schema in the template that this
    # draft refers to. If that fails, the JSONField remains a standard text input.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.fields['selection_doc'] = JSONEditorField(self.instance.selection_template.schema)
        except Exception as e:
            pass

    class Meta:
        model = models.TaskRelationBlueprint
        fields = '__all__'


class TaskSchedulingRelationDraftSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskSchedulingRelationDraft
        fields = '__all__'


class TaskSchedulingRelationBlueprintSerializer(RelationalHyperlinkedModelSerializer):
    class Meta:
        model = models.TaskSchedulingRelationBlueprint
        fields = '__all__'




