from lofar.sas.tmss.tmss.exceptions import *
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.models.specification import TaskBlueprint, SchedulingUnitBlueprint, TaskDraft, SchedulingRelationPlacement
from lofar.sas.tmss.tmss.tmssapp.subtasks import create_and_schedule_subtasks_from_task_blueprint
from lofar.sas.tmss.tmss.tmssapp.models.specification import TaskBlueprint, SchedulingUnitBlueprint
from lofar.sas.tmss.tmss.tmssapp.subtasks import create_and_schedule_subtasks_from_task_blueprint, \
    create_subtasks_from_task_blueprint, schedule_independent_subtasks_in_task_blueprint
from functools import cmp_to_key

import logging
logger = logging.getLogger(__name__)


def create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    """
    Create a SchedulingUnitBlueprint from the SchedulingUnitDraft
    :raises Exception if instantiate fails.
    """
    logger.debug("create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft.id=%s name='%s')", scheduling_unit_draft.pk, scheduling_unit_draft.name)

    # TODO: copy/fill-in the properties from the draft to the blueprint
    scheduling_unit_blueprint = SchedulingUnitBlueprint.objects.create(
        name="%s (SchedulingUnitBlueprint)" % (scheduling_unit_draft.name,),
        description="%s (SchedulingUnitBlueprint)" % (scheduling_unit_draft.description or "<no description>",),
        requirements_doc=scheduling_unit_draft.requirements_doc,
        do_cancel=False,
        draft=scheduling_unit_draft,
        requirements_template=scheduling_unit_draft.requirements_template)

    logger.info("create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft.id=%s name='%s') created scheduling_unit_blueprint id=%s name='%s'",
                scheduling_unit_draft.pk, scheduling_unit_draft.name, scheduling_unit_blueprint.pk, scheduling_unit_blueprint.name)
    return scheduling_unit_blueprint


def create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitDraft:
    """
    Create all defined task_drafts in the scheduling_unit_draft's requirements_doc, connect them, and return the updated scheduling_unit_draft.
    """
    logger.debug("create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft.id=%s, name='%s') ...", scheduling_unit_draft.pk, scheduling_unit_draft.name)

    if len(scheduling_unit_draft.requirements_doc.get("tasks",[])) == 0:
        raise BlueprintCreationException("create_task_drafts_from_scheduling_unit_draft: scheduling_unit_draft.id=%s has no tasks defined in its requirements_doc" % (scheduling_unit_draft.pk,))

    for task_definition in scheduling_unit_draft.requirements_doc["tasks"]:
        task_template_name = task_definition["specifications_template"]
        task_template = models.TaskTemplate.objects.get(name=task_template_name)

        if scheduling_unit_draft.task_drafts.filter(name=task_definition["name"], specifications_template=task_template).count() > 0:
            logger.debug("skipping creation of task draft because it is already in the scheduling_unit... task_name='%s', task_template_name='%s'", task_definition["name"], task_template_name)
            continue

        logger.debug("creating task draft... task_name='%s', task_template_name='%s'", task_definition["name"], task_template_name)

        task_draft = models.TaskDraft.objects.create(name=task_definition["name"],
                                                     description=task_definition.get("description",""),
                                                     tags=task_definition.get("tags",[]),
                                                     specifications_doc=task_definition["specifications_doc"],
                                                     copy_reason=models.CopyReason.objects.get(value='template'),
                                                     copies=None,
                                                     scheduling_unit_draft=scheduling_unit_draft,
                                                     specifications_template=task_template)

        logger.info("created task draft id=%s task_name='%s', task_template_name='%s'", task_draft.pk, task_definition["name"], task_template_name)

    # Now create task relations
    for task_relation_definition in scheduling_unit_draft.requirements_doc["task_relations"]:
        producer_task_draft = scheduling_unit_draft.task_drafts.get(name=task_relation_definition["producer"])
        consumer_task_draft = scheduling_unit_draft.task_drafts.get(name=task_relation_definition["consumer"])
        dataformat = models.Dataformat.objects.get(value=task_relation_definition["dataformat"])
        input_role = models.TaskConnectorType.objects.get(role=task_relation_definition["input"]["role"], datatype=task_relation_definition["input"]["datatype"])
        output_role = models.TaskConnectorType.objects.get(role=task_relation_definition["output"]["role"], datatype=task_relation_definition["output"]["datatype"])
        selection_template = models.TaskRelationSelectionTemplate.objects.get(name=task_relation_definition["selection_template"])

        if models.TaskRelationDraft.objects.filter(producer=producer_task_draft,
                                                   consumer=consumer_task_draft,
                                                   dataformat=dataformat,
                                                   input_role=input_role,
                                                   output_role=output_role,
                                                   selection_template=selection_template,
                                                   selection_doc=task_relation_definition["selection_doc"]).count() > 0:
            logger.debug("skipping creation of task_relation between task draft '%s' and '%s' because it is already in the scheduling_unit...", task_relation_definition["producer"], task_relation_definition["consumer"])
            continue

        task_relation = models.TaskRelationDraft.objects.create(tags=task_relation_definition.get("tags",[]),
                                                                selection_doc=task_relation_definition["selection_doc"],
                                                                dataformat=dataformat,
                                                                producer=producer_task_draft,
                                                                consumer=consumer_task_draft,
                                                                input_role=input_role,
                                                                output_role=output_role,
                                                                selection_template=selection_template)
        logger.info("created task_relation id=%s between task draft id=%s name='%s' and id=%s name='%s",
                    task_relation.pk, producer_task_draft.id, producer_task_draft.name, consumer_task_draft.id, consumer_task_draft.name)


    # task_scheduling_relation
    for task_scheduling_relation_definition in scheduling_unit_draft.requirements_doc["task_scheduling_relations"]:
        placement = models.SchedulingRelationPlacement.objects.get(value=task_scheduling_relation_definition["placement"])
        time_offset = task_scheduling_relation_definition["time_offset"]
        first_task_draft = scheduling_unit_draft.task_drafts.get(name=task_scheduling_relation_definition["first"])
        second_task_draft = scheduling_unit_draft.task_drafts.get(name=task_scheduling_relation_definition["second"])

        if models.TaskSchedulingRelationDraft.objects.filter(placement=placement,
                                                             time_offset=time_offset,
                                                             first=first_task_draft,
                                                             second=second_task_draft).count() > 0:
            logger.debug("skipping creation of task_scheduling_relation between task draft '%s' and '%s' because it is already in the scheduling_unit...",
                        task_scheduling_relation_definition["first"], task_scheduling_relation_definition["second"])
            continue

        task_scheduling_relation = models.TaskSchedulingRelationDraft.objects.create(placement=placement,
                                                                                     time_offset=time_offset,
                                                                                     first=first_task_draft,
                                                                                     second=second_task_draft)
        logger.info("created task_scheduling_relation id=%s between task draft id=%s name='%s' and id=%s name='%s",
                    task_scheduling_relation.pk, first_task_draft.id, first_task_draft.name, second_task_draft.id, second_task_draft.name)

    logger.info("create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft.id=%s, name='%s') ... done", scheduling_unit_draft.pk, scheduling_unit_draft.name)

    scheduling_unit_draft.refresh_from_db()
    return scheduling_unit_draft


def create_task_blueprint_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    """
    Create a task_blueprint from the task_draft
    :raises Exception if instantiate fails.
    """
    logger.debug("creating task_blueprint from task_draft id=%s", task_draft.pk)

    # get or create a scheduling_unit_blueprint from the scheduling_unit_draft
    scheduling_unit_blueprint = task_draft.scheduling_unit_draft.scheduling_unit_blueprints.last()
    if scheduling_unit_blueprint is None:
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(task_draft.scheduling_unit_draft)

    task_blueprint = TaskBlueprint.objects.create(
        description=task_draft.description,
        name=task_draft.name,
        do_cancel=False,
        draft=task_draft,
        scheduling_unit_blueprint=scheduling_unit_blueprint,
        specifications_doc=task_draft.specifications_doc,
        specifications_template=task_draft.specifications_template)

    logger.info("created task_blueprint id=%s from task_draft id=%s", task_blueprint.pk, task_draft.pk)

    # now that we have a task_blueprint, its time to refresh the task_draft so we get the non-cached fields
    task_draft.refresh_from_db()

    # loop over consumers/producers, and 'copy' the TaskRelationBlueprint from the TaskRelationDraft
    # this is only possible if both 'ends' of the task_relation are converted to a TaskBlueprint
    # so, when converting two TaskDrafts (for example an observation and a pipeline), then for the conversion
    # of the first TaskDraft->TaskBlueprint no relation is setup,
    # and for the second TaskDraft->TaskBlueprint conversion we have both endpoints available, so we can connect them.
    task_draft_relations = list(task_draft.consumed_by.all()) + list(task_draft.produced_by.all())
    for task_relation_draft in task_draft_relations:
        for producing_task_blueprint in task_relation_draft.producer.task_blueprints.all():
            for consuming_task_blueprint in task_relation_draft.consumer.task_blueprints.all():
                try:
                    # do nothing if task_relation_blueprint already exists...
                    models.TaskRelationBlueprint.objects.get(producer_id=producing_task_blueprint.id, consumer_id=consuming_task_blueprint.id)
                except models.TaskRelationBlueprint.DoesNotExist:
                    # ...'else' create it.
                    task_relation_blueprint = models.TaskRelationBlueprint.objects.create(draft=task_relation_draft,
                                                                                          input_role=task_relation_draft.input_role,
                                                                                          output_role=task_relation_draft.output_role,
                                                                                          producer=producing_task_blueprint,
                                                                                          consumer=consuming_task_blueprint,
                                                                                          selection_doc=task_relation_draft.selection_doc,
                                                                                          selection_template=task_relation_draft.selection_template,
                                                                                          dataformat=task_relation_draft.dataformat)
                    logger.info("created task_relation_blueprint id=%s which connects task_blueprints producer_id=%s and consumer_id=%s",
                                task_relation_blueprint.pk, producing_task_blueprint.pk, consuming_task_blueprint.pk,)

    # Do the same 'trick' for Task Scheduling Relation Draft to Blueprint
    task_draft_scheduling_relations = list(task_draft.first_to_connect.all()) + list(task_draft.second_to_connect.all())
    for task_scheduling_relation_draft in task_draft_scheduling_relations:
        for first_task_blueprint in task_scheduling_relation_draft.first.task_blueprints.all():
            for second_task_blueprint in task_scheduling_relation_draft.second.task_blueprints.all():
                try:
                    # do nothing if task_scheduling_relation_blueprint already exists...
                    models.TaskSchedulingRelationBlueprint.objects.get(first_id=first_task_blueprint.id,
                                                                       second_id=second_task_blueprint.id)
                except models.TaskSchedulingRelationBlueprint.DoesNotExist:
                    # ...'else' create it.
                    task_scheduling_relation_blueprint = models.TaskSchedulingRelationBlueprint.objects.create(first=first_task_blueprint,
                                                                                                               second=second_task_blueprint,
                                                                                                               time_offset=task_scheduling_relation_draft.time_offset,
                                                                                                               placement=task_scheduling_relation_draft.placement)
                    logger.info("created task_scheduling_relation_blueprint id=%s which connects task_blueprints first_id=%s and second_id=%s, placement=%s time_offset=%s[sec]",
                                task_scheduling_relation_blueprint.pk, first_task_blueprint.pk, second_task_blueprint.pk, task_scheduling_relation_draft.placement, task_scheduling_relation_draft.time_offset)

    return task_blueprint


def create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint, then create its child task_blueprint(s), then create the task_blueprint's subtasks'''

    # make sure we call create task_drafts first, so the task_blueprints can be created from the scheduling_unit_blueprint in the next step
    # already known task_drafts and/or relations are skipped automagically
    create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft)

    scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
    return create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)


def create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint, then create its child task_blueprint(s), then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
    return create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)


def create_task_blueprint_and_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    '''Convenience method: Create the task_blueprint, then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    task_blueprint =  create_task_blueprint_from_task_draft(task_draft)
    create_subtasks_from_task_blueprint(task_blueprint)
    task_blueprint.refresh_from_db()
    return task_blueprint


def create_task_blueprint_and_subtasks_and_schedule_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    '''Convenience method: Create the task_blueprint, then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    task_blueprint =  create_task_blueprint_from_task_draft(task_draft)
    create_and_schedule_subtasks_from_task_blueprint(task_blueprint)
    task_blueprint.refresh_from_db()
    return task_blueprint


def create_task_blueprints_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint's task_blueprint(s)'''

    # make sure we call create task_drafts first, so the task_blueprints can be created from the scheduling_unit_blueprint in the next step
    # already known task_drafts and/or relations are skipped automagically
    create_task_drafts_from_scheduling_unit_draft(scheduling_unit_blueprint.draft)

    task_drafts = list(scheduling_unit_blueprint.draft.task_drafts.all())

    # sort them in 'data-flow'-order,
    # because successors can depend on predecessors, so the first taskdraft's need to be blueprinted first.
    task_drafts.sort(key=cmp_to_key(lambda taskdraft_a, taskdraft_b: -1 if taskdraft_a in taskdraft_b.predecessors else 1 if taskdraft_b in taskdraft_a.predecessors else 0))

    # convert task_draft(s) to task_blueprint(s)
    for task_draft in task_drafts:
        create_task_blueprint_from_task_draft(task_draft)

    # refresh so all related fields are updated.
    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint


def create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint's task_blueprint(s), then create each task_blueprint's subtasks'''
    scheduling_unit_blueprint = create_task_blueprints_from_scheduling_unit_blueprint(scheduling_unit_blueprint)

    task_blueprints = list(scheduling_unit_blueprint.task_blueprints.all())

    # sort task_blueprint(s) in 'data-flow'-order,
    # because successors can depend on predecessors, so the first tbp's need to be subtask'd first.
    task_blueprints.sort(key=cmp_to_key(lambda tbp_a, tbp_b: -1 if tbp_a in tbp_b.predecessors else 1 if tbp_b in tbp_a.predecessors else 0))

    for task_blueprint in task_blueprints:
        create_subtasks_from_task_blueprint(task_blueprint)

    # refresh so all related fields are updated.
    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint


def create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    '''Convenience method: Create the scheduling_unit_blueprint's task_blueprint(s), then create the task_blueprint's subtasks, and schedule the ones that are not dependend on predecessors'''
    scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)
    task_blueprints = list(scheduling_unit_blueprint.task_blueprints.all())

    # sort task_blueprint(s) in 'data-flow'-order,
    # because successors can depend on predecessors, so the first tbp's need to be subtask'd first.
    task_blueprints.sort(key=cmp_to_key(lambda tbp_a, tbp_b: -1 if tbp_a in tbp_b.predecessors else 1 if tbp_b in tbp_a.predecessors else 0))

    for task_blueprint in task_blueprints:
        schedule_independent_subtasks_in_task_blueprint(task_blueprint)

    # refresh so all related fields are updated.
    scheduling_unit_blueprint.refresh_from_db()
    return scheduling_unit_blueprint
