"""
This file contains the viewsets (based on the elsewhere defined data models and serializers)
"""

from django.shortcuts import get_object_or_404
from rest_framework import viewsets
from .lofar_viewset import LOFARViewSet, LOFARNestedViewSet
from .. import models
from .. import serializers
from django_filters import rest_framework as filters
from lofar.sas.tmss.tmss.tmssapp.models.scheduling import Subtask
# Don't use Ordering class from the django_filters but use rest_framework instead !!
from rest_framework.filters import OrderingFilter
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from drf_yasg.inspectors import SwaggerAutoSchema

from rest_framework.decorators import action
from django.http import HttpResponse, JsonResponse
from rest_framework.response import Response as RestResponse

from lofar.sas.tmss.tmss.tmssapp.viewsets.lofar_viewset import LOFARViewSet
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import serializers

from datetime import datetime
from lofar.common.json_utils import get_default_json_object_for_schema
from lofar.common.datetimeutils import formatDatetime
from lofar.sas.tmss.tmss.tmssapp.adapters.parset import convert_to_parset
from drf_yasg.renderers import _SpecRenderer


from lofar.sas.tmss.tmss.tmssapp.renderers import PlainTextRenderer
from rest_framework.views import APIView
from rest_framework.decorators import api_view, renderer_classes


class TextPlainAutoSchema(SwaggerAutoSchema):
    def get_produces(self):
        return ["text/plain"]


class SubtaskStateViewSet(LOFARViewSet):
    queryset = models.SubtaskState.objects.all()
    serializer_class = serializers.SubtaskStateSerializer


class SubtaskStateLogViewSet(LOFARViewSet):
    queryset = models.SubtaskStateLog.objects.all()
    serializer_class = serializers.SubtaskStateLogSerializer

    def get_queryset(self):
        if 'subtask_id' in self.kwargs:
            subtask = get_object_or_404(models.Subtask, pk=self.kwargs['subtask_id'])
            return subtask.subtaskstatelog_set.order_by('created_at').all()

        queryset = models.SubtaskStateLog.objects.all()

        # query by subtask
        subtask = self.request.query_params.get('subtask', None)
        if subtask is not None:
            return queryset.filter(subtask=subtask)

        return queryset


class SubtaskTypeViewSet(LOFARViewSet):
    queryset = models.SubtaskType.objects.all()
    serializer_class = serializers.SubtaskTypeSerializer

class StationTypeViewSet(LOFARViewSet):
    queryset = models.StationType.objects.all()
    serializer_class = serializers.StationTypeSerializer

class AlgorithmViewSet(LOFARViewSet):
    queryset = models.Algorithm.objects.all()
    serializer_class = serializers.AlgorithmSerializer

class ScheduleMethodViewSet(LOFARViewSet):
    queryset = models.ScheduleMethod.objects.all()
    serializer_class = serializers.ScheduleMethodSerializer

class SubtaskTemplateFilter(filters.FilterSet):
    class Meta:
        model = models.SubtaskTemplate
        fields = {
            'name': ['exact'],
            'version': ['lt', 'gt', 'exact']
        }

class SubtaskTemplateViewSet(LOFARViewSet):
    queryset = models.SubtaskTemplate.objects.all()
    serializer_class = serializers.SubtaskTemplateSerializer
    filter_class = SubtaskTemplateFilter

    def get_queryset(self):
        queryset = models.SubtaskTemplate.objects.all()

        # query by name
        name = self.request.query_params.get('name', None)
        if name is not None:
            return queryset.filter(name=name)

        return queryset

    @swagger_auto_schema(responses={200: 'The schema as a JSON object',
                                    403: 'forbidden'},
                         operation_description="Get the schema as a JSON object.")
    @action(methods=['get'], detail=True)
    def schema(self, request, pk=None):
        subtask_template = get_object_or_404(models.SubtaskTemplate, pk=pk)
        return JsonResponse(subtask_template.schema)

    @swagger_auto_schema(responses={200: 'JSON object with all the defaults from the schema filled in',
                                    403: 'forbidden'},
                         operation_description="Get a JSON object with all the defaults from the schema filled in.")
    @action(methods=['get'], detail=True)
    def default_specification(self, request, pk=None):
        subtask_template = get_object_or_404(models.SubtaskTemplate, pk=pk)
        spec = get_default_json_object_for_schema(subtask_template.schema)
        return JsonResponse(spec)


class DefaultSubtaskTemplateViewSet(LOFARViewSet):
    queryset = models.DefaultSubtaskTemplate.objects.all()
    serializer_class = serializers.DefaultSubtaskTemplateSerializer


class DataproductSpecificationsTemplateViewSet(LOFARViewSet):
    queryset = models.DataproductSpecificationsTemplate.objects.all()
    serializer_class = serializers.DataproductSpecificationsTemplateSerializer


class DefaultDataproductSpecificationsTemplateViewSet(LOFARViewSet):
    queryset = models.DefaultDataproductSpecificationsTemplate.objects.all()
    serializer_class = serializers.DefaultDataproductSpecificationsTemplateSerializer

class DataproductFeedbackTemplateViewSet(LOFARViewSet):
    queryset = models.DataproductFeedbackTemplate.objects.all()
    serializer_class = serializers.DataproductFeedbackTemplateSerializer


class SubTaskFilter(filters.FilterSet):
    class Meta:
        model = Subtask
        fields = {
            'state__value': ['exact'],
            'start_time': ['lt', 'gt'],
            'stop_time': ['lt', 'gt'],
            'cluster__name': ['exact', 'icontains'],
        }


class SubtaskViewSet(LOFARViewSet):
    queryset = models.Subtask.objects.all()
    serializer_class = serializers.SubtaskSerializer
    filter_backends = (filters.DjangoFilterBackend, OrderingFilter,)
    filter_class = SubTaskFilter
    ordering = ('start_time',)

    @swagger_auto_schema(auto_schema=TextPlainAutoSchema,
                         responses={200: 'A LOFAR parset for this subtask (as plain text)',
                                    403: 'forbidden',
                                    404: 'Not found'},
                         produces='text/plain',
                         operation_description="Get a LOFAR parset for the specifications of this subtask")
    @action(methods=['get'], detail=True, renderer_classes=[PlainTextRenderer])
    def parset(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        parset = convert_to_parset(subtask)

        header = "# THIS PARSET WAS GENERATED BY TMSS FROM THE SPECIFICATION OF SUBTASK ID=%d ON %s\n" % (subtask.pk, formatDatetime(datetime.utcnow()))
        parset_str = header + str(parset)
        return HttpResponse(parset_str, content_type='text/plain')


    @swagger_auto_schema(responses={200: 'The predecessor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the predecessor subtasks of this subtask.")
    @action(methods=['get'], detail=True, url_name="predecessors")
    def predecessors(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        predecessors = self.filter_queryset(subtask.predecessors)
        serializer = self.get_serializer(predecessors, many=True)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The successor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the successor subtasks of this subtask.")
    @action(methods=['get'], detail=True, url_name="successors")
    def successors(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        successors = self.filter_queryset(subtask.successors)
        serializer = self.get_serializer(successors, many=True)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The scheduled version of this subtask',
                                    403: 'forbidden',
                                    500: 'The subtask could not be scheduled'},
                         operation_description="Try to schedule this subtask.")
    @action(methods=['get'], detail=True, url_name="schedule")
    def schedule(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        from lofar.sas.tmss.tmss.tmssapp.subtasks import schedule_subtask
        scheduled_subtask = schedule_subtask(subtask)
        serializer = self.get_serializer(scheduled_subtask)
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The state log for this Subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the state log for this Subtask.")
    @action(methods=['get'], detail=True)
    def state_log(self, request, pk=None):
        subtask = get_object_or_404(models.Subtask, pk=pk)
        serializer = serializers.SubtaskStateLogSerializer(subtask.subtaskstatelog_set.order_by('created_at').all(),
                                                           many=True,
                                                           context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The input dataproducts of this subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the input dataproducts of this subtask.")
    @action(methods=['get'], detail=True, url_name='input_dataproducts')
    def input_dataproducts(self, request, pk=None):
        dataproducts = models.Dataproduct.objects.filter(subtaskinput__subtask_id=pk)
        serializer = serializers.DataproductSerializer(dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


    @swagger_auto_schema(responses={200: 'The output dataproducts of this subtask.',
                                    403: 'forbidden'},
                         operation_description="Get the output dataproducts of this subtask.")
    @action(methods=['get'], detail=True, url_name='output_dataproducts')
    def output_dataproducts(self, request, pk=None):
        dataproducts = models.Dataproduct.objects.filter(producer__subtask_id=pk)
        serializer = serializers.DataproductSerializer(dataproducts, many=True, context={'request': request})
        return RestResponse(serializer.data)


class SubtaskNestedViewSet(LOFARNestedViewSet):
    queryset = models.Subtask.objects.all()
    serializer_class = serializers.SubtaskSerializer
    filter_backends = (filters.DjangoFilterBackend,)
    filter_class = SubTaskFilter
    ordering = ('start_time',)

    def get_queryset(self):
        if 'task_blueprint_id' in self.kwargs:
            task_blueprint = get_object_or_404(models.TaskBlueprint, pk=self.kwargs['task_blueprint_id'])
            return task_blueprint.subtasks.all()


class SubtaskInputViewSet(LOFARViewSet):
    queryset = models.SubtaskInput.objects.all()
    serializer_class = serializers.SubtaskInputSerializer


class SubtaskOutputViewSet(LOFARViewSet):
    queryset = models.SubtaskOutput.objects.all()
    serializer_class = serializers.SubtaskOutputSerializer


class DataproductViewSet(LOFARViewSet):
    queryset = models.Dataproduct.objects.all()
    serializer_class = serializers.DataproductSerializer


class AntennaSetViewSet(LOFARViewSet):
    queryset = models.AntennaSet.objects.all()
    serializer_class = serializers.AntennaSetSerializer


class DataproductTransformViewSet(LOFARViewSet):
    queryset = models.DataproductTransform.objects.all()
    serializer_class = serializers.DataproductTransformSerializer


class FilesystemViewSet(LOFARViewSet):
    queryset = models.Filesystem.objects.all()
    serializer_class = serializers.FilesystemSerializer


class ClusterViewSet(LOFARViewSet):
    queryset = models.Cluster.objects.all()
    serializer_class = serializers.ClusterSerializer


class DataproductArchiveInfoViewSet(LOFARViewSet):
    queryset = models.DataproductArchiveInfo.objects.all()
    serializer_class = serializers.DataproductArchiveInfoSerializer


class DataproductHashViewSet(LOFARViewSet):
    queryset = models.DataproductHash.objects.all()
    serializer_class = serializers.DataproductHashSerializer


