"""
This file contains the viewsets (based on the elsewhere defined data models and serializers)
"""

from django.shortcuts import get_object_or_404
from django.http import JsonResponse
from django.contrib.auth.models import User
from django_filters import rest_framework as filters
from rest_framework.viewsets import ReadOnlyModelViewSet
from rest_framework import status
from rest_framework.response import Response

from rest_framework.decorators import permission_classes
from rest_framework.permissions import IsAuthenticatedOrReadOnly, DjangoModelPermissions
from rest_framework.decorators import action

from drf_yasg.utils import swagger_auto_schema

from lofar.sas.tmss.tmss.tmssapp.viewsets.lofar_viewset import LOFARViewSet, LOFARNestedViewSet
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import serializers

from datetime import datetime
from lofar.common.json_utils import get_default_json_object_for_schema
from lofar.common.datetimeutils import formatDatetime
from lofar.sas.tmss.tmss.tmssapp.tasks import *
from lofar.sas.tmss.tmss.tmssapp.subtasks import *




# This is required for keeping a user reference as ForeignKey in other models
# (I think so that the HyperlinkedModelSerializer can generate a URI)
class UserViewSet(ReadOnlyModelViewSet):
    queryset = User.objects.all()
    serializer_class = serializers.UserSerializer


class TagsViewSet(LOFARViewSet):
    queryset = models.Tags.objects.all()
    serializer_class = serializers.TagsSerializer


class GeneratorTemplateViewSet(LOFARViewSet):
    queryset = models.GeneratorTemplate.objects.all()
    serializer_class = serializers.GeneratorTemplateSerializer


class DefaultGeneratorTemplateViewSet(LOFARViewSet):
    queryset = models.DefaultGeneratorTemplate.objects.all()
    serializer_class = serializers.DefaultGeneratorTemplateSerializer

class SchedulingUnitTemplateFilter(filters.FilterSet):
    class Meta:
        model = models.SchedulingUnitTemplate
        fields = {
            'name': ['exact'],
            'version': ['lt', 'gt', 'exact']
        }

class SchedulingUnitTemplateViewSet(LOFARViewSet):
    queryset = models.SchedulingUnitTemplate.objects.all()
    serializer_class = serializers.SchedulingUnitTemplateSerializer
    filter_class = SchedulingUnitTemplateFilter

    @swagger_auto_schema(responses={200: 'JSON object with all the defaults from the schema filled in',
                                    403: 'forbidden'},
                         operation_description="Get a JSON object with all the defaults from the schema filled in.")
    @action(methods=['get'], detail=True)
    def default_specification(self, request, pk=None):
        schedulingunit_template = get_object_or_404(models.SchedulingUnitTemplate, pk=pk)
        spec = get_default_json_object_for_schema(schedulingunit_template.schema)
        return JsonResponse(spec)


class DefaultSchedulingUnitTemplateViewSet(LOFARViewSet):
    queryset = models.DefaultSchedulingUnitTemplate.objects.all()
    serializer_class = serializers.DefaultSchedulingUnitTemplateSerializer

class TaskTemplateFilter(filters.FilterSet):
    class Meta:
        model = models.TaskTemplate
        fields = {
            'name': ['exact'],
            'version': ['lt', 'gt', 'exact']
        }

class TaskTemplateViewSet(LOFARViewSet):
    queryset = models.TaskTemplate.objects.all()
    serializer_class = serializers.TaskTemplateSerializer
    filter_class = TaskTemplateFilter

    @swagger_auto_schema(responses={200: 'The schema as a JSON object',
                                    403: 'forbidden'},
                         operation_description="Get the schema as a JSON object.")
    @action(methods=['get'], detail=True)
    def schema(self, request, pk=None):
        template = get_object_or_404(models.TaskTemplate, pk=pk)
        return JsonResponse(template.schema)

    @swagger_auto_schema(responses={200: 'JSON object with all the defaults from the schema filled in',
                                    403: 'forbidden'},
                         operation_description="Get a JSON object with all the defaults from the schema filled in.")
    @action(methods=['get'], detail=True)
    def default_specification(self, request, pk=None):
        template = get_object_or_404(models.TaskTemplate, pk=pk)
        spec = get_default_json_object_for_schema(template.schema)
        return JsonResponse(spec)


class DefaultTaskTemplateViewSet(LOFARViewSet):
    queryset = models.DefaultTaskTemplate.objects.all()
    serializer_class = serializers.DefaultTaskTemplateSerializer


class TaskRelationSelectionTemplateViewSet(LOFARViewSet):
    queryset = models.TaskRelationSelectionTemplate.objects.all()
    serializer_class = serializers.TaskRelationSelectionTemplateSerializer


class DefaultTaskRelationSelectionTemplateViewSet(LOFARViewSet):
    queryset = models.DefaultTaskRelationSelectionTemplate.objects.all()
    serializer_class = serializers.DefaultTaskRelationSelectionTemplateSerializer


class RoleViewSet(LOFARViewSet):
    queryset = models.Role.objects.all()
    serializer_class = serializers.RoleSerializer


class SchedulingRelationPlacement(LOFARViewSet):
    queryset = models.SchedulingRelationPlacement.objects.all()
    serializer_class = serializers.SchedulingRelationPlacementSerializer


class DatatypeViewSet(LOFARViewSet):
    queryset = models.Datatype.objects.all()
    serializer_class = serializers.DatatypeSerializer


class DataformatViewSet(LOFARViewSet):
    queryset = models.Dataformat.objects.all()
    serializer_class = serializers.DataformatSerializer


class CopyReasonViewSet(LOFARViewSet):
    queryset = models.CopyReason.objects.all()
    serializer_class = serializers.CopyReasonSerializer

    
class TaskConnectorTypeViewSet(LOFARViewSet):
    queryset = models.TaskConnectorType.objects.all()
    serializer_class = serializers.TaskConnectorTypeSerializer


@permission_classes((DjangoModelPermissions,))   # example override of default permissions per viewset | todo: review for production
class CycleViewSet(LOFARViewSet):
    queryset = models.Cycle.objects.all()
    serializer_class = serializers.CycleSerializer
    ordering = ['start']


class CycleQuotaViewSet(LOFARViewSet):
    queryset = models.CycleQuota.objects.all()
    serializer_class = serializers.CycleQuotaSerializer

    def get_queryset(self):
        queryset = models.CycleQuota.objects.all()

        # query by project
        project = self.request.query_params.get('project', None)
        if project is not None:
            return queryset.filter(project=project)

        return queryset


class ProjectViewSet(LOFARViewSet):
    queryset = models.Project.objects.all()
    serializer_class = serializers.ProjectSerializer
    ordering = ['name']

    def get_queryset(self):
        queryset = models.Project.objects.all()

        # query by cycle
        cycle = self.request.query_params.get('cycle', None)
        if cycle is not None:
            return queryset.filter(cycles__name=cycle)

        return queryset


class ProjectNestedViewSet(LOFARNestedViewSet):
    queryset = models.Project.objects.all()
    serializer_class = serializers.ProjectSerializer

    def get_queryset(self):
        if 'cycle_id' in self.kwargs:
            cycle = get_object_or_404(models.Cycle, pk=self.kwargs['cycle_id'])
            return cycle.projects.all()
        else:
            return models.Project.objects.all()


class ProjectQuotaViewSet(LOFARViewSet):
    queryset = models.ProjectQuota.objects.all()
    serializer_class = serializers.ProjectQuotaSerializer

    def get_queryset(self):
        queryset = models.ProjectQuota.objects.all()

        # query by project
        project = self.request.query_params.get('project', None)
        if project is not None:
            return queryset.filter(project=project)

        return queryset
    

class ResourceTypeViewSet(LOFARViewSet):
    queryset = models.ResourceType.objects.all()
    serializer_class = serializers.ResourceTypeSerializer


class SchedulingSetViewSet(LOFARViewSet):
    queryset = models.SchedulingSet.objects.all()
    serializer_class = serializers.SchedulingSetSerializer


class FlagViewSet(LOFARViewSet):
    queryset = models.Flag.objects.all()
    serializer_class = serializers.FlagSerializer


class SettingViewSet(LOFARViewSet):
    queryset = models.Setting.objects.all()
    serializer_class = serializers.SettingSerializer


class PeriodCategoryViewSet(LOFARViewSet):
    queryset = models.PeriodCategory.objects.all()
    serializer_class = serializers.PeriodCategorySerializer


class ProjectCategoryViewSet(LOFARViewSet):
    queryset = models.ProjectCategory.objects.all()
    serializer_class = serializers.ProjectCategorySerializer


class SchedulingUnitDraftViewSet(LOFARViewSet):
    queryset = models.SchedulingUnitDraft.objects.all()
    serializer_class = serializers.SchedulingUnitDraftSerializer

    @swagger_auto_schema(responses={201: 'The Created SchedulingUnitBlueprint, see Location in Response header',
                                    403: 'forbidden'},
                         operation_description="Carve SchedulingUnitDraft in stone, and make an (uneditable) blueprint out of it.")
    @action(methods=['get'], detail=True, url_name="create_task_blueprint", name="Create SchedulingUnitBlueprint")
    def create_scheduling_unit_blueprint(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)

        # url path magic to construct the new scheduling_unit_blueprint_path url
        scheduling_unit_draft_path = request._request.path
        base_path = scheduling_unit_draft_path[:scheduling_unit_draft_path.find('/scheduling_unit_draft')]
        scheduling_unit_blueprint_path = '%s/scheduling_unit_blueprint/%s/' % (base_path, scheduling_unit_blueprint.id,)

        # return a response with the new serialized SchedulingUnitBlueprintSerializer, and a Location to the new instance in the header
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED,
                        headers={'Location': scheduling_unit_blueprint_path})

    @swagger_auto_schema(responses={201: 'The Created SchedulingUnitBlueprint, see Location in Response header',
                                    403: 'forbidden'},
                         operation_description="Carve this SchedulingUnitDraft and its TaskDraft(s) in stone, and make blueprint(s) out of it and create their subtask(s), and schedule the ones that are not dependend on predecessors")
    @action(methods=['get'], detail=True, url_name="create_blueprints_and_schedule", name="Create Blueprints-Tree and Schedule")
    def create_blueprints_and_schedule(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        scheduling_unit_blueprint = create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # url path magic to construct the new scheduling_unit_blueprint_path url
        scheduling_unit_draft_path = request._request.path
        base_path = scheduling_unit_draft_path[:scheduling_unit_draft_path.find('/scheduling_unit_draft')]
        scheduling_unit_blueprint_path = '%s/scheduling_unit_blueprint/%s/' % (base_path, scheduling_unit_blueprint.id,)

        # return a response with the new serialized SchedulingUnitBlueprintSerializer, and a Location to the new instance in the header
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED,
                        headers={'Location': scheduling_unit_blueprint_path})
    @swagger_auto_schema(responses={201: 'The Created SchedulingUnitBlueprint, see Location in Response header',
                                    403: 'forbidden'},
                         operation_description="Carve this SchedulingUnitDraft and its TaskDraft(s) in stone, and make blueprint(s) out of it and create their subtask(s)")
    @action(methods=['get'], detail=True, url_name="create_blueprints_and_subtasks", name="Create Blueprints-Tree")
    def create_blueprints_and_subtasks(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        # url path magic to construct the new scheduling_unit_blueprint_path url
        scheduling_unit_draft_path = request._request.path
        base_path = scheduling_unit_draft_path[:scheduling_unit_draft_path.find('/scheduling_unit_draft')]
        scheduling_unit_blueprint_path = '%s/scheduling_unit_blueprint/%s/' % (base_path, scheduling_unit_blueprint.id,)

        # return a response with the new serialized SchedulingUnitBlueprintSerializer, and a Location to the new instance in the header
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED,
                        headers={'Location': scheduling_unit_blueprint_path})


    @swagger_auto_schema(responses={201: 'The updated scheduling_unit_draft with references to its created task_drafts',
                                    403: 'forbidden'},
                         operation_description="Create Task Drafts from SchedulingUnitDraft.")
    @action(methods=['get'], detail=True, url_name="create_task_drafts", name="Create Task Drafts from Requirement doc")
    def create_task_drafts(self, request, pk=None):
        scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=pk)
        create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft)

        # just return as a response the serialized scheduling_unit_draft (with references to the created task_draft(s)
        return Response(serializers.SchedulingUnitDraftSerializer(scheduling_unit_draft,
                        context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


class SchedulingUnitDraftNestedViewSet(LOFARNestedViewSet):
    queryset = models.SchedulingUnitDraft.objects.all()
    serializer_class = serializers.SchedulingUnitDraftSerializer

    def get_queryset(self):
        if 'scheduling_set_id' in self.kwargs:
            scheduling_set = get_object_or_404(models.SchedulingSet, pk=self.kwargs['scheduling_set_id'])
            return scheduling_set.scheduling_unit_drafts.all()
        else:
            return models.SchedulingUnitDraft.objects.all()


class SchedulingUnitBlueprintViewSet(LOFARViewSet):
    queryset = models.SchedulingUnitBlueprint.objects.all()
    serializer_class = serializers.SchedulingUnitBlueprintSerializer

    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint, with references to its created TaskBlueprints and (scheduled) Subtasks.",
                                    403: 'forbidden'},
                         operation_description="Create TaskBlueprint(s) for this scheduling unit, create subtasks, and schedule the ones that are not dependend on predecessors.")
    @action(methods=['get'], detail=True, url_name="create_taskblueprints_subtasks_and_schedule", name="Create TaskBlueprint(s), their Subtask(s) and schedule them.")
    def create_taskblueprints_subtasks_and_schedule(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        scheduling_unit_blueprint = create_task_blueprints_and_subtasks_and_schedule_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)

        # return a response with the new serialized scheduling_unit_blueprint (with references to the created task_blueprint(s) and (scheduled) subtasks)
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint, with references to its created TaskBlueprints and Subtasks.",
                                    403: 'forbidden'},
                         operation_description="Create TaskBlueprint(s) for this scheduling unit and create subtasks.")
    @action(methods=['get'], detail=True, url_name="create_taskblueprints_subtasks", name="Create TaskBlueprint(s) and their Subtask(s)")
    def create_taskblueprints_subtasks(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)

        # return a response with the new serialized scheduling_unit_blueprint (with references to the created task_blueprint(s) and (scheduled) subtasks)
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={201: "This SchedulingUnitBlueprint, with references to its created TaskBlueprints.",
                                    403: 'forbidden'},
                         operation_description="Create the TaskBlueprint(s).")
    @action(methods=['get'], detail=True, url_name="create_taskblueprints", name="Create TaskBlueprint(s)")
    def create_taskblueprints(self, request, pk=None):
        scheduling_unit_blueprint = get_object_or_404(models.SchedulingUnitBlueprint, pk=pk)
        scheduling_unit_blueprint = create_task_blueprints_from_scheduling_unit_blueprint(scheduling_unit_blueprint)

        # return a response with the new serialized scheduling_unit_blueprint (with references to the created task_blueprint(s) and (scheduled) subtasks)
        return Response(serializers.SchedulingUnitBlueprintSerializer(scheduling_unit_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)


class SchedulingUnitBlueprintNestedViewSet(LOFARNestedViewSet):
    queryset = models.SchedulingUnitBlueprint.objects.all()
    serializer_class = serializers.SchedulingUnitBlueprintSerializer

    def get_queryset(self):
        if 'scheduling_unit_draft_id' in self.kwargs:
            scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=self.kwargs['scheduling_unit_draft_id'])
            return scheduling_unit_draft.scheduling_unit_blueprints.all()
        else:
            return models.SchedulingUnitBlueprint.objects.all()


class TaskDraftViewSet(LOFARViewSet):
    queryset = models.TaskDraft.objects.all()
    serializer_class = serializers.TaskDraftSerializer

    @swagger_auto_schema(responses={201: 'The created task blueprint, see Location in Response header',
                                    403: 'forbidden'},
                         operation_description="Carve this draft task specification in stone, and make an (uneditable) blueprint out of it.")
    @action(methods=['get'], detail=True, url_name="create_task_blueprint", name="Create TaskBlueprint")
    def create_task_blueprint(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        task_blueprint = create_task_blueprint_from_task_draft(task_draft)

        # url path magic to construct the new task_blueprint_path url
        task_draft_path = request._request.path
        base_path = task_draft_path[:task_draft_path.find('/task_draft')]
        task_blueprint_path = '%s/task_blueprint/%s/' % (base_path, task_blueprint.id,)

        # return a response with the new serialized TaskBlueprint, and a Location to the new instance in the header
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED,
                        headers={'Location': task_blueprint_path})

    @swagger_auto_schema(responses={201: "This TaskBlueprint, with its created (and some scheduled) subtasks",
                                    403: 'forbidden'},
                         operation_description="Create subtasks, and schedule the ones that are not dependend on predecessors.")
    @action(methods=['get'], detail=True, url_name="create_task_blueprint_subtasks_and_schedule", name="Create TaskBlueprint, its Subtask(s) and Schedule")
    def create_task_blueprint_subtasks_and_schedule(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        task_blueprint = create_task_blueprint_and_subtasks_and_schedule_subtasks_from_task_draft(task_draft)

        # url path magic to construct the new task_blueprint_path url
        task_draft_path = request._request.path
        base_path = task_draft_path[:task_draft_path.find('/task_draft')]
        task_blueprint_path = '%s/task_blueprint/%s/' % (base_path, task_blueprint.id,)

        # return a response with the new serialized TaskBlueprint, and a Location to the new instance in the header
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED,
                        headers={'Location': task_blueprint_path})


    @swagger_auto_schema(responses={201: "This TaskBlueprint, with its created subtask(s)",
                                    403: 'forbidden'},
                         operation_description="Create subtasks.")
    @action(methods=['get'], detail=True, url_name="create_task_blueprint_subtasks", name="Create TaskBlueprint and its Subtask(s)")
    def create_task_blueprint_and_subtasks(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        task_blueprint = create_task_blueprint_and_subtasks_from_task_draft(task_draft)

        # url path magic to construct the new task_blueprint_path url
        task_draft_path = request._request.path
        base_path = task_draft_path[:task_draft_path.find('/task_draft')]
        task_blueprint_path = '%s/task_blueprint/%s/' % (base_path, task_blueprint.id,)

        # return a response with the new serialized TaskBlueprint, and a Location to the new instance in the header
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED,
                        headers={'Location': task_blueprint_path})


    @swagger_auto_schema(responses={200: 'The predecessor task draft of this task draft',
                                    403: 'forbidden'},
                         operation_description="Get the predecessor task draft of this task draft.")
    @action(methods=['get'], detail=True, url_name="predecessors")
    def predecessors(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        predecessors = self.filter_queryset(task_draft.predecessors)
        serializer = self.get_serializer(predecessors, many=True)
        return Response(serializer.data)

    @swagger_auto_schema(responses={200: 'The successor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the successor subtasks of this subtask.")
    @action(methods=['get'], detail=True, url_name="successors")
    def successors(self, request, pk=None):
        task_draft = get_object_or_404(models.TaskDraft, pk=pk)
        successors = self.filter_queryset(task_draft.successors)
        serializer = self.get_serializer(successors, many=True)
        return Response(serializer.data)


class TaskDraftNestedViewSet(LOFARNestedViewSet):
     queryset = models.TaskDraft.objects.all()
     serializer_class = serializers.TaskDraftSerializer

     def get_queryset(self):
         if 'scheduling_unit_draft_id' in self.kwargs:
             scheduling_unit_draft = get_object_or_404(models.SchedulingUnitDraft, pk=self.kwargs['scheduling_unit_draft_id'])
             return scheduling_unit_draft.task_drafts.all()
         else:
             return models.TaskDraft.objects.all()


class TaskBlueprintViewSet(LOFARViewSet):
    queryset = models.TaskBlueprint.objects.all()
    serializer_class = serializers.TaskBlueprintSerializer

    @swagger_auto_schema(responses={201: "This TaskBlueprint, with it is created subtasks",
                                    403: 'forbidden'},
                         operation_description="Create subtasks.")
    @action(methods=['get'], detail=True, url_name="create_subtasks", name="Create Subtasks")
    def create_subtasks(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        subtasks = create_subtasks_from_task_blueprint(task_blueprint)
        task_blueprint.refresh_from_db()

        # return a response with the new serialized task_blueprint (with references to the created subtasks)
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={201: "This TaskBlueprint, with it's created (and some scheduled) subtasks",
                                    403: 'forbidden'},
                         operation_description="Create subtasks, and schedule the ones that are not dependend on predecessors.")
    @action(methods=['get'], detail=True, url_name="create_subtasks_and_schedule", name="Create Subtasks and Schedule")
    def create_subtasks_and_schedule(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        subtasks = create_and_schedule_subtasks_from_task_blueprint(task_blueprint)
        task_blueprint.refresh_from_db()

        # return a response with the new serialized task_blueprint (with references to the created (and scheduled) subtasks)
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={201: "This TaskBlueprint, with the scheduled subtasks",
                                    403: 'forbidden'},
                         operation_description="Schedule the Subtasks that are not dependend on predecessors.")
    @action(methods=['get'], detail=True, url_name="schedule_independent_subtasks", name="Schedule independend Subtasks")
    def schedule_independent_subtasks(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        schedule_independent_subtasks_in_task_blueprint(task_blueprint)
        task_blueprint.refresh_from_db()

        # return a response with the new serialized task_blueprint (with references to the created (and scheduled) subtasks)
        return Response(serializers.TaskBlueprintSerializer(task_blueprint, context={'request':request}).data,
                        status=status.HTTP_201_CREATED)

    @swagger_auto_schema(responses={200: 'The predecessor task draft of this task draft',
                                    403: 'forbidden'},
                         operation_description="Get the predecessor task draft of this task draft.")
    @action(methods=['get'], detail=True, url_name="predecessors")
    def predecessors(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        predecessors = self.filter_queryset(task_blueprint.predecessors)
        serializer = self.get_serializer(predecessors, many=True)
        return Response(serializer.data)

    @swagger_auto_schema(responses={200: 'The successor subtasks of this subtask',
                                    403: 'forbidden'},
                         operation_description="Get the successor subtasks of this subtask.")
    @action(methods=['get'], detail=True, url_name="successors")
    def successors(self, request, pk=None):
        task_blueprint = get_object_or_404(models.TaskBlueprint, pk=pk)
        successors = self.filter_queryset(task_blueprint.successors)
        serializer = self.get_serializer(successors, many=True)
        return Response(serializer.data)


class TaskBlueprintNestedViewSet(LOFARNestedViewSet):
    queryset = models.TaskBlueprint.objects.all()
    serializer_class = serializers.TaskBlueprintSerializer

    def get_queryset(self):
        if 'task_draft_id' in self.kwargs:
            task_draft = get_object_or_404(models.TaskDraft, pk=self.kwargs['task_draft_id'])
            return task_draft.task_blueprints.all()
        else:
            return models.TaskBlueprint.objects.all()


class TaskRelationDraftViewSet(LOFARViewSet):
    queryset = models.TaskRelationDraft.objects.all()
    serializer_class = serializers.TaskRelationDraftSerializer


class TaskRelationDraftNestedViewSet(LOFARNestedViewSet):
    queryset = models.TaskRelationDraft.objects.all()
    serializer_class = serializers.TaskRelationDraftSerializer

    def get_queryset(self):
        if 'task_draft_id' in self.kwargs:
            task_draft = get_object_or_404(models.TaskDraft, pk=self.kwargs['task_draft_id'])
            return task_draft.produced_by.all() | task_draft.consumed_by.all()
        else:
            return models.TaskRelationDraft.objects.all()


class TaskRelationBlueprintViewSet(LOFARViewSet):
    queryset = models.TaskRelationBlueprint.objects.all()
    serializer_class = serializers.TaskRelationBlueprintSerializer


class TaskSchedulingRelationBlueprintViewSet(LOFARViewSet):
    queryset = models.TaskSchedulingRelationBlueprint.objects.all()
    serializer_class = serializers.TaskSchedulingRelationBlueprintSerializer


class TaskSchedulingRelationDraftViewSet(LOFARViewSet):
    queryset = models.TaskSchedulingRelationDraft.objects.all()
    serializer_class = serializers.TaskSchedulingRelationDraftSerializer


class TaskRelationBlueprintNestedViewSet(LOFARNestedViewSet):
    queryset = models.TaskRelationBlueprint.objects.all()
    serializer_class = serializers.TaskRelationBlueprintSerializer


    def get_queryset(self):
        if 'task_blueprint_id' in self.kwargs:
            task_blueprint = get_object_or_404(models.TaskBlueprint, pk=self.kwargs['task_blueprint_id'])
            return task_blueprint.produced_by.all() | task_blueprint.consumed_by.all()
        elif 'task_relation_draft_id' in self.kwargs:
            task_relation_draft = get_object_or_404(models.TaskRelationDraft, pk=self.kwargs['task_relation_draft_id'])
            return task_relation_draft.related_task_relation_blueprint.all()
        else:
            return models.TaskRelationBlueprint.objects.all()
