#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

from lofar.common.test_utils import skip_integration_tests
if skip_integration_tests():
    exit(3)

# before we import any django modules the DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS need to be known/set.
# import and start an isolated RATestEnvironment and TMSSTestEnvironment (with fresh database and attached django and ldap server on free ports)
# this automagically sets the required  DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS envvars.
from lofar.sas.resourceassignment.resourceassigner.test.ra_test_environment import RATestEnvironment
from lofar.sas.tmss.test.test_utils import TMSSTestEnvironment

ra_test_env = RATestEnvironment()
tmss_test_env = TMSSTestEnvironment()
try:
    ra_test_env.start()
    tmss_test_env.start()
except:
    ra_test_env.stop()
    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()
    ra_test_env.stop()


from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
test_data_creator = TMSSRESTTestDataCreator(tmss_test_env.django_server.url, (tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password))
from datetime import datetime, timedelta
from lofar.sas.resourceassignment.resourceassigner.rarpc import RARPC

class SchedulingTest(unittest.TestCase):
    def setUp(self):
        # clean all specs/tasks/claims in RADB (cascading delete)
        for spec in ra_test_env.radb.getSpecifications():
            ra_test_env.radb.deleteSpecification(spec['id'])

    def test_schedule_observation_subtask_with_enough_resources_available(self):
        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observationcontrol schema")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url=cluster_url)
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url']), '/subtask_output/')

            client.set_subtask_status(subtask_id, 'defined')
            subtask = client.schedule_subtask(subtask_id)

            self.assertEqual('scheduled', subtask['state_value'])
            self.assertEqual('scheduled', ra_test_env.radb.getTask(tmss_id=subtask_id)['status'])

    def test_schedule_observation_subtask_with_blocking_reservations(self):

        # create a reservation on station CS001
        with RARPC.create() as rarpc:
            ra_spec = { 'task_type': 'reservation',
                        'task_subtype': 'maintenance',
                        'status': 'prescheduled',
                        'starttime': datetime.utcnow()-timedelta(hours=1),
                        'endtime': datetime.utcnow() + timedelta(hours=1),
                        'cluster': None,
                        'specification': {} }
            inner_spec = { 'Observation.VirtualInstrument.stationList': ['CS001'],
                           'Observation.startTime': '2020-01-08 06:30:00',
                           'Observation.endTime': '2021-07-08 06:30:00' }
            ra_spec['specification'] = inner_spec
            assigned = rarpc.do_assignment(ra_spec)
            self.assertTrue(assigned)

        with tmss_test_env.create_tmss_client() as client:
            subtask_template = client.get_subtask_template("observationcontrol schema")
            spec = get_default_json_object_for_schema(subtask_template['schema'])
            spec['stations']['digital_pointings'][0]['subbands'] = [0]
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            subtask_data = test_data_creator.Subtask(specifications_template_url=subtask_template['url'],
                                                     specifications_doc=spec,
                                                     cluster_url=cluster_url)
            subtask = test_data_creator.post_data_and_get_response_as_json_object(subtask_data, '/subtask/')
            subtask_id = subtask['id']
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=subtask['url']), '/subtask_output/')

            client.set_subtask_status(subtask_id, 'defined')

            with self.assertRaises(Exception):
                client.schedule_subtask(subtask_id)

            subtask = client.get_subtask(subtask_id)
            self.assertEqual('error', subtask['state_value'])
            self.assertEqual('conflict', ra_test_env.radb.getTask(tmss_id=subtask_id)['status'])

    def test_schedule_pipeline_subtask_with_enough_resources_available(self):
        with tmss_test_env.create_tmss_client() as client:
            cluster_url = client.get_path_as_json_object('/cluster/1')['url']

            # setup: first create an observation, so the pipeline can have input.
            obs_subtask_template = client.get_subtask_template("observationcontrol schema")
            obs_spec = get_default_json_object_for_schema(obs_subtask_template['schema'])
            obs_spec['stations']['digital_pointings'][0]['subbands'] = [0]

            obs_subtask_data = test_data_creator.Subtask(specifications_template_url=obs_subtask_template['url'],
                                                         specifications_doc=obs_spec,
                                                         cluster_url=cluster_url)
            obs_subtask = test_data_creator.post_data_and_get_response_as_json_object(obs_subtask_data, '/subtask/')
            obs_subtask_output_url = test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=obs_subtask['url']), '/subtask_output/')
            test_data_creator.post_data_and_get_url(test_data_creator.Dataproduct(filename="L%s_SB000.MS"%obs_subtask['id'],
                                                                                  subtask_output_url=obs_subtask_output_url), '/dataproduct/')

            # now create the pipeline...
            pipe_subtask_template = client.get_subtask_template("pipelinecontrol schema")
            pipe_spec = get_default_json_object_for_schema(pipe_subtask_template['schema'])

            pipe_subtask_data = test_data_creator.Subtask(specifications_template_url=pipe_subtask_template['url'],
                                                          specifications_doc=pipe_spec,
                                                          task_blueprint_url=obs_subtask['task_blueprint'],
                                                          cluster_url=cluster_url)
            pipe_subtask = test_data_creator.post_data_and_get_response_as_json_object(pipe_subtask_data, '/subtask/')

            # ...and connect it to the observation
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskInput(subtask_url=pipe_subtask['url'], subtask_output_url=obs_subtask_output_url), '/subtask_input/')
            test_data_creator.post_data_and_get_url(test_data_creator.SubtaskOutput(subtask_url=pipe_subtask['url']), '/subtask_output/')

            for predecessor in client.get_subtask_predecessors(pipe_subtask['id']):
                client.set_subtask_status(predecessor['id'], 'finished')

            client.set_subtask_status(pipe_subtask['id'], 'defined')
            subtask = client.schedule_subtask(pipe_subtask['id'])

            self.assertEqual('scheduled', subtask['state_value'])
            self.assertEqual('scheduled', ra_test_env.radb.getTask(tmss_id=pipe_subtask['id'])['status'])


    def test_schedule_schedulingunit_enough_resources_available(self):
        '''similar test as test_schedule_pipeline_subtask_with_enough_resources_available, but now created from a scheduling_unit'''
        with tmss_test_env.create_tmss_client() as client:
            scheduling_unit_template = client.get_schedulingunit_template("scheduling unit schema")
            scheduling_unit_doc = get_default_json_object_for_schema(scheduling_unit_template['schema'])

            # define an observation without QA
            obs_task = get_default_json_object_for_schema(client.get_task_template(name="observation schema")['schema'])
            obs_task['QA']['plots']['enabled'] = False
            obs_task['QA']['file_conversion']['enabled'] = False
            obs_task['SAPs'][0]['subbands'] = [0,1]
            scheduling_unit_doc['tasks'].append({"name": "Observation",
                                                 "specifications_doc": obs_task,
                                                 "specifications_template": "observation schema"})

            # define a pipeline
            scheduling_unit_doc['tasks'].append({"name": "Pipeline",
                                                 "specifications_doc": get_default_json_object_for_schema(client.get_task_template(name="preprocessing schema")['schema']),
                                                 "specifications_template": "preprocessing schema"})

            # connect obs to pipeline
            scheduling_unit_doc['task_relations'].append({"producer": "Observation",
                                                          "consumer": "Pipeline",
                                                          "input": { "role": "input", "datatype": "visibilities" },
                                                          "output": { "role": "correlator", "datatype": "visibilities" },
                                                          "dataformat": "MeasurementSet",
                                                          "selection_doc": {},
                                                          "selection_template": "All" })

            # submit
            scheduling_unit_draft_data = test_data_creator.SchedulingUnitDraft(template_url=scheduling_unit_template['url'],
                                                                               requirements_doc=scheduling_unit_doc)
            scheduling_unit_draft = test_data_creator.post_data_and_get_response_as_json_object(scheduling_unit_draft_data, '/scheduling_unit_draft/')

            # create the whole blueprints tree...
            scheduling_unit_blueprint = client.create_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft['id'])

            # fetch the created task_blueprints
            task_blueprints = [client.get_url_as_json_object(task_blueprint_url) for task_blueprint_url in scheduling_unit_blueprint['task_blueprints']]
            self.assertEqual(2, len(task_blueprints))
            # and make sure they are ordered correctly
            if "Pipeline" in task_blueprints[0]['name']:
                task_blueprints.reverse()

            for task_blueprint in task_blueprints:
                self.assertEqual(1, len(task_blueprint['subtasks']))
                subtask = client.get_url_as_json_object(task_blueprint['subtasks'][0])

                client.set_subtask_status(subtask['id'], 'defined')
                subtask = client.schedule_subtask(subtask['id'])

                self.assertEqual('scheduled', subtask['state_value'])
                self.assertEqual('scheduled', ra_test_env.radb.getTask(tmss_id=subtask['id'])['status'])

                client.set_subtask_status(subtask['id'], 'finished')


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
