#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *


from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator

from lofar.sas.tmss.tmss.tmssapp import models

from lofar.sas.tmss.tmss.tmssapp.subtasks import *

# The following methods should be tested too

# schedule_subtask
# check_prerequities_for_scheduling
# schedule_qafile_subtask
# schedule_qaplots_subtask
# schedule_observation_subtask
# schedule_pipeline_subtask
# create_and_schedule_subtasks_from_task_blueprint





class SubTasksCreationFromSubTask(unittest.TestCase):

    @staticmethod
    def create_subtask_object(subtask_type_value, subtask_state_value):
        """
        Helper function to create a subtask object for testing with given subtask value and subtask state value
        as string (no object)
        """
        template_type = models.SubtaskType.objects.get(value=subtask_type_value)
        subtask_template_obj = SubTasksCreationFromSubTask.create_subtask_template(template_type)
        subtask_state_obj = models.SubtaskState.objects.get(value=subtask_state_value)
        subtask_data = Subtask_test_data(subtask_template=subtask_template_obj, state=subtask_state_obj)
        return models.Subtask.objects.create(**subtask_data)

    @staticmethod
    def create_subtask_template(template_type: object):
        subtask_template_data = SubtaskTemplate_test_data()
        subtask_template_data['type'] = template_type
        return models.SubtaskTemplate.objects.create(**subtask_template_data)

    def test_create_qafile_subtask_from_observation_subtask_failed(self):
        """
        Test if creation of subtask qafile failed due to wrong state or wrong type of the predecessor subtask
        Correct state should be 'defined' and correct type should be 'observation' (for this test of course it is not)
        """
        subtasks = [self.create_subtask_object("pipeline", "defined"),
                    self.create_subtask_object("observation", "defining"),
                    self.create_subtask_object("observation", "defining") ]
        for subtask in subtasks:
            with self.assertRaises(ValueError):
                create_qafile_subtask_from_observation_subtask(subtask)

    def test_create_qafile_subtask_from_observation_subtask_succeed(self):
        """
        Test if creation of subtask qafile succeed
        Check if the created subtask has correct subtask state and value (TODO)
        """
        predecessor_subtask = self.create_subtask_object("observation", "defined")
        subtask = create_qafile_subtask_from_observation_subtask(predecessor_subtask)
        # subtask object is None because QA file conversion is  by default not enabled!!!!
        self.assertEqual(None, subtask)

    def test_create_qaplots_subtask_from_qafile_subtask_failed(self):
        """
        Test if creation of subtask qaplots failed due to wrong state or wrong type of the predecessor subtask
        Correct type should be 'qa_files' (for this test of course it is not)
        """
        subtasks = [self.create_subtask_object("pipeline", "defined"),
                    self.create_subtask_object("observation", "defining"),
                    self.create_subtask_object("observation", "defining") ]
        for subtask in subtasks:
            with self.assertRaises(ValueError):
                create_qaplots_subtask_from_qafile_subtask(subtask)

    def test_create_qaplots_subtask_from_qafile_subtask_succeed(self):
        """
        Test if creation of subtask qaplots succeed
        Check if the created subtask has correct subtask state and value (TODO)
        """
        predecessor_subtask = self.create_subtask_object("qa_files", "defined")
        subtask = create_qaplots_subtask_from_qafile_subtask(predecessor_subtask)
        # subtask object is None because QA plots is by default not enabled!!!!
        self.assertEqual(None, subtask)


class SubTasksCreationFromTaskBluePrint(unittest.TestCase):

    @staticmethod
    def create_task_blueprint_object(task_template_name="observation schema", QA_enabled=False):
        """
        Helper function to create a task blueprint object for testing with given task template name value
        as string (no object)
        """
        task_template = TaskTemplate.objects.get(name=task_template_name)
        task_spec = get_default_json_object_for_schema(task_template.schema)
        if 'QA' in task_spec:
            task_spec["QA"]['plots']['enabled'] = QA_enabled
            task_spec["QA"]['file_conversion']['enabled'] = QA_enabled

        task_draft_data = TaskDraft_test_data(specifications_template=task_template, specifications_doc=task_spec)
        task_draft_obj = models.TaskDraft.objects.create(**task_draft_data)

        task_blueprint_data = TaskBlueprint_test_data(task_draft=task_draft_obj)
        task_blueprint_obj = models.TaskBlueprint.objects.create(**task_blueprint_data)
        return task_blueprint_obj

    def test_create_sequence_of_subtask_from_task_blueprint(self):
        """
        Create multiple subtasks from a task blueprint, executed in correct order.
        No exception should occur, check name, type and state of the subtask
        """
        task_blueprint = self.create_task_blueprint_object()

        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observationcontrol schema", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))

        # Next call requires an observation subtask already created
        subtask = create_qafile_subtask_from_task_blueprint(task_blueprint)
        # subtask object is None because QA file conversion is by default not enabled!!!!
        self.assertEqual(None, subtask)

        # Next call will fail due to no qa_files object
        # ValueError: Cannot create qa_plots subtask for task_blueprint id=1 because it has no qafile subtask(s)
        with self.assertRaises(SubtaskCreationException):
            subtask = create_qaplots_subtask_from_task_blueprint(task_blueprint)

        # subtask = create_preprocessing_subtask_from_task_blueprint(task_blueprint)

    def test_create_sequence_of_subtask_from_task_blueprint_with_QA_enabled(self):
        """
        Create multiple subtasks from a task blueprint, executed in correct order.
        QA plots and QA file conversion enabled
        No exception should occur, check name, type and state of the subtasks
        """
        # Enable QA plot and QA conversion
        task_blueprint = self.create_task_blueprint_object(QA_enabled=True)
        task_blueprint_preprocessing = self.create_task_blueprint_object("preprocessing schema")

        subtask = create_observation_control_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("observationcontrol schema", str(subtask.specifications_template.name))
        self.assertEqual("observation", str(subtask.specifications_template.type))
        # Next call requires an observation subtask already created
        subtask = create_qafile_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("QA file conversion", str(subtask.specifications_template.name))
        self.assertEqual("qa_files", str(subtask.specifications_template.type))
        # Next call requires an qaplots subtask already created
        subtask = create_qaplots_subtask_from_task_blueprint(task_blueprint)
        self.assertEqual("defined", str(subtask.state))
        self.assertEqual("QA plots", str(subtask.specifications_template.name))
        self.assertEqual("qa_plots", str(subtask.specifications_template.type))
        # TODO: check why next call failed?
        #subtask = create_preprocessing_subtask_from_task_blueprint(task_blueprint_preprocessing)
        #self.assertEqual("defined", str(subtask.state))


    def test_create_subtasks_from_task_blueprint_succeed(self):
        """
        """
        task_blueprint = self.create_task_blueprint_object(QA_enabled=True)
        subtasks = create_subtasks_from_task_blueprint(task_blueprint)
        self.assertEqual(3, len(subtasks))

# TODO Test the Schedule calls


class SubtaskInputSelectionFilteringTest(unittest.TestCase):

    # todo: merge in tests from TMSS-207 and deduplicate staticmethods

    def setUp(self) -> None:
        # make sure we're allowed to schedule
        setting = Setting.objects.get(name='allow_scheduling_observations')
        setting.value = True
        setting.save()

    @staticmethod
    def create_subtask_object(subtask_type_value, subtask_state_value):
        """
        Helper function to create a subtask object for testing with given subtask value and subtask state value
        as string (no object)
        """
        template_type = models.SubtaskType.objects.get(value=subtask_type_value)
        subtask_template_obj = SubtaskInputSelectionFilteringTest.create_subtask_template(template_type)
        subtask_state_obj = models.SubtaskState.objects.get(value=subtask_state_value)
        subtask_data = Subtask_test_data(subtask_template=subtask_template_obj, state=subtask_state_obj)
        return models.Subtask.objects.create(**subtask_data)

    @staticmethod
    def create_subtask_template(template_type: object):
        subtask_template_data = SubtaskTemplate_test_data()
        subtask_template_data['type'] = template_type
        return models.SubtaskTemplate.objects.create(**subtask_template_data)

    def test_specifications_doc_meets_selection_doc_returns_true_on_empty_filter(self):
        specs = {}
        selection = {}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

    def test_specifications_doc_meets_selection_doc_returns_true_when_filter_applies(self):
        # simple selection matches specs
        specs = {'sap': [0]}
        selection = {'sap': [0]}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

        # extra specs are ignored
        specs = {'sap': [0], 'not': 'relevant'}
        selection = {'sap': [0]}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

        # complex selection matches specs; multiple keys and values
        specs = {'sap': [0], 'is_relevant': True}
        selection = {'sap': [0, 1], 'is_relevant': True}
        self.assertTrue(specifications_doc_meets_selection_doc(specs, selection))

    def test_specifications_doc_meets_selection_doc_returns_true_when_filter_does_not_apply(self):
        # selection mismatches specs
        specs = {'sap': [0]}
        selection = {'sap': [1]}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

        # spec only partially selected
        specs = {'sap': [0, 1]}
        selection = {'sap': [1]}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

        # selection not in specs
        specs = {'sap': [0]}
        selection = {'sap': [0], 'is_relevant': True}
        self.assertFalse(specifications_doc_meets_selection_doc(specs, selection))

    # TODO: move to t_scheduling, cause during scheduling the output dataproducts are assigned
    # def test_schedule_pipeline_subtask_filters_predecessor_output_dataproducts_for_input(self):
    #
    #     # setup:
    #     #   create observation subtask and outputs and dataproducts
    #     obs_st = self.create_subtask_object('observation', 'finished')
    #     obs_out1 = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))
    #     obs_out2 = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=obs_st))
    #
    #     #   create connected pipeline subtask and inputs, specify input filtering
    #     pipe_st = self.create_subtask_object('pipeline', 'defined')
    #     pipe_out = models.SubtaskOutput.objects.create(**SubtaskOutput_test_data(subtask=pipe_st)) # required by scheduling function
    #     pipe_in1 = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out1, selection_doc={'sap': [0]}))
    #     pipe_in2 = models.SubtaskInput.objects.create(**SubtaskInput_test_data(subtask=pipe_st, producer=obs_out2, selection_doc={'sap': [1]}))
    #
    #     #   create obs output dataproducts with specs we can filter on
    #     dp1_1 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_doc={'sap': [0]}))
    #     dp1_2 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_doc={'sap': [1]}))
    #     dp1_3 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out1, specifications_doc={'sap': [0]}))
    #
    #     dp2_1 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out2, specifications_doc={'sap': [0]}))
    #     dp2_2 = models.Dataproduct.objects.create(**Dataproduct_test_data(producer=obs_out2, specifications_doc={'sap': [1]}))
    #
    #     # trigger:
    #     #   schedule pipeline, which should attach the correct subset of dataproducts to the pipeline inputs
    #     schedule_pipeline_subtask(pipe_st)
    #
    #     # assert:
    #     #   check correct input filtering
    #     self.assertEqual(set(pipe_in1.dataproducts.all()), {dp1_1, dp1_3})
    #     self.assertEqual(set(pipe_in2.dataproducts.all()), {dp2_2})


class SettingTest(unittest.TestCase):

    def test_schedule_observation_subtask_raises_when_flag_is_false(self):
        setting = Setting.objects.get(name='allow_scheduling_observations')
        setting.value = False
        setting.save()
        obs_st = SubtaskInputSelectionFilteringTest.create_subtask_object('observation', 'defined')

        with self.assertRaises(SubtaskSchedulingException):
            schedule_observation_subtask(obs_st)

if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
