#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
import requests

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

# before we import any django modules the DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS need to be known/set.
# import and start an isolated RATestEnvironment and TMSSTestEnvironment (with fresh database and attached django and ldap server on free ports)
# this automagically sets the required  DJANGO_SETTINGS_MODULE, TMSS_LDAPCREDENTIALS and TMSS_DBCREDENTIALS envvars.
from lofar.sas.resourceassignment.resourceassigner.test.ra_test_environment import RATestEnvironment
from lofar.sas.tmss.test.test_utils import TMSSTestEnvironment

ra_test_env = RATestEnvironment()
tmss_test_env = TMSSTestEnvironment()
try:
    ra_test_env.start()
    tmss_test_env.start()
except:
    ra_test_env.stop()
    tmss_test_env.stop()
    exit(1)

# tell unittest to stop (and automagically cleanup) the test database once all testing is done.
def tearDownModule():
    tmss_test_env.stop()
    ra_test_env.stop()

from lofar.sas.tmss.test.tmss_test_data_django_models import *

# import and setup rest test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator
rest_data_creator = TMSSRESTTestDataCreator(tmss_test_env.django_server.url, (tmss_test_env.ldap_server.dbcreds.user, tmss_test_env.ldap_server.dbcreds.password))

from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp.tasks import *


class CreationFromSchedulingUnitDraft(unittest.TestCase):
    """
    From scheduling_unit_draft should test:
    1. create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    6. create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> [TaskDraft]:
    3. create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft: models.SchedulingUnitDraft) -> models.SchedulingUnitBlueprint:
    """
    @staticmethod
    def create_scheduling_unit_draft_object(scheduling_unit_draft_name, requirements_doc=None):
        """
        Helper function to create a scheduling unit object for testing
        """
        scheduling_unit_draft_data = SchedulingUnitDraft_test_data(name=scheduling_unit_draft_name,
                                                                   requirements_doc=requirements_doc,
                                                                   template=models.SchedulingUnitTemplate.objects.get(name="scheduling unit schema"))
        draft_obj = models.SchedulingUnitDraft.objects.create(**scheduling_unit_draft_data)
        return draft_obj

    def test_create_scheduling_unit_blueprint_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if NO tasks are created
        """
        scheduling_unit_draft = self.create_scheduling_unit_draft_object("Test Scheduling Unit 1")

        scheduling_unit_blueprint = create_scheduling_unit_blueprint_from_scheduling_unit_draft(scheduling_unit_draft)
        self.assertEqual(scheduling_unit_draft.name, scheduling_unit_blueprint.draft.name)
        self.assertEqual(0, len(scheduling_unit_blueprint.task_blueprints.all()))

    def test_create_task_drafts_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft (with empty requirements_doc)
        Check if NO tasks are created
        Check with REST-call if NO tasks are created
        """
        scheduling_unit_draft = self.create_scheduling_unit_draft_object("Test Scheduling Unit 2", requirements_doc={'tasks': []})
        with self.assertRaises(BlueprintCreationException):
            create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft.refresh_from_db()
        task_drafts = scheduling_unit_draft.task_drafts.all()
        self.assertEqual(0, len(task_drafts))
        self.assertEqual(0, len(task_drafts))

    def test_create_task_drafts_from_scheduling_unit_draft_with_UC1_requirements(self):
        """
        Create Scheduling Unit Draft with requirements_doc (read from file)
        Check if tasks (7) are created
        Check with REST-call if tasks are created
        """
        working_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(working_dir, "testdata/example_UC1_scheduling_unit.json")) as json_file:
            json_requirements_doc = json.loads(json_file.read())

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   requirements_doc=json_requirements_doc,
                                   requirements_template=models.SchedulingUnitTemplate.objects.get(name="scheduling unit schema"),
                                   copy_reason=models.CopyReason.objects.get(value='template'),
                                   generator_instance_doc="para",
                                   copies=None,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        create_task_drafts_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft.refresh_from_db()
        task_drafts = scheduling_unit_draft.task_drafts.all()
        self.assertEqual(7, len(task_drafts))


    def test_create_task_blueprints_and_subtasks_subtasks_from_scheduling_unit_draft_with_UC1_requirements(self):
        """
        Create Scheduling Unit Draft with requirements_doc (read from file)
        Check if tasks (7) are created
        Check with REST-call if tasks are created
        """
        working_dir = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(working_dir, "testdata/example_UC1_scheduling_unit.json")) as json_file:
            json_requirements_doc = json.loads(json_file.read())

        scheduling_unit_draft = models.SchedulingUnitDraft.objects.create(
                                   name="Test Scheduling Unit UC1",
                                   requirements_doc=json_requirements_doc,
                                   requirements_template=models.SchedulingUnitTemplate.objects.get(name="scheduling unit schema"),
                                   copy_reason=models.CopyReason.objects.get(value='template'),
                                   generator_instance_doc="para",
                                   copies=None,
                                   scheduling_set=models.SchedulingSet.objects.create(**SchedulingSet_test_data()))

        create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        scheduling_unit_draft.refresh_from_db()
        task_drafts = scheduling_unit_draft.task_drafts.all()
        self.assertEqual(7, len(task_drafts))

        scheduling_unit_blueprints = scheduling_unit_draft.scheduling_unit_blueprints.all()
        self.assertEqual(1, len(scheduling_unit_blueprints))

        scheduling_unit_blueprint = scheduling_unit_blueprints[0]
        task_blueprints = scheduling_unit_blueprint.task_blueprints.all()
        self.assertEqual(7, len(task_blueprints))

    def test_create_task_blueprints_and_subtasks_subtasks_from_scheduling_unit_draft(self):
        """
        Create Scheduling Unit Draft
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if NO tasks are created
        """
        scheduling_unit_draft = self.create_scheduling_unit_draft_object("Test Scheduling Unit 3", {'tasks': []})

        with self.assertRaises(BlueprintCreationException):
            create_task_blueprints_and_subtasks_from_scheduling_unit_draft(scheduling_unit_draft)

        self.assertEqual(0, len(scheduling_unit_draft.scheduling_unit_blueprints.all()))


class CreationFromSchedulingUnitBluePrint(unittest.TestCase):
    """
    From scheduling_unit_blueprint should test:
    5. create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint: models.SchedulingUnitBlueprint) -> models.SchedulingUnitBlueprint:
    """

    def test_create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(self):
        """
        Create Scheduling Unit BluePrint
        Check with REST-call if NO tasks are created
        """
        scheduling_unit_blueprint_data = SchedulingUnitBlueprint_test_data(name="Test Scheduling Unit BluePrint")
        scheduling_unit_blueprint = models.SchedulingUnitBlueprint.objects.create(**scheduling_unit_blueprint_data)

        with self.assertRaises(BlueprintCreationException):
            scheduling_unit_blueprint = create_task_blueprints_and_subtasks_from_scheduling_unit_blueprint(scheduling_unit_blueprint)

        self.assertEqual(0, scheduling_unit_blueprint.task_blueprints.count())


class CreationFromTaskDraft(unittest.TestCase):
    """
    From task draft should test:
     2. create_task_blueprint_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
     5. create_task_blueprint_and_subtasks_from_task_draft(task_draft: models.TaskDraft) -> models.TaskBlueprint:
    """
    @staticmethod
    def create_task_object(task_draft_name):
        """
        Helper function to create a task object for testing
        TODO change schema to observation schema, correlator schema wil be removed
             using the observation schema results in  jsonschema.exceptions.ValidationError: 'stations' is a required property
             so somehow its does not fill in the required fields ??
        """
        obs_task_template = models.TaskTemplate.objects.get(name='observation schema')
        task_draft_data = TaskDraft_test_data(name=task_draft_name, specifications_template=obs_task_template)
        models.TaskDraft.objects.create(**task_draft_data)


    def test_create_task_blueprint_and_subtasks(self):
        """
        Create task draft
        Check if the name draft (specified) is equal to name blueprint (created)
        Check with REST-call if 3 subtasks are created and if these subtasks have state value 'defined'
        """
        self.create_task_object("Test Target Observation 1")

        task_draft = models.TaskDraft.objects.get(name="Test Target Observation 1")
        task_blueprint = create_task_blueprint_and_subtasks_from_task_draft(task_draft)
        self.assertEqual(task_draft.name, task_blueprint.draft.name)
        self.assertEqual(3, task_blueprint.subtasks.count())
        for subtask in task_blueprint.subtasks.all():
            subtask.state.value == 'defined'

    def test_create_task_blueprint(self):
        """
        Create task draft
        Check if the task draft name is equal to the task draft name specified in the created task blueprint
        Check with REST-call if NO subtasks are created
        """
        self.create_task_object("Test Target Observation 2")

        task_draft = models.TaskDraft.objects.get(name="Test Target Observation 2")
        task_blueprint = create_task_blueprint_from_task_draft(task_draft)
        self.assertEqual(task_draft.name, task_blueprint.draft.name)
        self.assertEqual(0, task_blueprint.subtasks.count())


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
