#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import unittest
from datetime import datetime
import uuid

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

# todo: Tags? -> Decide how to deal with them first.
# todo: Immutability of Blueprints on db level?

# Do Mandatory setup:
# use setup/teardown magic for tmss test database
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_database_unittest_setup module)
from lofar.sas.tmss.test.tmss_database_unittest_setup import *

from lofar.sas.tmss.test.tmss_test_data_django_models import *

from django.db.utils import IntegrityError
from django.core.exceptions import ValidationError


class GeneratorTemplateTest(unittest.TestCase):
    def test_GeneratorTemplate_gets_created_with_correct_creation_timestamp(self):
        # setup
        before = datetime.utcnow()
        entry = models.GeneratorTemplate.objects.create(**GeneratorTemplate_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_GeneratorTemplate_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.GeneratorTemplate.objects.create(**GeneratorTemplate_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)


class DefaultGeneratorTemplateTest(unittest.TestCase):
    def test_DefaultGeneratorTemplate_prevents_same_name(self):
        common_forbidden_name = "my_name"
        template = models.GeneratorTemplate.objects.create(**GeneratorTemplate_test_data())

        test_data_1 = DefaultGeneratorTemplate_test_data(common_forbidden_name, template)
        models.DefaultGeneratorTemplate.objects.create(**test_data_1)

        test_data_2 = DefaultGeneratorTemplate_test_data(common_forbidden_name, template)
        with self.assertRaises(IntegrityError):
            models.DefaultGeneratorTemplate.objects.create(**test_data_2)


class SchedulingUnitTemplateTest(unittest.TestCase):
    def test_SchedulingUnitTemplate_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.SchedulingUnitTemplate.objects.create(**SchedulingUnitTemplate_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_SchedulingUnitTemplate_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.SchedulingUnitTemplate.objects.create(**SchedulingUnitTemplate_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)


class TaskTemplateTest(unittest.TestCase):
    def test_TaskTemplate_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.TaskTemplate.objects.create(**TaskTemplate_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_TaskTemplate_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.TaskTemplate.objects.create(**TaskTemplate_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)

    def test_TaskTemplate_name_version_unique(self):
        test_data = TaskTemplate_test_data(name="my_name", version="1")
        entry1 = models.TaskTemplate.objects.create(**test_data)

        with self.assertRaises(IntegrityError):
            entry2 = models.TaskTemplate.objects.create(**test_data)

        test_data2 = dict(**test_data)
        test_data2['version'] = "2"
        entry2 = models.TaskTemplate.objects.create(**test_data2)

        with self.assertRaises(IntegrityError):
            entry2.version = '1'
            entry2.save()


class TaskRelationSelectionTemplateTest(unittest.TestCase):
    def test_TaskRelationSelectionTemplate_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.TaskRelationSelectionTemplate.objects.create(**TaskRelationSelectionTemplate_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_TaskRelationSelectionTemplate_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.TaskRelationSelectionTemplate.objects.create(**TaskRelationSelectionTemplate_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)


class TaskConnectorTest(unittest.TestCase):

    def test_POST_TaskConnector_prevents_missing_input_of(self):

        # setup
        test_data_1 = dict(TaskConnectorType_test_data())
        test_data_1['input_of'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskConnectorType.objects.create(**test_data_1)

    def test_POST_TaskConnector_prevents_missing_output_of(self):

        # setup
        test_data_1 = dict(TaskConnectorType_test_data())
        test_data_1['output_of'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskConnectorType.objects.create(**test_data_1)


class CycleTest(unittest.TestCase):

    def test_Cycle_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.Cycle.objects.create(**Cycle_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_Cycle_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.Cycle.objects.create(**Cycle_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)


class ProjectTest(unittest.TestCase):

    def test_Project_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.Project.objects.create(**Project_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_Project_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.Project.objects.create(**Project_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)


class ProjectQuotaTest(unittest.TestCase):
    def test_ProjectQuota_prevents_missing_project(self):
        # setup
        test_data = dict(ProjectQuota_test_data())
        test_data['project'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.ProjectQuota.objects.create(**test_data)


class SchedulingSetTest(unittest.TestCase):

    def test_SchedulingSet_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.SchedulingSet.objects.create(**SchedulingSet_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_SchedulingSet_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.SchedulingSet.objects.create(**SchedulingSet_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)

    def test_SchedulingSet_prevents_missing_project(self):

        # setup
        test_data = dict(SchedulingSet_test_data())
        test_data['project'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.SchedulingSet.objects.create(**test_data)


class SchedulingUnitDraftTest(unittest.TestCase):

    def test_SchedulingUnitDraft_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.SchedulingUnitDraft.objects.create(**SchedulingUnitDraft_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_SchedulingUnitDraft_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.SchedulingUnitDraft.objects.create(**SchedulingUnitDraft_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)

    def test_SchedulingUnitDraft_prevents_missing_template(self):

        # setup
        test_data = dict(SchedulingUnitDraft_test_data())
        test_data['requirements_template'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.SchedulingUnitDraft.objects.create(**test_data)

    def test_SchedulingUnitDraft_prevents_missing_scheduling_set(self):

        # setup
        test_data = dict(SchedulingUnitDraft_test_data())
        test_data['scheduling_set'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.SchedulingUnitDraft.objects.create(**test_data)


class TaskDraftTest(unittest.TestCase):

    def test_TaskDraft_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.TaskDraft.objects.create(**TaskDraft_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_TaskDraft_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.TaskDraft.objects.create(**TaskDraft_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)

    def test_TaskDraft_prevents_missing_template(self):

        # setup
        test_data = dict(TaskDraft_test_data())
        test_data['specifications_template'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskDraft.objects.create(**test_data)

    def test_TaskDraft_prevents_missing_scheduling_unit_draft(self):

        # setup
        test_data = dict(TaskDraft_test_data())
        test_data['scheduling_unit_draft'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskDraft.objects.create(**test_data)

    def test_TaskDraft_predecessors_and_successors_none(self):
        task_draft_1:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())
        task_draft_2:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())

        self.assertEqual(set(), set(task_draft_1.predecessors.all()))
        self.assertEqual(set(), set(task_draft_2.predecessors.all()))
        self.assertEqual(set(), set(task_draft_1.successors.all()))
        self.assertEqual(set(), set(task_draft_2.successors.all()))

    def test_TaskDraft_predecessors_and_successors_simple(self):
        task_draft_1:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())
        task_draft_2:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())

        models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data(producer=task_draft_1,
                                                                              consumer=task_draft_2))

        self.assertEqual(task_draft_1, task_draft_2.predecessors.all()[0])
        self.assertEqual(task_draft_2, task_draft_1.successors.all()[0])

    def test_TaskDraft_predecessors_and_successors_complex(self):
        task_draft_1:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())
        task_draft_2:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())
        task_draft_3:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())
        task_draft_4:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())
        task_draft_5:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())
        task_draft_6:models.TaskDraft = models.TaskDraft.objects.create(**TaskDraft_test_data())

        # ST1 ---> ST3 ---> ST4
        #      |        |
        # ST2 -          -> ST5 ---> ST6

        models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data(producer=task_draft_1,
                                                                              consumer=task_draft_3))
        models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data(producer=task_draft_2,
                                                                              consumer=task_draft_3))
        models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data(producer=task_draft_3,
                                                                              consumer=task_draft_4))
        models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data(producer=task_draft_3,
                                                                              consumer=task_draft_5))
        models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data(producer=task_draft_5,
                                                                              consumer=task_draft_6))

        self.assertEqual(set((task_draft_1, task_draft_2)), set(task_draft_3.predecessors.all()))
        self.assertEqual(set((task_draft_4, task_draft_5)), set(task_draft_3.successors.all()))
        self.assertEqual(set((task_draft_3,)), set(task_draft_4.predecessors.all()))
        self.assertEqual(set((task_draft_3,)), set(task_draft_5.predecessors.all()))
        self.assertEqual(set((task_draft_3,)), set(task_draft_1.successors.all()))
        self.assertEqual(set((task_draft_3,)), set(task_draft_2.successors.all()))
        self.assertEqual(set(), set(task_draft_1.predecessors.all()))
        self.assertEqual(set(), set(task_draft_2.predecessors.all()))
        self.assertEqual(set(), set(task_draft_4.successors.all()))
        self.assertEqual(set((task_draft_6,)), set(task_draft_5.successors.all()))

class TaskRelationDraftTest(unittest.TestCase):

    def test_TaskRelationDraft_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_TaskRelationDraft_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.TaskRelationDraft.objects.create(**TaskRelationDraft_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)


    def test_TaskRelationDraft_prevents_missing_template(self):

        # setup
        test_data = dict(TaskRelationDraft_test_data())
        test_data['selection_template'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationDraft.objects.create(**test_data)

    def test_TaskRelationDraft_prevents_missing_consumer(self):

        # setup
        test_data = dict(TaskRelationDraft_test_data())
        test_data['consumer'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationDraft.objects.create(**test_data)

    def test_TaskRelationDraft_prevents_missing_producer(self):

        # setup
        test_data = dict(TaskRelationDraft_test_data())
        test_data['producer'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationDraft.objects.create(**test_data)


class SchedulingUnitBlueprintTest(unittest.TestCase):

    def test_SchedulingUnitBlueprint_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.SchedulingUnitBlueprint.objects.create(**SchedulingUnitBlueprint_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_SchedulingUnitBlueprint_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.SchedulingUnitBlueprint.objects.create(**SchedulingUnitBlueprint_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)


    def test_SchedulingUnitBlueprint_prevents_missing_template(self):

        # setup
        test_data = dict(SchedulingUnitBlueprint_test_data())
        test_data['requirements_template'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.SchedulingUnitBlueprint.objects.create(**test_data)

    def test_SchedulingUnitBlueprint_prevents_missing_draft(self):

        # setup
        test_data = dict(SchedulingUnitBlueprint_test_data())
        test_data['draft'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.SchedulingUnitBlueprint.objects.create(**test_data)


class TaskBlueprintTest(unittest.TestCase):

    def test_TaskBlueprint_gets_created_with_correct_creation_timestamp(self):

        # setup
        before = datetime.utcnow()
        entry = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_TaskBlueprint_update_timestamp_gets_changed_correctly(self):

        # setup
        entry = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)

    def test_TaskBlueprint_prevents_missing_template(self):

        # setup
        test_data = dict(TaskBlueprint_test_data())
        test_data['specifications_template'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskBlueprint.objects.create(**test_data)

    def test_TaskBlueprint_prevents_missing_draft(self):

        # setup
        test_data = dict(TaskBlueprint_test_data())
        test_data['draft'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskBlueprint.objects.create(**test_data)

    def test_TaskBlueprint_prevents_missing_scheduling_unit_blueprint(self):

        # setup
        test_data = dict(TaskBlueprint_test_data())
        test_data['scheduling_unit_blueprint'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskBlueprint.objects.create(**test_data)

    def test_TaskBlueprint_predecessors_and_successors_none(self):
        task_blueprint_1: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        task_blueprint_2: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())

        self.assertEqual(set(), set(task_blueprint_1.predecessors.all()))
        self.assertEqual(set(), set(task_blueprint_2.predecessors.all()))
        self.assertEqual(set(), set(task_blueprint_1.successors.all()))
        self.assertEqual(set(), set(task_blueprint_2.successors.all()))

    def test_TaskBlueprint_predecessors_and_successors_simple(self):
        task_blueprint_1: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        task_blueprint_2: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())

        models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data(producer=task_blueprint_1,
                                                                                      consumer=task_blueprint_2))

        self.assertEqual(task_blueprint_1, task_blueprint_2.predecessors.all()[0])
        self.assertEqual(task_blueprint_2, task_blueprint_1.successors.all()[0])

    def test_TaskBlueprint_predecessors_and_successors_complex(self):
        task_blueprint_1: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        task_blueprint_2: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        task_blueprint_3: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        task_blueprint_4: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        task_blueprint_5: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())
        task_blueprint_6: models.TaskBlueprint = models.TaskBlueprint.objects.create(**TaskBlueprint_test_data())

        # ST1 ---> ST3 ---> ST4
        #      |        |
        # ST2 -          -> ST5 ---> ST6

        models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data(producer=task_blueprint_1,
                                                                                      consumer=task_blueprint_3))
        trb1 = models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data(producer=task_blueprint_2,
                                                                                      consumer=task_blueprint_3))
        models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data(producer=task_blueprint_3,
                                                                                      consumer=task_blueprint_4))
        models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data(producer=task_blueprint_3,
                                                                                      consumer=task_blueprint_5))
        models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data(producer=task_blueprint_5,
                                                                                      consumer=task_blueprint_6))

        self.assertEqual(set((task_blueprint_1, task_blueprint_2)), set(task_blueprint_3.predecessors.all()))
        self.assertEqual(set((task_blueprint_4, task_blueprint_5)), set(task_blueprint_3.successors.all()))
        self.assertEqual(set((task_blueprint_3,)), set(task_blueprint_4.predecessors.all()))
        self.assertEqual(set((task_blueprint_3,)), set(task_blueprint_5.predecessors.all()))
        self.assertEqual(set((task_blueprint_3,)), set(task_blueprint_1.successors.all()))
        self.assertEqual(set((task_blueprint_3,)), set(task_blueprint_2.successors.all()))
        self.assertEqual(set(), set(task_blueprint_1.predecessors.all()))
        self.assertEqual(set(), set(task_blueprint_2.predecessors.all()))
        self.assertEqual(set(), set(task_blueprint_4.successors.all()))
        self.assertEqual(set((task_blueprint_6,)), set(task_blueprint_5.successors.all()))


class TaskRelationBlueprintTest(unittest.TestCase):

    def test_TaskRelationBlueprint_gets_created_with_correct_creation_timestamp(self):
        # setup
        before = datetime.utcnow()
        entry = models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data())

        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.created_at)
        self.assertGreater(after, entry.created_at)

    def test_TaskRelationBlueprint_update_timestamp_gets_changed_correctly(self):
        # setup
        entry = models.TaskRelationBlueprint.objects.create(**TaskRelationBlueprint_test_data())
        before = datetime.utcnow()
        entry.save()
        after = datetime.utcnow()

        # assert
        self.assertLess(before, entry.updated_at)
        self.assertGreater(after, entry.updated_at)

    def test_TaskRelationBlueprint_prevents_missing_selection_template(self):
        # setup
        test_data = dict(TaskRelationBlueprint_test_data())
        test_data['selection_template'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationBlueprint.objects.create(**test_data)

    def test_TaskRelationBlueprint_prevents_missing_draft(self):
        # setup
        test_data = dict(TaskRelationBlueprint_test_data())
        test_data['draft'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationBlueprint.objects.create(**test_data)

    def test_TaskRelationBlueprint_prevents_missing_producer(self):
        # setup
        test_data = dict(TaskRelationBlueprint_test_data())
        test_data['producer'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationBlueprint.objects.create(**test_data)

    def test_TaskRelationBlueprint_prevents_missing_consumer(self):
        # setup
        test_data = dict(TaskRelationBlueprint_test_data())
        test_data['consumer'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationBlueprint.objects.create(**test_data)

    def test_TaskRelationBlueprint_prevents_missing_input(self):
        # setup
        test_data = dict(TaskRelationBlueprint_test_data())
        test_data['input_role'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationBlueprint.objects.create(**test_data)

    def test_TaskRelationBlueprint_prevents_missing_output(self):
        # setup
        test_data = dict(TaskRelationBlueprint_test_data())
        test_data['output_role'] = None

        # assert
        with self.assertRaises(IntegrityError):
            models.TaskRelationBlueprint.objects.create(**test_data)


if __name__ == "__main__":
    os.environ['TZ'] = 'UTC'
    unittest.main()
