#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import logging
logger = logging.getLogger(__name__)
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s', level=logging.INFO)

# Do Mandatory setup step:
# use setup/teardown magic for tmss test database, ldap server and django server
# (ignore pycharm unused import statement, python unittests does use at RunTime the tmss_test_environment_unittest_setup module)
from lofar.sas.tmss.test.tmss_test_environment_unittest_setup import *

# import and setup test data creator
from lofar.sas.tmss.test.tmss_test_data_rest import TMSSRESTTestDataCreator

from lofar.sas.tmss.tmss.tmssapp import models

from django.contrib.auth.models import User, Group, Permission
from datetime import datetime
import unittest
import requests

class CyclePermissionTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_data_creator = TMSSRESTTestDataCreator(BASE_URL, requests.auth.HTTPBasicAuth('paulus', 'pauluspass'))
        response = requests.get(cls.test_data_creator.django_api_url, auth=cls.test_data_creator.auth)

        cls.support_group = Group.objects.create(name='support')
        cls.support_group.permissions.add(Permission.objects.get(codename='add_cycle'))

        cls.admin_group = Group.objects.create(name='admin')
        cls.admin_group.permissions.add(Permission.objects.get(codename='delete_cycle'))

    def test_Cycle_cannot_be_added_without_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while user.has_perm('tmssapp.add_cycle'):
            user = User.objects.get(username='paulus')

        self.assertFalse(user.has_perm('tmssapp.add_cycle'))

        test_data = self.test_data_creator.Cycle()
        res = self.test_data_creator.post_data_and_get_response(test_data, '/cycle/')
        self.assertEqual(res.status_code, 403)

    def test_Cycle_can_be_added_by_support(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.support_group])

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.add_cycle'):
            user = User.objects.get(username='paulus')

        self.assertTrue(user.has_perm('tmssapp.add_cycle'))

        test_data = self.test_data_creator.Cycle()
        res = self.test_data_creator.post_data_and_get_response(test_data, '/cycle/')
        self.assertEqual(res.status_code, 201)

    def test_Cycle_cannot_be_deleted_without_group(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.support_group]) # can add, cannot delete

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.add_cycle'):
            user = User.objects.get(username='paulus')

        # add
        count = len(models.Cycle.objects.all())
        test_data = self.test_data_creator.Cycle()
        url = self.test_data_creator.post_data_and_get_url(test_data, '/cycle/')
        self.assertEqual(count+1, len(models.Cycle.objects.all()))

        # delete
        response = requests.delete(url, auth=self.test_data_creator.auth)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(count + 1, len(models.Cycle.objects.all()))

    def test_Cycle_can_be_deleted_by_admin(self):
        user = User.objects.get(username='paulus')
        user.groups.set([self.support_group, self.admin_group]) # can add and delete

        # refresh user to update cache, see: https://docs.djangoproject.com/en/3.0/topics/auth/default/#permission-caching
        user = User.objects.get(username='paulus')
        while not user.has_perm('tmssapp.add_cycle'):
            user = User.objects.get(username='paulus')

        # add
        count = len(models.Cycle.objects.all())
        test_data = self.test_data_creator.Cycle()
        url = self.test_data_creator.post_data_and_get_url(test_data, '/cycle/')
        self.assertEqual(count+1, len(models.Cycle.objects.all()))

        # delete
        response = requests.delete(url, auth=self.test_data_creator.auth)
        self.assertEqual(response.status_code, 204)
        self.assertEqual(count, len(models.Cycle.objects.all()))

if __name__ == "__main__":
    unittest.main()

