#!/usr/bin/env python3

# Copyright (C) 2018    ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

# $Id:  $

import os
import time
from multiprocessing import Process, Event
import django

import logging
logger = logging.getLogger(__name__)

from lofar.common.testing.postgres import PostgresTestMixin, PostgresTestDatabaseInstance
from lofar.common.dbcredentials import Credentials, DBCredentials
from lofar.common.util import find_free_port, waitForInterrupt
from lofar.sas.tmss.test.ldap_test_service import TestLDAPServer
from lofar.sas.tmss.tmss.exceptions import TMSSException
from lofar.messaging.config import DEFAULT_BROKER, DEFAULT_BUSNAME
from lofar.common.testing.dbcredentials import TemporaryCredentials
from lofar.sas.tmss.client.tmss_http_rest_client import TMSSsession
from lofar.sas.resourceassignment.resourceassigner.test.ra_test_environment import RATestEnvironment


def assertDataWithUrls(self, data, expected):
    """
    object instances get returned as urls, check that the value is part of that url
    """
    # TODO: Make this smarter, this only checks for matching pk!

    from django.db import models

    for k, v in expected.items():
        if isinstance(v, models.Model):
            v = str(v.pk)
            v = v.replace(' ', '%20')
            err_msg = "The value '%s' (key is %s) is not in expected %s" % (str(v), str(data[k]), k)
            self.assertTrue(str(v) in data[k], err_msg)

        else:
            self.assertEqual(v, data[k])

def assertUrlList(self, url_list, expected_objects):
    """
    object instances get returned as urls, check that the expected projects are in that list
    """

    # TODO: Make this smarter, this only checks for matching pk!

    from django.db import models
    self.assertEqual(len(url_list), len(expected_objects))
    for v in expected_objects:
        if isinstance(v, models.Model):
            v = str(v.pk)
            v = v.replace(' ', '%20')
            self.assertTrue(any(str(v) in myurl for myurl in url_list))
        else:
            raise ValueError('Expected item is not a Django model instance: %s' % v)


class TMSSTestDatabaseInstance(PostgresTestDatabaseInstance):
    '''
    Creates an isolated postgres database instance and initializes the database with a django tmss migration.
    Destroys the isolated postgres database instance upon exit automagically.
    '''
    def __init__(self) -> None:
        super().__init__(user='test_tmss_user')

    def apply_database_schema(self):
        logger.info('applying TMSS sql schema to %s', self.dbcreds)

        # a TMSSTestDatabaseInstance needs to run in a clean env,
        # with these variables set to the current test values.
        import os
        os.environ["TMSS_DBCREDENTIALS"] = self.dbcreds_id
        os.environ["DJANGO_SETTINGS_MODULE"] = "lofar.sas.tmss.tmss.settings"

        # run migrate in a seperate process so the needed django setup does not pollute our current apps environment
        def _migrate_helper():
            # use django management modules to apply database schema via initial migration
            import django
            django.setup()
            django.core.management.call_command('migrate')

        migrate_process = Process(target=_migrate_helper, daemon=True)
        migrate_process.start()
        migrate_process.join()

        if migrate_process.exitcode != 0:
            raise TMSSException("Could not initialize TMSS database with django migrations")


class TMSSPostgresTestMixin(PostgresTestMixin):
    '''
    A common test mixin class from which you can derive to get a freshly setup postgres testing instance with the latest TMSS sql schema.
    '''
    @classmethod
    def create_test_db_instance(cls) -> TMSSTestDatabaseInstance:
        return TMSSTestDatabaseInstance()


class TMSSDjangoServerInstance():
    ''' Creates a running django TMSS server at the requested port with the requested database credentials.
    '''
    def __init__(self, db_dbcreds_id: str="TMSS", ldap_dbcreds_id: str="TMSS_LDAP", host: str='127.0.0.1', port: int=8000,
                 exchange: str=os.environ.get("TMSS_EXCHANGE", DEFAULT_BUSNAME), broker: str=os.environ.get("TMSS_EXCHANGE", DEFAULT_BUSNAME)):
        self._db_dbcreds_id = db_dbcreds_id
        self._ldap_dbcreds_id = ldap_dbcreds_id
        self.host = host
        self.port = port
        self._server_process = None
        self._exchange = exchange
        self._broker = broker

    @property
    def address(self):
        ''':returns the address and port of the django server'''
        return "%s:%d" % (self.host, self.port)

    @property
    def url(self):
        ''':returns the http url to the django server'''
        return "http://%s/api" % self.address

    @property
    def oidc_url(self):
        ''':returns the http url to the django server'''
        return "http://%s/oidc" % self.address

    @property
    def database_dbcreds_id(self) -> str:
        ''':returns the uuid of the temporary database credentials'''
        return self._db_dbcreds_id

    @property
    def database_dbcreds(self) -> Credentials:
        ''':returns the temporary database Credentials'''
        return DBCredentials().get(self._db_dbcreds_id)

    @property
    def ldap_dbcreds_id(self) -> str:
        ''':returns the uuid of the temporary LDAP server credentials'''
        return self._ldap_dbcreds_id

    @property
    def ldap_dbcreds(self) -> Credentials:
        ''':returns the temporary LDAP Credentials'''
        return DBCredentials().get(self._ldap_dbcreds_id)

    def setup_django(self):
        # (tmss)django is initialized via many environment variables.
        # set these here, run django setup, and start the server
        os.environ["TMSS_LDAPCREDENTIALS"] = self.ldap_dbcreds_id
        os.environ["TMSS_DBCREDENTIALS"] = self.database_dbcreds_id
        os.environ["TMSS_EXCHANGE"] = self._exchange
        os.environ["TMSS_BROKER"] = self._broker
        os.environ["DJANGO_SETTINGS_MODULE"] = "lofar.sas.tmss.tmss.settings"
        django.setup()

    def start(self):
        '''
        Start the Django server with a test-LDAP server in the background.
        Best used in a 'with'-context
        '''
        def _helper_runserver_loop():
            logger.info("Starting Django server at port=%d with database: %s and LDAP: %s",
                        self.port, self.database_dbcreds, self.ldap_dbcreds)

            self.setup_django()
            django.core.management.call_command('runserver',
                                                use_reloader=False,
                                                addrport=self.address)

        self._server_process = Process(target=_helper_runserver_loop, daemon=True)
        self._server_process.start()

        # wait for server to be up and running....
        # or exit via TimeoutError
        self.check_running_server(timeout=10)

    def stop(self):
        '''
        Stop the running Django and LDAP servers.
        '''
        if self._server_process is not None:
            logger.info("Stopping Django server...")
            try:
                self._server_process.kill() # new in python 3.7
            except AttributeError:
                self._server_process.terminate() # < python 3.7

            self._server_process = None
            logger.info("Django server stopped.")

    def check_running_server(self, timeout: float = 10) -> bool:
        '''Check the running django server for a valid response'''
        try:
            import requests, urllib3
            from _datetime import datetime, timedelta
            start = datetime.utcnow()
            while True:
                try:
                    response = requests.get(self.url, auth=(self.ldap_dbcreds.user, self.ldap_dbcreds.password), timeout=max(1, timeout/10))

                    if response.status_code in [200, 401, 403]:
                        logger.info("TMSS Django server is up and running at %s with database: %s and LDAP: %s",
                                    self.url, self.database_dbcreds, self.ldap_dbcreds)

                        if response.status_code in [401, 403]:
                            logger.warning("TMSS Django server at %s could not autenticate with LDAP creds: %s", self.url, self.ldap_dbcreds)

                        # TODO: logout, otherwise django remembers our login session.
                        return True
                except Exception as e:
                    time.sleep(0.25)

                if datetime.utcnow() - start > timedelta(seconds=timeout):
                    raise TimeoutError("Could not get a valid response from the django server at %s" % self.url)
        except Exception as e:
            logger.error(e)
            return False

    def __enter__(self):
        try:
            self.start()
        except Exception as e:
            logger.error(e)
            self.stop()
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


class TMSSTestEnvironment:
    '''Create and run a test django TMSS server against a newly created test database and a test ldap server (and cleanup automagically)'''
    def __init__(self, host: str='127.0.0.1', preferred_django_port: int=8000,
                 exchange: str=os.environ.get("TMSS_EXCHANGE", DEFAULT_BUSNAME), broker: str=os.environ.get("TMSS_BROKER", DEFAULT_BROKER)):
        self.ldap_server = TestLDAPServer(user='test', password='test')
        self.database = TMSSTestDatabaseInstance()
        self.django_server = TMSSDjangoServerInstance(db_dbcreds_id=self.database.dbcreds_id,
                                                      ldap_dbcreds_id=self.ldap_server.dbcreds_id,
                                                      host=host,
                                                      port=find_free_port(preferred_django_port),
                                                      exchange=exchange,
                                                      broker=broker)
        self.client_credentials = TemporaryCredentials(user=self.ldap_server.dbcreds.user,
                                                       password=self.ldap_server.dbcreds.password)

    def start(self):
        self.ldap_server.start()
        self.database.create()
        self.django_server.start()

        # store client credentials in the TemporaryCredentials file...
        self.client_credentials.dbcreds.host = self.django_server.host
        self.client_credentials.dbcreds.port = self.django_server.port
        self.client_credentials.dbcreds.type = "http"
        self.client_credentials.create()
        # ... and set TMSS_CLIENT_DBCREDENTIALS environment variable, sp anybody or anything (any test) can use it automagically
        os.environ['TMSS_CLIENT_DBCREDENTIALS'] = self.client_credentials.dbcreds_id

        # apart from the running django server with a REST API,
        # it is also convenient to provide a working django setup for the 'normal' django API (via models.objects)
        # so: do setup_django
        self.django_server.setup_django()

        # now that the ldap and django server are running, and the django set has been done,
        # we can announce our test user as superuser, so the test user can do anythin via the API.
        # (there are also other tests, using other (on the fly created) users with restricted permissions, which is fine but not part of this generic setup.
        from django.contrib.auth.models import User
        user, _ = User.objects.get_or_create(username=self.ldap_server.dbcreds.user)
        user.is_superuser = True
        user.save()

    def stop(self):
        self.django_server.stop()
        self.ldap_server.stop()
        self.database.destroy()
        self.client_credentials.destroy()

    def __enter__(self):
        try:
            self.start()
        except Exception as e:
            logger.error(e)
            self.stop()
            raise
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def create_tmss_client(self):
        return TMSSsession.create_from_dbcreds_for_ldap(self.client_credentials.dbcreds_id)

def main_test_database():
    """instantiate, run and destroy a test postgress django database"""
    os.environ['TZ'] = 'UTC'
    logging.basicConfig(format = '%(asctime)s %(levelname)s %(message)s', level = logging.INFO)

    with TMSSTestDatabaseInstance() as db:
        # print some nice info for the user to use the test servers...
        # use print instead of log for clean lines.
        for h in logging.root.handlers:
            h.flush()
        print()
        print()
        print("**********************************")
        print("Test-TMSS database up and running.")
        print("**********************************")
        print("DB Credentials ID: %s (for example to run tmms against this test db, call 'tmss -C %s')" % (db.dbcreds_id, db.dbcreds_id))
        print()
        print("Press Ctrl-C to exit (and remove the test database automatically)")
        waitForInterrupt()

def main_test_environment():
    """instantiate, run and destroy a full tmss test environment (postgress database, ldap server, django server)"""
    from optparse import OptionParser, OptionGroup
    os.environ['TZ'] = 'UTC'

    parser = OptionParser('%prog [options]',
                          description='setup/run/teardown a full TMSS test environment including a fresh and isolated database, LDAP server and DJANGO REST server.')
    parser.add_option("-H", "--host", dest="host", type="string", default='127.0.0.1',
                      help="expose the TMSS Django REST API via this host. [default=%default]")
    parser.add_option("-p", "--port", dest="port", type="int", default=find_free_port(8000),
                      help="try to use this port for the DJANGO REST API. If not available, then a random free port is used and logged. [default=%default]")
    group = OptionGroup(parser, 'Messaging options')
    group.add_option('-b', '--broker', dest='broker', type='string', default=DEFAULT_BROKER, help='Address of the message broker, default: %default')
    group.add_option('-e', "--exchange", dest="exchange", type="string", default=DEFAULT_BUSNAME, help="Bus or queue where the TMSS messages are published. [default: %default]")
    parser.add_option_group(group)
    (options, args) = parser.parse_args()

    logging.basicConfig(format = '%(asctime)s %(levelname)s %(message)s', level = logging.INFO)

    with RATestEnvironment(exchange=options.exchange, broker=options.broker):
        with TMSSTestEnvironment(host=options.host, preferred_django_port=options.port, exchange=options.exchange, broker=options.broker) as instance:
            # print some nice info for the user to use the test servers...
            # use print instead of log for clean lines.
            for h in logging.root.handlers:
                h.flush()
            print()
            print()
            print("*****************************************************")
            print("Test-TMSS database, LDAP and Django up and running...")
            print("*****************************************************")
            print("DB Credentials ID: %s" % (instance.database.dbcreds_id, ))
            print("LDAP Credentials ID: %s" % (instance.django_server.ldap_dbcreds_id, ))
            print("TMSS Client Credentials ID: %s" % (instance.client_credentials.dbcreds_id, ))
            print("Django URL: %s" % (instance.django_server.url))
            print()
            print("Example cmdlines to run tmss or tmss_manage_django:")
            print("TMSS_DBCREDENTIALS=%s TMSS_LDAPCREDENTIALS=%s tmss" % (instance.database.dbcreds_id, instance.django_server.ldap_dbcreds_id))
            print("TMSS_DBCREDENTIALS=%s TMSS_LDAPCREDENTIALS=%s tmss_manage_django" % (instance.database.dbcreds_id, instance.django_server.ldap_dbcreds_id))
            print()
            print("Example cmdline to run tmss client call:")
            print("TMSS_CLIENT_DBCREDENTIALS=%s tmss_set_subtask_state <id> <state>" % (instance.client_credentials.dbcreds_id, ))
            print()
            print("Press Ctrl-C to exit (and remove the test database and django server automatically)")
            waitForInterrupt()


if __name__ == '__main__':
    main_test_environment()